#if !defined(ptf_h)
//! \internal header guard
#define ptf_h

//! \file ptf.h
//! \brief Basic functions to deal with the low-level structure of PTF 
//! container files on disk.
//!
//! \note General note: PTF files are little endian.
//!
//! For more information on how to start working with PTF, see \ref ptf_intro "PTF Introduction".

#include <stdlib.h>

//! \addtogroup PtfFormat The PTF File Format
//! @{
//!
//! \copydoc ptf.h

#ifdef WIN32
  #if defined(PTF_EXPORTS)
    #define PTF_API __declspec(dllexport)
  #endif
#endif
#ifndef PTF_API
  #define PTF_API
#endif

#if defined(__cplusplus)
extern "C" {
#endif

//  I know that MSVC and x86 GCC has a 64-bit long long type and 32-bit int type.
#if defined(_MSC_VER) || defined(__GNUC__) || defined(DOXYGEN_SKIP)
//! ptf_64 is known to be 64 bits unsigned, and stored as such in the file.
//! The element is stored little-endian in the file, as all elements.
typedef unsigned long long ptf_64;
//! ptf_32 is known to be 32 bits unsigned, and stored as such in the file.
//! The element is stored little-endian in the file, as all elements.
typedef unsigned int ptf_32;
//! ptf_type stores a well-defined data type identifier (as 32 bits).
//! The element is stored little-endian in the file, as all elements.
typedef unsigned int ptf_type;
//! ptf_16 is known to be 16 bits unsigned, and stored as such in the file.
//! The element is stored little-endian in the file, as all elements.
typedef unsigned short ptf_16;
//! ptf_8 is known to be 8 bits unsigned, and stored as such in the file.
typedef unsigned char ptf_8;
//! ptf_flag contains the values 0, or non-0.
typedef unsigned char ptf_flag;
//! ptf_char is known to be 8 bits unsigned, and stored as such in the file.
//! Text is stored as UTF-8 in the file, and may need translation to wchar_t 
//! when read/written. Strings should always be 0 terminated (and as such, 
//! a string of size X can use at most X-1 bytes). With UTF-8, bytes are 
//! different from characters or glyphs.
typedef unsigned char ptf_char;
#else
#error "please port me"
#endif


//! Dates are stored using western calendar, as UST, split into fields.
struct ptf_date {
  ptf_16                      year;       //!< 2006 for the year 2006
  ptf_8                       month;      //!< 1 for January
  ptf_8                       day;        //!< 1 for the 1st of the month
  ptf_8                       hour;       //!< 00 through 23
  ptf_8                       minute;     //!< 00 through 59
  ptf_8                       second;     //!< 00 through 59
  ptf_8                       hundredth;  //!< 00 through 99
};
typedef struct ptf_date ptf_date;
//! In the PTF octree, nodes can be identified by "cube extents".
//! This means the leftmost and rightmost coordinates within the cube 
//! tree space (described elsewhere).
struct ptf_ocube {
  ptf_32                      x_min;      //!< Left edge of the cube in X dimension
  ptf_32                      x_max;      //!< Right edge of the cube in X dimension
  ptf_32                      y_min;      //!< Bottom edge of the cube in Y dimension
  ptf_32                      y_max;      //!< Top edge of the cube in Y dimension
  ptf_32                      z_min;      //!< Yon edge of the cube in Z dimension
  ptf_32                      z_max;      //!< Hither edge of the cube in Z dimension
};
typedef struct ptf_ocube ptf_ocube;
//! Constants used in the data type declarations.
enum ptf_constants {
  //! The byte size of a fixed-length string of ptf_char.
  PTF_NAMESIZE = 48,
  //! The byte size of a metadata value.
  PTF_METAVALUESIZE = 80,
  //! The byte size of a full path. I have never seen actual paths this long.
  PTF_PATHSIZE = 256,
  //! Version of the specification.
  PTF_SPECIFICATION_VERSION = 0x10000,
  //! Minimum version needed to understand (at least parts of) the file.
  PTF_MIN_VERSION = 0x10000,
  //! Value for the 'compression' member of record ref
  PTF_COMPRESSION_NONE = 0,
  //! Value for the 'compression' member of record ref.
  //! The compression method is Zlib deflate compression, using 
  //! the maximum (9) compression level setting.
  PTF_COMPRESSION_ZLIB = 1,
  //! Value for the 'encryption' member of record ref
  PTF_ENCRYPTION_NONE = 0,
  //! Value for the 'encryption' member of record ref. 
  //! The encryption protocol is X-Tea with 128 bit key and 32 rounds
  //! in CBC mode. The Initialization Vector is the 64-bit Tiger hash 
  //! of the string formed by concatenating the encryption key with the 
  //! 64-bit Tiger hash of the data.
  PTF_ENCRYPTION_XTEA128_32 = 1,
};
//! ptf_name is used for storing names of "reasonable" length, such as 
//! user names, file names, etc when records need to be fixed size. 
//! Longer pieces of text are stored as text records that are just arrays of ptf_char.
typedef ptf_char ptf_name[PTF_NAMESIZE];
//! ptf_value is used for storing values of "reasonable" length, such 
//! as street addresses, proper names, web URLs etc when records need 
//! to be fixed size.
//! Longer pieces of text are stored as text records that are just arrays of ptf_char.
typedef ptf_char ptf_value[PTF_METAVALUESIZE];
//! ptf_path is used for storing path names. Path names are always normalized 
//! to forward-slash format with letter-colon transformed to slash-letter. 
//! Thus, the Windows path C:\WINDOWS\SYSTEM would be stored as /C/WINDOWS/SYSTEM.
//! File paths should be case correct, even if the underlying file system is 
//! case insensitive. I e, creating applications should make efforts to preserve 
//! the actual case of the file being referenced, so that transport to a case 
//! sensitive file system is simplified.
typedef ptf_char ptf_path[PTF_PATHSIZE];

//! A position is stored in geocentric coordinates, with X out to 
//! the right (approx 90 degree east), Z out hither (Greenwich meridian)
//! and Y up (through the approximate north pole). The position stores 
//! how to get from object space to global space.
//! If you want another coordinate space, you can multiply in whatever 
//! static transform takes the octree world space into your desired 
//! space.
//! \see PTF_MIN_GEOPOS, PTF_MAX_GEOPOS
struct ptf_geopos {
  double           x;
  double           y;
  double           z;
};

//!  Minimum is inclusive. \see ptf_geopos
#define PTF_MIN_GEOPOS (-(double)(64*1024*1024))
//!  Maximum is exclusive (half-open interval). \see ptf_geopos
#define PTF_MAX_GEOPOS ((double)(64*1024*1024))



//! An orientation is stored as a normalized quaternion, and is expressed
//! in post-translated space (i e, local to the referenced object's center).
//! The ori stores how to get from object space to global space.
//! \note The quaternion is stored with the vector first, and the scalar 
//! second. About half of the libraries out there store the quaternion the 
//! other way around, so check yours to make sure it matches!
struct ptf_ori {
  float            x;   //!< Vector/Imaginary part 1 (sin theta times axis x)
  float            y;   //!< Vector/Imaginary part 2 (sin theta times axis y)
  float            z;   //!< Vector/Imaginary part 3 (sin theta times axis z)
  float            w;   //!< Scalar (cos theta)
};

//! The magic number for a PTF file is a C string, with a terminating 0.
#define PTF_MAGIC_CHARS "PTF 1.0"

//! Inside a TOC for a given type, records of the file are addressed using 
//! size/offset pairs. "offset" is always a multiple of ptf_file_header::alignment 
//! whereas "size" is the logical size of the data. This means that there will 
//! be unused padding after records whose size are not a multiple of the alignment.
struct ptf_toc_entry {
  ptf_64                      offset;         //!< Where in the file the record lives. If zero, the record is deleted.
                                              //!< \see ptf_disk_file_header
                                              //!< \see PTF_DISK_FILE_OFFSET
  ptf_64                      hash;           //!< Checksummed hash of record data -- 0 means data is not checksummed
  ptf_32                      size_disk;      //!< The number of bytes the record occupies on disk.  If zero, the record is empty.
  ptf_32                      size_mem;       //!< The number of bytes the record occupies in memory, after decryption and decompression.
                                              //!< If zero, the record is empty.
  ptf_16                      encryption;     //!< Encryption type used by record -- 0 means no encryption.
                                              //!< Note the generation of the initialization vector in CBC mode.
  ptf_16                      compression;    //!< Compression type used by record -- 0 means no compression.
                                              //!< Note that compression happens before encryption when writing.
  union {
    ptf_32                    reserved;       //!< Reserved for internal use
    ptf_type                  type;           //!< The well-defined type for this entry
  };
};
typedef struct ptf_toc_entry ptf_toc_entry;

//! The meta-TOC is pointed at by the ptf_file_header at the start of the 
//! file, and contains pointers to all the different type TOCs in the file.
//! There is ptf_file_header::meta_count number of metatable elements in 
//! the table. The types may be present in any order within the meta-TOC.
//! Note that the meta-TOC cannot be accessed through the regular TOC/record_ref
//! API, it is entirely managed by the PTF library. Also, the meta-TOC can 
//! have a 0-th record that is not null, which regular TOCs do not.
//!
//! Include <ptf_types.h> to get the definition of standard PTF types.
typedef struct ptf_toc_entry ptf_meta_toc_entry;

//! sizeof(ptf_file_header) is intended to be 256 bytes.
//! This describes the layout of the first 256 bytes of a PTF file, from 
//! which everything else about the file can be traced.
struct ptf_file_header {
  ptf_char                    magic[8];       //!< "PTF 1.0\0"
  ptf_name                    file_name;      //!< canonical name of file
  ptf_32                      spec_version;   //!< what specification version is written
  ptf_32                      min_version;    //!< minimum specification version needed to read
  ptf_date                    cr_date;        //!< date initially created
  ptf_date                    mod_date;       //!< date last modified
  ptf_32                      alignment;      //!< each record is aligned to this # bytes (must be power of two)
  ptf_32                      spare1;         //!< reserved; set to 0
  ptf_toc_entry               meta_toc;       //!< pointer to the meta-TOC
  ptf_ocube                   range;          //!< what range of the world the file covers
  ptf_64                      size;           //!< size of the file, as of last commit
  ptf_char                    schema[32];     //!< what schema is used for data stored in the PTF file (advisory)
  ptf_32                      spare2[18];     //!< reserved for future use; set to 0
};
typedef struct ptf_file_header ptf_file_header;

//! A full reference to a specific file/record is stored using ptf_record_ref.
//! This allows for reference to records in other files, too. For less flexible 
//! references, see types ptf_local_ref and ptf_fixed_ref.
struct ptf_record_ref {
  ptf_32                      file;           //!< The file (using this file's local ref table).
  ptf_32                      spare1;         //!< Reserved for future use; set to 0.
  ptf_type                    type;           //!< The type of the record.
  ptf_32                      index;          //!< The index within that type.
};
typedef struct ptf_record_ref ptf_record_ref;

//! A reference to a record within the current file is stored as ptf_local_record_ref.
//! Such a reference cannot reference data in a file other than the current file.
//! \see ptf_record_ref
struct ptf_local_ref {
  ptf_type                    type;           //!< The type of the record being referenced
  ptf_32                      index;          //!< The index within the TOC for that type
};
typedef struct ptf_local_ref ptf_local_ref;

//! A remote reference of known type can be represented as an 8-byte entity. 
//! This comes in handy for things like the octree, where we know that remote 
//! references are for other octree nodes.
struct ptf_remote_ref {
  ptf_32                      file;           //!< The file (using this file's local ref table).
  ptf_32                      index;          //!< The index within the known type of that file.
};

//! A reference to a record where you know the type, and know that it is local within 
//! the referencing file, can be made using a single index. This is useful when you 
//! store things like arrays of mesh record pointers, for example.
//! \see ptf_record_ref
struct ptf_fixed_ref {
  ptf_32                      index;          //!< The index within the TOC for the type being referenced 
                                              //!< (type which must be given from context in the specific 
                                              //!< reference maker).
};
typedef struct ptf_fixed_ref ptf_fixed_ref;

//!}@

#if defined(__cplusplus)
}
#endif


#endif    //  ptf_h
