
#if !defined(ptf_collection_h)
//! \internal header guard
#define ptf_collection_h

#if !defined(ptf_file_h)
#include "ptf_file.h"
#endif
#if !defined(ptf_types_h)
#include "ptf_types.h"
#endif
//! \addtogroup PtfGeometry The PTF Geometry Library

//! A collection is a mini-octree for a mesh that has been split and 
//! simplified into a piecewise multi-resolution mesh.
//! All collection instances are identified with a ptf_collection_handle. 
//! There are functions to read a collection from a PTF file, and 
//! write a collection handle to a PTF file. The specific data layout is 
//! documented in <ptf_types.h>.
typedef class ptf_collection_t *ptf_collection_handle;

//! Information about records within a collection.
struct ptf_crec_t {
  ptf_crec_t      * parent;           //!< Pointer to parent of this node.  If this is a root node, NULL.
  ptf_crec_t      * children;         //!< Pointer to first child of this node.  If no children, NULL. All children are contiguous.
  ptf_32            left, bottom, yon;//!< Extent of this record
  ptf_16            num_children;     //!< Number of children this node has.
  ptf_16            child_idx;        //!< The child index this node represents within its parent.
  ptf_16            level;            //!< The octree level this node represents.
  ptf_16            reserved1;        //!< Not used; write as 0.
  float             feature_size;     //!< Feature size of the mesh represented by this node.
  ptf_remote_ref    mesh;             //!< The mesh this node represents (actual type comes from "type" in header)
};

//! Create a collection instance. You need an instance to be able to read data 
//! into it, or add data and write out to a file. A collection instance organizes 
//! a set of internally related geometries, split into an octree of content where 
//! deeper levels mean higher-resolution data. The geometry is typically a mesh, 
//! but can also be collision data, or something else. It's also allowed for a 
//! collection to not contain data at a given level, which means you have to look 
//! further down the tree to get data for that area.
//! \param o_coll The collection instance handle will be stored here. You must 
//! call ptf_collection_destroy() on it once you're done with it.
//! \param name The name to assign to this collection instance.
//! \param type The type of data stored in the collection (ptMeshData or 
//! ptOpcodeCollisionData, typically).
//! \return ptf_err status.
PTF_API ptf_err ptf_collection_create(ptf_collection_handle *o_coll, 
    ptf_char const * name, ptf_type type);

//! Deallocate all data allocated by a collection instance.
//! \param coll The allocated instance to destroy. It will not be valid 
//! after this call returns.
//! \return ptf_err status.
PTF_API ptf_err ptf_collection_destroy(ptf_collection_handle coll);

//! Return a copy of a collection's header data.
//! \param coll The allocated collection instance.
//! \param hdr Pointer to a collection header structure to populate.
//! \param o_related A pointer to an array of related ptf_record_ref will be stored 
//! here. For example, a collision collection may have the visual collection that 
//! goes with it as a related record. This pointer is only valid until the next 
//! operation on the collection.
//! \param o_related_count The number of related ptf_record_ref pointed at by 
//! o_related is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_collection_get_header(ptf_collection_handle coll, 
    ptf_collection_header * hdr, ptf_record_ref const **o_related, ptf_32 *o_related_count);

//! Set the array of related ptf_record_ref for a collection. This replaces 
//! any previous array.
//! \param coll The collection to update.
//! \param related A pointer to an array of ptf_record_ref to store in the collection.
//! \param count The number of related ptf_record_ref pointed at by related. If 0, 
//! remove any related ptf_record_ref.
//! \return ptf_err status
PTF_API ptf_err ptf_collection_set_related(ptf_collection_handle coll,
    ptf_record_ref const *related, ptf_32 count);

//! Create the top-level of a collection.
//! \param coll The instance handle to add this node to.
//! \param f_size The feature size this node represents.
//! \param mesh Reference to the mesh contained within this node.
//! \param bounds The octree bounds of this node.  Note that this does not need
//  to represent a valid octree cube, but must align on octree cube boundaries.
//  The cube must be cubical, and must have an edge length that is a power of 2.
//  A collection node with an edge of 256 units must have a bottom,left,yon
//  position that is a multiple of 128.  Likewise for 1024 lining up on 512, etc.
//! \param o_id The ID to use when identifying this node in calls to ptf_collection_node_add.
//! \return ptf_err status.
PTF_API ptf_err ptf_collection_set_root(ptf_collection_handle coll, 
    float f_size, ptf_remote_ref mesh, ptf_ocube bounds, ptf_32 *o_id);

//! Add a child node to a collection.
//! \param coll The instance handle to add this node to.
//! \param parent_id The ID of the node this node descends from.
//! \param f_size The feature size this node represents.
//! \param mesh Reference to the mesh contained within this node.
//! \param child_idx Which child this node represents.  See ptf_child_index in ptf_types.h.
//! \param o_id The ID to use when identifying this node in calls to ptf_collection_node_add.
//! \return ptf_err status.
PTF_API ptf_err ptf_collection_node_add(ptf_collection_handle coll, 
    ptf_32 parent_id, float f_size, ptf_remote_ref mesh, ptf_8 child_idx, ptf_32 *o_id);

//! Returns the collection tree as a flat array.  Subsequent additions to the collection
//! or calls to ptf_collection_get_records will invalidate this array. The first record 
//! (at index 0) is the root.
//! \param coll The instance handle to get the array from.
//! \param o_recs The array of ptf_crec_t records representing this collection tree.
//! \param o_num_recs The total number of entries in the array, including the root entry (index 0).
//! \return ptf_err status.
PTF_API ptf_err ptf_collection_get_records(ptf_collection_handle coll, 
    ptf_crec_t **o_recs, ptf_32 *o_num_recs);

//! Read a ptCollectionData block from a PTF file and populate the collection instance 
//! handle with the data found therein.
//! \param coll The collection handle to populate with data from the file.
//! \param file The PTF file to read from.
//! \param ref The ptCollectionData type record index to read from.
//! \return ptf_err status.
PTF_API ptf_err ptf_collection_read_from_file(ptf_collection_handle coll, 
    ptf_file_handle file, ptf_fixed_ref const &fref);

//! Given a populated collection handle, and an existing collection data record in a 
//! PTF file, replace that record with the data of the populated collection.
//! The collection in memory is left intact; this is a pure "write" operation.
//! \param coll The collection to write to file.
//! \param file The file to receive the collection data.
//! \param ref The index of the ptCollectionData record that will receive the data.
//! \return ptf_err status.
PTF_API ptf_err ptf_collection_update_to_file(ptf_collection_handle coll, 
    ptf_file_handle file, ptf_fixed_ref const &ref);

//! Given a populated collection handle, append a new ptCollectionData record to the 
//! given PTF file, and return the index of this new record.
//! \param coll The collection handle to write to the file.
//! \param file The file to receive the collection handle.
//! \param o_ref The index of the collection record will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_collection_append_to_file(ptf_collection_handle coll, 
    ptf_file_handle file, ptf_fixed_ref * o_ref);

#endif  //  ptf_mesh_h
