#if !defined(ptf_content_h)
//! \internal header guard
#define ptf_content_h

#if !defined(ptf_file_h)
#include "ptf_file.h"
#endif
//! \addtogroup PtfGeometry The PTF Geometry Library
//!
//! For more information about the thoughts behind storing of content 
//! in PTF files, see the sections:
//! \ref terrainmaterial
//! \ref terainoctree
//! \ref terrainoctreecontent
//! \ref terrainoctreeraster
//! @{

//! PTF content records are lists of different kinds of content, 
//! with per-instance transforms. A ptf_content_handle is a handle 
//! to the representation of this record, which can serialize and 
//! deserialize to/from PTF blocks.
//!
//! Usually, the content records are referenced by the ptf_octree, 
//! using the ptf_onode_ref::content_index field. However, a content 
//! record can exist stand-alone (without any tie to a PTF file).
//! You will manually call read and write to transfer between a file 
//! and the ptf_content_handle; applied changes are not automatically 
//! serialized back.
//!
//! Content records reference specific geometries (trees, earth skin 
//! blocks, etc); each such reference has a position and an 
//! orientation, which is expected to move that geometry into the 
//! bounds of the octree node that references the content.
//!
//! As a peculiarity, ptf content objects cannot have record index 0; 
//! it is reserved in the file format.
typedef struct ptf_content_t *ptf_content_handle;

//! Create a new content instance in memory.
//! You must dispose this instance by calling ptf_content_destroy().
//! \param o_cont The created content.
//! \return ptf_err status.
PTF_API ptf_err ptf_content_create(ptf_content_handle * o_cont);

//! Dispose an in-memory content record.
//! \param cont The handle to dispose.
//! \return ptf_err status.
PTF_API ptf_err ptf_content_destroy(ptf_content_handle cont);

//! Read a PTF file record into an existing content handle. Any 
//! previous contents of that handle will be disposed. The type 
//! of the record will be ptOctreeContent.
//! \param cont The content handle to read into.
//! \param fh The file to read from.
//! \param ref The content record index to read.
//! \return ptf_err status.
PTF_API ptf_err ptf_content_read_from_file(ptf_content_handle cont, 
    ptf_file_handle fh, ptf_fixed_ref const & ref);

//! Write a PTF file record into an existing record (replacing 
//! that records contents). This does not flush the file.
//! \param cont The contents to write out.
//! \param fh The file to write to.
//! \param ref The specific record to replace (of type ptOctreeContent).
//! \return ptf_err status.
PTF_API ptf_err ptf_content_update_to_file(ptf_content_handle cont, 
    ptf_file_handle fh, ptf_fixed_ref const & ref);

//! Store a copy of the given content handle in an entirely new PTF 
//! file record. This does not flush the file.
//! \param cont The handle to store.
//! \param fh The file to write to.
//! \param o_ref The index of the created record (will never be 0).
//! The type will always be ptOctreeContent.
PTF_API ptf_err ptf_content_append_to_file(ptf_content_handle cont, 
    ptf_file_handle fh, ptf_fixed_ref * o_ref);

//! Retrieve the information for this content record, if any. 
//! \param cont The record handle to retrieve data for.
//! \param o_ptr A pointer to the data is returned here. This will be 
//! valid until the record is disposed, or a call to ptf_content_set_annotation().
//! \param o_size The size of the data is returned here.
//! \param return ptf_err status.
PTF_API ptf_err ptf_content_get_annotation(ptf_content_handle cont, 
    void const ** o_ptr, ptf_32 * o_size, ptf_type * o_type);

//! Each content record can store arbitrary annotation data about the content. 
//! The format of this info is not yet defined, so this is only for 
//! experimental use. Set the type to a value in the "experimental" 
//! range, until such time as a well-defined format can be established.
//! To delete the annotation, set a 0-size data block.
//! \param cont The content record to update the information for.
//! \param ptr Pointer to the data to copy into the record, replacing 
//! any previous such data. The first four bytes must be 0.
//! \param size The size of the data (including the 0 bytes).
//! return ptf_err status.
PTF_API ptf_err ptf_content_set_annotation(ptf_content_handle cont, 
    void const * ptr, ptf_32 size, ptf_type type);

//! Each content record rerefences some number of geometries of various 
//! kind and source (e g, a tree from some external library file, or a 
//! building from another record in this file). Count how many separate 
//! geometries are being referenced.
//! \param cont The content record to count references in.
//! \param o_count The number of references in the record.
//! \return ptf_err status.
PTF_API ptf_err ptf_content_count_references(ptf_content_handle cont, 
    ptf_32 * o_count);

//! You can iterate through the references in a content record.
//! \param cont The record to iterate over.
//! \param index The index (between 0, inclusive, and count, exclusive, 
//! from ptf_content_count_references).
//! \param o_type The record reference for the goemetry being referenced.
//! \return ptf_err status.
PTF_API ptf_err ptf_content_get_reference(ptf_content_handle cont, 
    ptf_32 index, ptf_record_ref * o_type);

//! Add a new geometry reference to a content record. The record 
//! reference does not have to be unique -- you can have the same 
//! record_ref added twice to the same block, with different instances. 
//! This makes sense, beacuse the additional information is per 
//! reference, not per instance.
//! The typical types of records added as content include
//! ptMesh, ptRaster and ptCollection, but could be anything. Readers 
//! will skip over content references of types they are not prepared to 
//! handle.
//! \note One special kind of content is a ptOctreeNode record_ref, which means 
//! to replace all descendents of this octree node with the descendents 
//! of that other octree node (possibly in a separate PTF file).
//! The size and location of the referenced octree node should be the same 
//! as the size and location of the replaced octree node.
//! The number of instances of that ref is ignored. If there is more than one 
//! ptOctreeNode reference, the contents of those references are unioned together, 
//! but the content of the referencing node (and its children) is still replaced.
//! \param cont The content handle to add to.
//! \param ref The geometry reference to add to the handle.
//! \param o_index The index allocated for the reference.
//! \return ptf_err status.
PTF_API ptf_err ptf_content_add_reference(ptf_content_handle cont, 
    ptf_record_ref const & ref, ptf_32 * o_index);

//! Remove a specific reference from the content (and all its instances, 
//! if any).
//! \note This may re-order the other references, and will definitely 
//! re-number some of them in an unpredictable fashion.
//! \param cont The content record to edit.
//! \param index The index of the reference to remove.
//! \return ptf_err status.
PTF_API ptf_err ptf_content_remove_reference(ptf_content_handle cont,
    ptf_32 index);

//! Get the metadata record reference from a content record.
//! \param cont The record to get metadata from.
//! \param index The index (between 0, inclusive, and count, exclusive, 
//! from ptf_content_count_references).
//! \param o_ref The metadata reference to populate.
//! \return ptf_err status.
PTF_API ptf_err ptf_content_get_reference_meta_data(ptf_content_handle cont, 
    ptf_32 index, ptf_record_ref * o_ref);

//! Set the metadata record reference on a content record. The metadata 
//! reference record points to a record specifying well-defined metadata 
//! about all the instances of this record. This can be used, for example, 
//! to define "layers" of data, if you share the same record_ref between 
//! many references.
//! \param cont The record to add metadata to.
//! \param index The index (between 0, inclusive, and count, exclusive, 
//! from ptf_content_count_references).
//! \param ref The metadata reference.
//! \return ptf_err status.
PTF_API ptf_err ptf_content_set_reference_meta_data(ptf_content_handle cont, 
    ptf_32 index, ptf_record_ref const & ref);

//! References can have an arbitrary information block attached to them. 
//! The format of this information is not currently defined. If you use 
//! this feature, make sure to set the type to something in the experimental 
//! range, until a well-defined format can be established.
//! \param cont The content record to get data from.
//! \param index The index of the reference to get data for (if present).
//! \param o_ptr A pointer to the data will be returned here. This pointer 
//! will be valid until the next mutation operation on cont.
//! \param o_size The size of the data will be returned here.
//! \param o_type The format of the annotation will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_content_get_reference_annotation(ptf_content_handle cont, 
    ptf_32 index, void const ** o_ptr, ptf_32 * o_size, ptf_type * o_type);

//! Add a block of comment data to a specific reference. If you need 
//! unique comments for each instance, you have to create a new reference 
//! per instance needing unique comments.
//!
//! If the annotation were meta-data, then the type for this block is ptLocalRef, 
//! which points at a record containing sideband data for the instantiations of 
//! this reference (all instantiations share the same sideband data). Typically, 
//! the type of that local_ref will be ptGeometryInstanceMetaData, which 
//! allows you to tag the instances with a UUID, and a SEDRIS classification.
//! However, that special case is taken care of by the special-case reference_meta_data 
//! functions above. reference_annotation() allows you to annotate the reference 
//! with additional "free-form" data.
//!
//! To delete a reference, set a 0-size data block.
//! \param cont The content record to edit.
//! \param index The index of the reference to annotate.
//! \param ptr A pointer to the data to copy into the record.
//! \param size The size of the data to copy.
//! \param type The format of the annotation.
//! \return ptf_err status.
PTF_API ptf_err ptf_content_set_reference_annotation(ptf_content_handle cont, 
    ptf_32 index, void const * ptr, ptf_32 size, ptf_type type);

//! Count how many individual instances of a given geometry reference 
//! exist within this content handle. Each instance has its own 
//! position and orientation.
//! \param cont The content handle to count instances in.
//! \param index The index of the reference to count instances of.
//! \param o_count The number of references is returned here.
//! return ptf_err status.
PTF_API ptf_err ptf_content_count_reference_instances(ptf_content_handle cont, 
    ptf_32 index, ptf_32 * o_count);

//! Retrieve a specific content reference instance.
//! \param cont The content handle to retrieve from.
//! \param index The reference index to retrieve from.
//! \param instance The specific instance index to retrieve.
//! \param o_pos The position of the instance in world coordinates.
//! \param o_ori The orientation of the instance.
//! return ptf_err status.
PTF_API ptf_err ptf_content_get_reference_instance(ptf_content_handle cont, 
    ptf_32 index, ptf_32 instance, ptf_geopos * o_pos, ptf_ori * o_ori);

//! Add a new instance of a given reference to the content.
//! \param cont The content handle to edit.
//! \param index The reference to add an instance of.
//! \param pos The position of the instance, in global coordinates.
//! \param ori The orientation of the instance.
//! \param o_instance The index of the added instance is returned here.
//! Note that instance indices are not persistent across mutations of 
//! the content record.
//! return ptf_err status.
PTF_API ptf_err ptf_content_add_reference_instance(ptf_content_handle cont, 
    ptf_32 index, ptf_geopos const & pos, ptf_ori const & ori, ptf_32 * o_instance);

//! Update an existing instance of a given reference to the content.
//! \param cont The content handle to edit.
//! \param index The reference to add an instance of.
//! \param instance The index of the instance of the reference.
//! \param pos The position of the instance, in global coordinates.
//! \param ori The orientation of the instance.
//! return ptf_err status.
PTF_API ptf_err ptf_content_update_reference_instance(ptf_content_handle cont, 
    ptf_32 index, ptf_32 instance, ptf_geopos const & pos, ptf_ori const & ori);

//! @}
#endif  //  ptf_content_h
