#if !defined(ptf_raster_h)
//! \internal header guard
#define ptf_raster_h

#include "ptf_types.h"
#include "ptf_file.h"
//! \addtogroup PtfGeometry The PTF Geometry Library
//! @{

//! All raster elevation data instances are identified by an abstract ptf_dem_handle.
//! In this case, "dem" refers to a generic Digital Elevation Model, not necessarily
//! the USGS DEM file format.  The terms "dem" and "raster elevation" are used interchangably
//! throughout this API.  Elevation data instances can exist separate from any ptf_file instance, 
//! and it is up to the programmer to manage the lifetime of these objects. The API does provide 
//! functions to read and write these instances from/to PTF files.
typedef class ptf_dem_t *ptf_dem_handle;

//! In-memory structure for passing elevation sample values to/from the PTF elevation raster API.
typedef union {
  char          int8;
  short         int16;
  float         float32;
  ptf_8         uint8;
  ptf_16        uint16;
} ptf_dem_val;

//! Create a raster elevation instance. You need an instance to be able to read data 
//! into it, or add data and write out to a file.
//! \param o_dem The raster elevation instance handle will be stored here. You must 
//! call ptf_content_destroy() on it once you're done with it.
//! \param name A string used as an identifier for this instance.
//! \return ptf_err status.
PTF_API ptf_err ptf_dem_create(ptf_dem_handle *o_dem, ptf_char const * name);

//! Deallocate all data allocated by a raster elevation instance.
//! \param dem The allocated instance to destroy. The handle will not be valid 
//! after this call returns.
//! \return ptf_err status.
PTF_API ptf_err ptf_dem_destroy(ptf_dem_handle dem);

//! Read a ptElevationData block from a PTF file and populate the raster elevation instance 
//! handle with the data found therein.
//! \param dem The dem handle to populate with data from the file.
//! \param file The PTF file from which to read.
//! \param ref The ptElevationData type record index from which to read.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_read_from_file(ptf_dem_handle dem, 
    ptf_file_handle file, ptf_fixed_ref const &ref);

//! Given a populated dem handle, and an existing dem data record in a 
//! PTF file, replace that record with the data of the populated dem.
//! The dem in memory is left intact; this is a pure "write" operation.
//! \param dem The dem to write to file.
//! \param file The file to receive the dem data.
//! \param ref The index of the ptElevationData record that will receive the data.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_update_to_file(ptf_dem_handle dem, 
    ptf_file_handle file, ptf_fixed_ref const &ref);

//! Given a populated raster elevation handle, append a new ptElevationData record to the 
//! given PTF file, and return the index of this new record.
//! \param dem The dem handle to write to the file.
//! \param file The file to receive the dem handle.
//! \param o_ref The index of the dem record will be returned here.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_append_to_file(ptf_dem_handle dem, 
    ptf_file_handle file, ptf_fixed_ref *o_ref);

//! Retrieve the identifier name for this instance.
//! \param dem The dem to query.
//! \param name Identifier name for this instance. 
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_get_name( ptf_dem_handle dem, ptf_name name );

//! Set the number of rows and columns for the given dem handle, as well as the data type for the underlying
//! elevation data.  Elevation instances can have 8-bit integer, 16-bit integer, or 32-bit floating point
//! samples.  This function should typically be called when a new raster elevation instance is created.  
//! Calling this function for a populated instance will have undefined effects on the instance's data.
//! \param dem The dem to modify.
//! \param cols The number of columns to be used by this dem.
//! \param rows The number of rows to be used by this dem. 
//! \param type The data type for samples in this dem.  Supported types are: pvtInt8, pvtInt16, and pvtFloat32.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_set_col_row_type( ptf_dem_handle dem, ptf_16 cols, ptf_16 rows, ptf_8 type );

//! Retrieve the number of rows and columns for the given dem handle.
//! \param dem The dem to query.
//! \param rows The number of rows used by this dem. 
//! \param cols The number of columns used by this dem.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_get_col_row( ptf_dem_handle dem, ptf_16 * cols, ptf_16 * rows );

//! Retrieve the storage type for samples in this raster elevation instance.  
//! \param dem The dem to query.
//! \param type Storage type for samples in this dem.  Valid types are: pvtInt8, pvtInt16, and pvtFloat32.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_get_sample_type( ptf_dem_handle dem, ptf_8 * type );

//! Set the geographic extents of this raster elevation instance.  Extents are measured from the
//! projection center of the instance.
//! \param dem The dem to write to file.
//! \param half_width Distance from instance center to east/west edge, in meters.
//! \param half_height Distance from instance center to north/south edge, in meters.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_set_extents( ptf_dem_handle dem, float half_width, float half_height );

//! Retrieve geographic extents for this raster elevation instance.
//! \param dem The dem to query.
//! \param half_width Distance from instance center to east/west edge, in meters.
//! \param half_height Distance from instance center to north/south edge, in meters.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_get_extents( ptf_dem_handle dem, float * half_width, float * half_height );

//! Set the projection origin this raster elevation instance.  Projection is Lambert Azimuthal Equal Area
//! \param dem The dem to modify.
//! \param lon Projection origin longitude, in degrees.
//! \param lat Projection origin latitude, in degrees.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_set_proj_org( ptf_dem_handle dem, float lon, float lat );

//! Retrieve projection origin this raster elevation instance.
//! \param dem The dem to query.
//! \param lon Projection origin longitude, in degrees.
//! \param lat Projection origin latitude, in degrees.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_get_proj_org( ptf_dem_handle dem, float * lon, float * lat );

//! Set "null data" value for this instance.  The null value is used to indicate samples in the dem
//! that do not contain any data.
//! \param dem The dem to write to file.
//! \param sentinel Null value for this instance.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_set_null_data( ptf_dem_handle dem, ptf_dem_val sentinel );

//! Retrieve the "null data" value for this instance.  
//! \param dem The dem to query.
//! \param result Indicates whether this dem has an established Null data value.
//! \param sentinel Null value for this instance.  Only set when result is "true".
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_get_null_data( ptf_dem_handle dem, bool * result, ptf_dem_val * sentinel );

//! Set the X, Y, and Z scale value for this instance.  X and Y scale adjust the extents.  Z scale
//! scales the underlying sample values (e.g. to provide larger dynamic range at the expense of resolution).
//! \param dem The dem to modify.  Scale of 1.0 
//! \param x_scale Scale for east-west direction.
//! \param y_scale Scale for north-south direction.
//! \param z_scale Scale for elevation samples.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_set_scale( ptf_dem_handle dem, float x_scale, float y_scale, float z_scale );

//! Retrieve the X, Y, and Z scale value for this instance.
//! \param dem The dem to query.
//! \param x_scale Scale for east-west direction.
//! \param y_scale Scale for north-south direction.
//! \param z_scale Scale for elevation samples.
//! \return ptf_err status.
//! \see ptf_elev_header
PTF_API ptf_err ptf_dem_get_scale( ptf_dem_handle dem, float * x_scale, float * y_scale, float * z_scale );

//! Set the data offset for this instance.  This offset is applied to the underlying sample values.  It is
//! especially useful when dealing with unsigned samples.  Sample value = ( raw value * scale ) + offset
//! \param dem The dem to modify.
//! \param z_offset Height offset of underlying sample data.
//! \return ptf_err status.
PTF_API ptf_err ptf_dem_set_data_offset( ptf_dem_handle dem, float z_offset );

//! Retrieve the data offset for this instance.  This offset is applied to the underlying sample values.  
//! It is especially useful when dealing with unsigned samples.  Sample value = ( raw value * scale ) + offset
//! \param dem The dem to query.
//! \param z_offset Height offset of underlying sample data.
//! \return ptf_err status.
PTF_API ptf_err ptf_dem_get_data_offset( ptf_dem_handle dem, float *z_offset );

//! Retrieve the min and max sample values for this instance.
//! \param dem The dem to query.
//! \param min Minimum post value (excluding NULL data).
//! \param max Maximum post value.
//! \return ptf_err status.
PTF_API ptf_err ptf_dem_get_min_max( ptf_dem_handle dem, ptf_dem_val * min, ptf_dem_val * max );

//! Sets the sample value at the specified row/column.  Sample is assumed to be "raw", that is without
//! the z scale or z offset applied.
//! \param dem The dem to modify.
//! \param row The row at which to set the sample
//! \param col The column at which to set the sample
//! \param sample The new sample value.
//! \return ptf_err status.
//! 
PTF_API ptf_err ptf_dem_set_sample_val( ptf_dem_handle dem, ptf_16 col, ptf_16 row, ptf_dem_val sample );

//! Retrieves the sample value at the specified row/column.  Sample is assumed to be "raw", that is without
//! the z scale or z offset applied.
//! \param dem The dem to query.
//! \param row The row from which to retrieve the sample
//! \param col The column from which to retrieve the sample
//! \param sample The sample value.
//! \return ptf_err status.
//!
PTF_API ptf_err ptf_dem_get_sample_val( ptf_dem_handle dem, ptf_16 col, ptf_16 row, ptf_dem_val * sample );

//! 
//! Sets an array of sample values.  Data in buffer must be arranged in row-order.  Post data in buffer
//! must be the same data type as the DEM.  Samples are assumed to be "raw", that is without
//! the z scale or z offset applied.
//! \param dem The dem to modify.
//! \param buffer The array containing the sample values.
//! \param startCol DEM column where data in buffer starts.
//! \param startRow DEM row where data in buffer starts.
//! \param width Number of columns of data in buffer.
//! \param height Number of rows of data in buffer.
//! \return ptf_err status.
//! 
PTF_API ptf_err ptf_dem_set_sample_buffer( ptf_dem_handle dem, void * buffer, ptf_16 startCol, ptf_16 startRow, 
                                          ptf_16 width, ptf_16 height, ptf_16 buf_stride );

//! Gets an array of sample values.  Data in buffer will be returned in row-order, with the southwestern-most 
//! post first in the buffer and the northeastern-most post last.  Post data in buffer will be the same data 
//! type as the DEM.  Buffer must be large enough to fit all of the data:
//! width * height * sizeof(type), where "type" is the underlying data type for the samples.
//! See ptf_dem_get_sample_type().
//! Samples are assumed to be "raw", that is without the z scale or z offset applied.
//! \param dem The dem to query.
//! \param buffer The array in which the sample values will be returned.
//! \param startCol DEM column to begin retrieving data.
//! \param startRow DEM row to begin retrieving data.
//! \param width Number of columns of data to retrieve.
//! \param height Number of rows of data to retrieve.
//! \param buf_stride Offset within the buffer to the next row.
//! \return ptf_err status.
//! 
PTF_API ptf_err ptf_dem_get_sample_buffer( ptf_dem_handle dem, void * buffer, ptf_16 startCol, ptf_16 startRow, 
                                          ptf_16 width, ptf_16 height, ptf_16 buf_stride );

//! @}
#endif  //  ptf_raster_h
