#if !defined(ptf_file_h)
//! \internal header guard
#define ptf_file_h

#if !defined(ptf_h)
#include "ptf.h"
#endif
#if !defined(ptf_err_h)
#include "ptf_err.h"
#endif
#if !defined(ptf_types_h)
#include "ptf_types.h"
#endif

//! \addtogroup PtfLibrary The PTF File Library
//! @{

#if defined(__cplusplus)
extern "C" {
#endif

//! A ptf_file_handle represents an open PTF file. It is needed 
//! for all other operations on PTF files.
typedef struct ptf_file_t * ptf_file_handle;

//! There are several modes of open for a PTF file.
enum ptf_open_mode {
  ptf_open_clobber,         //!< Always create a new writable file, overwriting any old file.
  ptf_open_create,          //!< Create a new writable file if one doesn't already exist, else open writable.
  ptf_open_update,          //!< Open an existing file for writing; if not existing, fail.
  ptf_open_read,            //!< Open an existing file for reading, if it exists.
};

//! I/O through the PTF library can be re-directed to your own functions.
//! By default, PTF will use fopen() and friends to do I/O.
//! A file accessor handle is passed to the file accessors, letting you 
//! indicate what underlying object to do I/O on. This could be a pointer 
//! to an object, or a HFILE, or a FILE*, or something else that's up to you.
typedef void * ptf_file_accessor_handle;
//! The ptf_file_accessors structure can be passed to ptf_file_open_accessors() 
//! to abstract out I/O to the PTF file object. The structure is copied into the 
//! PTF file object, so need not be valid after ptf_file_open_accessors() returns.
struct ptf_file_accessors {
  //! When read() is called, you should transfer exactly the amount of data 
  //! indicated, into the buffer indicated. If you cannot transfer that amount,
  //! return an error. If transferred, return peSuccess. Partial reads are not 
  //! acceptable.
  ptf_err (*read)(void * buffer, ptf_64 len, ptf_file_accessor_handle);
  //! When seek() is called, seek to the indicated offset in the file. The 
  //! "whence" value is 0 for offset from beginning of file, 1 for offset 
  //! from current location (not currently used), and 2 for offset from 
  //! end of file. Return peSuccess on success, else an error.
  ptf_err (*seek)(ptf_file_accessor_handle, ptf_64 offset, int whence);
  //! When write() is called, transfer exactly the amount of requested 
  //! bytes into the file (or into a buffer of the file). If you cannot 
  //! fit that many bytes, return an error -- partial writes are not 
  //! acceptable.
  ptf_err (*write)(void const * buffer, ptf_64 len, ptf_file_accessor_handle);
  //! Return the current offset in the file, measured from the start of 
  //! the file, updated through any seek/read/write operations.
  ptf_64 (*tell)(ptf_file_accessor_handle);
  //! Flush all buffering to disk. This is the mechanism by which transactional 
  //! commits to PTF files are implemented. All dirty data will be written and 
  //! flushed, then the header will be updated, written, and flushed.
  ptf_err (*sync)(ptf_file_accessor_handle);
  //! Open a handle, given the path and open mode passed into ptf_file_open_accessors.
  //! The param is passed along from open_accessors. Return NULL on failure, and 
  //! non-NULL on success.
  ptf_file_accessor_handle (*open)( ptf_char const * path, ptf_open_mode, void * param);
  //! Close a handle that you previously returned from open(). There is an implicit 
  //! flush of any outstanding dirty data as part of a close().
  ptf_err (*close)(ptf_file_accessor_handle);
};
//! Return the accessors for fopen()-style I/O that will be used when no other 
//! accessors are passed into ptf_file_open().
PTF_API ptf_file_accessors const * ptf_file_platform_accessors();

//! Open an existing PTF file on disk. The returned handle must be 
//! later disposed with ptf_file_close().
//! \param path The path to the file to open.
//! \param mode The mode to open the file in (writable, clobber, etc).
//! \o_file Set to NULL on failure, or non-NULL on success.
//! \return ptf_err status.
//! \note If both create and open are set, the file will be opened if 
//! existing, else it will be created. If neither is set, this function 
//! returns NULL.
PTF_API ptf_err ptf_file_open(ptf_char const * path, ptf_open_mode mode, ptf_file_handle * o_file);

//! Open an existing PTF file on disk. The returned handle must be 
//! later disposed with ptf_file_close().
//! \param path The path to the file to open.
//! \param mode The mode to open the file in (writable, clobber, etc).
//! \param accessors The platform-specific accessors to use for this file.
//! \param param User parameter to pass into the accessor functions.
//! \o_file Set to NULL on failure, or non-NULL on success.
//! \return ptf_err status.
//! \note If both create and open are set, the file will be opened if 
//! existing, else it will be created. If neither is set, this function 
//! returns NULL.
PTF_API ptf_err ptf_file_open_accessors(ptf_char const * path, ptf_open_mode mode, ptf_file_accessors const * accessors, void * param, ptf_file_handle * o_file);

//! Close a previously open PTF file. It is legal to call ptf_file_close() 
//! on NULL (which does nothing).
//! \param file The PTF file handle, from a previous call to ptf_file_open().
//! \return ptf_err status.
PTF_API ptf_err ptf_file_close(ptf_file_handle file);

//! Flush all pending changes, update the file header, and sync the underlying 
//! file to disk. If no change are pending, this function is a no-op.
//! \param file An open PTF file with pending changes.
//! \return ptf_err status.
//!
//! \note It is possible for an implementation of this interface to produce 
//! "safe" transactional semantics through the interface, by re-writing any 
//! changed index in a new place, syncing the file, updating the file header and 
//! syncing the file again. The reference attempts this but hasn't been extensively 
//! stress tested.
//!
//! The API does not address attempts to use the same file from more than one 
//! process, or attempts to open the same file more than once to generate more 
//! than one ptf_file_handle.
PTF_API ptf_err ptf_file_flush(ptf_file_handle file);

//! Count the number of types within the meta TOC of the file.
//! \param file The file to count types within.
//! \param o_count The number of types within the file.
//! \return ptf_err status.
PTF_API ptf_err ptf_file_count_types(ptf_file_handle file, ptf_32 * o_count);

//! Enumerate the available types within a file.
//! \param file The file to enumerate.
//! \param index The N-th type to return (must be between 0 and the value 
//! returned by ptf_file_count_types() minus one).
//! \param o_type A given type, or ptNoType for failure.
//! \return ptf_err status.
PTF_API ptf_err ptf_file_get_type_at(ptf_file_handle file, ptf_32 index, ptf_type * o_type);

//! Count the number of records within a given type for a given file.
//! Note that some of these records may have been removed, or be empty, but still 
//! count for the sake of indexing.
//! Also note that by convention, record 0 is always empty and cannot be written to.
//! This record (0) is still included in the count of TOC entries returned.
//! \param file The file to enumerate.
//! \param type The type for which to count records.
//! \param o_count The number of records for the given type, or 0 for failure.
//! \return ptf_err status.
PTF_API ptf_err ptf_file_count_type_records(ptf_file_handle file, ptf_type type, ptf_32 * o_count);

//! Retrieve the size of a given record of a given type.
//! \param file The file to examine.
//! \param ref A reference to the record to retrieve.
//! \param o_size The size of the given record.
//! \return ptf_err status. If the record has been removed, peRecordNotFound is returned.
//! If the index of the record is too large, peRecordNotFound is also returned. You can 
//! tell these two cases apart by checking ptf_file_count_type_records() for the type.
PTF_API ptf_err ptf_file_get_record_size(ptf_file_handle file, ptf_record_ref const * ref, ptf_64 * o_size);

//! Read a (potential part of) record, given the identifying type and index.
//! \param file The file to read from.
//! \param ref A reference to the record to read.
//! \param offset Start this many bytes into the record.
//! \param buf Put data in this buffer.
//! \param size Read this much data.
//! \return ptf_err status.
//! \note It is an error to attempt to read past the end of a record. 
//! A read that extends at least partially past the end of a record will 
//! return an error. You always get all of what you ask for, or nothing.
PTF_API ptf_err ptf_file_read_record(ptf_file_handle file, ptf_record_ref const * ref, ptf_64 offset, void * buf, ptf_64 size);

//! Allocate a new record within the file.
//! You can then start writing data to the record with ptf_file_write_record(), 
//! and finally commit the changes with ptf_file_commit_record(). It is illegal 
//! to keep more than one record open (uncommitted) at a time for a given file.
//! \param file The file to append the record to.
//! \param type The type of the record to allocate.
//! \return ptf_err status.
//! \note The "file" part of the ptf_record_ref will always be 0, because the 
//! record is always appended to the indicated file, which is always 0 in its own 
//! local namespace.
PTF_API ptf_err ptf_file_new_record(ptf_file_handle file, ptf_type type);

//! Append data to the end of the currently open record (if any).
//! If the function fails to write the data, the entire record transaction 
//! is aborted, and another call to ptf_file_write_record() or ptf_file_commit_record() 
//! will fail, until a new call to ptf_file_new_record() has been successfully made.
//! \param file A file with a currently open record to append to.
PTF_API ptf_err ptf_file_write_record(ptf_file_handle file, void const * data, ptf_32 size);

//! Commit data appended to a record into the file. As of this call, the 
//! data will be visible to any future library call using the same file handle.
//! However, changes may not (but might) be visible to other potential users 
//! of the file, either in a different process, or using a different ptf_file_handle. 
//! \param file The file with outstanding changes to commit.
//! \param o_ref The record reference of the committed record (can be NULL).
//! \return ptf_err status.
//! \note If you are committing a record replacement, then the returned reference 
//! will be the same as you passed into ptf_file_replace_record(). If you are committing 
//! a new record, you should not pass NULL for o_ref, or you will lose the record index 
//! of the new record. Record index 0 is never a valid record in PTF, and won't ever be 
//! returned.
PTF_API ptf_err ptf_file_commit_record(ptf_file_handle file, ptf_record_ref * o_ref);

//! If you have all the data you want to write already in memory, you can do the 
//! work of ptf_file_new_record(), ptf_file_write_record() and ptf_file_commit_record()
//! all in one call. In addition, this may allow the implementation to make better 
//! determinations about where to put the data in the file.
//! \param file The file to create the new record in.
//! \param type The type of the record to create.
//! \param data A pointer to the user data to put in the record.
//! \param size The size of the data to put in the record.
//! \param o_ref The actual record ref will be returned here.
//! \return ptf_err status.
//! \see Notes about ptf_file_commit_record().
PTF_API ptf_err ptf_file_new_complete_record(ptf_file_handle file, ptf_type type, void const * data, ptf_32 size, ptf_record_ref * o_ref);

//! Remove a given record from the file. Removed records are moved into the "unused 
//! record" table, and may be re-used sometime later (or not, depending on implementation).
//! The index of the record will still be reserved; another call to ptf_file_new_record() 
//! will not re-use the index.
//! \param file The file to remove the record from.
//! \param ref The record to remove. The 'file' parameter must be 0.
//! \return ptf_err status.
PTF_API ptf_err ptf_file_remove_record(ptf_file_handle file, ptf_record_ref const * ref);

//! Allocate one or more records in the "removed" state, and return the record_ref of
//! the first such record. Sequential records after that first record (by index) will 
//! also be allocated as "removed" records, up to the count indicated.
//! \param file The file to allocate record indices in.
//! \param type The type of record index to allocate.
//! \param count The number of record indices to allocate.
//! \param o_ref The record_ref for the first record index allocated. Note that this 
//! record will be in a "removed" state. You can later re-write this record with 
//! content using ptf_file_replace_record().
//! \return ptf_err status.
PTF_API ptf_err ptf_file_allocate_records(ptf_file_handle file, ptf_type type, ptf_32 count, ptf_record_ref *o_ref);

//! Re-allocate space for an existing record within the file.
//! You can then start writing data to the record with ptf_file_write_record(), 
//! and finally commit the changes with ptf_file_commit_record(). It is illegal 
//! to keep more than one record open (uncommitted) at a time for a given file.
//! ptf_file_replace_record() works even on removed records, and will then make 
//! the record of the given index not removed anymore.
//! \param file The file to rewrite the record data in.
//! \param ref The existing record that will have its data replaced.
//! \return ptf_err status.
//! \note The "file" part of the ptf_record_ref will always be 0, because the 
//! record is always appended to the indicated file, which is always 0 in its own 
//! local namespace.
PTF_API ptf_err ptf_file_replace_record(ptf_file_handle file, ptf_record_ref const * ref);


//! Add a reference to an external file. The key look-up is based on a case 
//! insensitive comparison of the canonical (header) name of the file. If the 
//! file is already being referenced, return the index of that reference.
//! \param file The open file to add the reference to.
//! \param def The definition to make, or look up.
//! \param o_index The index of the file reference record will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_file_make_reference(ptf_file_handle file, ptf_file_definition_record const * def, ptf_32 * o_index);

//! Given a file reference index, get the file definition record for that index.
//! There is no guarantee that the file actually exists.
//! \param file The open file to look up the reference in.
//! \param index The file reference index, typically from a ptf_record_ref.
//! \param o_def The definition, as stored in the file.
//! \return ptf_err status.
PTF_API ptf_err ptf_file_get_reference(ptf_file_handle file, ptf_32 index, ptf_file_definition_record * o_def);

//! Given a file definition record, find the full path of the file that is 
//! being referenced. The searching algorithm is implementation defined, although 
//! the current working directory of the referencing file is recommended as one 
//! search location. 
//! Update the ptf_file_definition_record 'fixed_path' member with the path at 
//! the file was found, if found.
//! \param def The definition to resolve to a path. The path is returned in 
//! this parameter as well. Only the "name" parameter is significant on input.
//! \param path NULL, or a NULL-terminated array of search paths to try when 
//! looking for the file.  Note: the search paths should not have a trailing '\'
//! \return ptf_err status.
//! \note The file found is verified as being a PTF file with the given 
//! canonical name, but the file version or consistency is not further checked.
PTF_API ptf_err ptf_reference_resolve_file(ptf_file_definition_record * def, ptf_char const * const * path );

//! Given a file definition record, find the full path of the file that is 
//! being referenced. The searching algorithm is not defined at this time. 
//! Update the ptf_file_definition_record 'path' member with the path at 
//! the file was found, if found. That pointer will be valid at least until 
//! the next call into the PTF library.
//! \param def The definition to resolve to a path. The path is returned in 
//! this parameter as well.
//! \param path NULL, or a NULL-terminate array of search paths to try when 
//! looking for the file.
//! \param accessors The platform-specific accessors to use for this file.
//! \param param User parameter to pass into the accessor functions.
//! \return ptf_err status.
//! \note The file found is verified as being a PTF file with the given 
//! canonical name, but the file version or consistency is not further checked.
PTF_API ptf_err ptf_reference_resolve_file_accessors(ptf_file_definition_record * def, ptf_char const * const * path, ptf_file_accessors const * accessors, void* param );

//! Given a PTF file, return the current header in effect for that file.
//! \param file The open file to get the header for.
//! \param o_hdr The header will be returned here.
//! \return ptf_err status.
//! \note Not all fields (having to do with meta-toc and size) are guaranteed to 
//! be the same as the file on disk, unless the file has just been flushed.
PTF_API ptf_err ptf_file_get_header(ptf_file_handle file, ptf_file_header * o_hdr);

//! Given a PTF file, update some fields of the header. The fields being 
//! updated include the canonical name, the creation and modification date and 
//! the extents. The other fields are not updated.
//! \param file The open file to update information about.
//! \param i_hdr The header information to update.
//! \return ptf_err status.
//! \note Typically, you'll call ptf_file_get_header(), then modify the fields 
//! you want to modify, and call ptf_file_update_header. Not all fields are 
//! writable in the header.
PTF_API ptf_err ptf_file_update_header(ptf_file_handle file, ptf_file_header const * i_hdr);

//! Retrieve the path used to open or create the PTF file.
//! \param file The file to get the open path from.
//! \param o_buf The open path will be stored in this buffer.
//! \param buf_size The number of bytes of storage space in the output buffer.
//! \return ptf_err status.
PTF_API ptf_err ptf_file_get_open_path(ptf_file_handle file, ptf_char *o_buf, size_t buf_size);

//! Retrieve a ptf_file_definition_record for a given, open, ptf_file_handle.
//! This record is suitable for looking up or creating a reference to the file using 
//! ptf_file_make_reference() (called on another ptf_file_handle, as files don't 
//! reference themselves!)
//! \param file The file handle to extract a definition record for.
//! \param io_def The definition to make, or return the already made value in.
//! \return ptf_err status.
PTF_API ptf_err ptf_file_get_definition_record(ptf_file_handle file, ptf_file_definition_record *io_def);

#if defined(__cplusplus)
}
#endif

//! @}

#endif    //  ptf_file_h
