
#if !defined(ptf_lod_h)
//! \internal header guard
#define ptf_lod_h

#if !defined(ptf_file_h)
#include "ptf_file.h"
#endif
#if !defined(ptf_types_h)
#include "ptf_types.h"
#endif
//! \addtogroup PtfGeometry The PTF Geometry Library

//! All LOD group instances are identified with a ptf_lod_handle. 
//! There are functions to read a LOD handle from a PTF file, and 
//! write a LOD handle to a PTF file. The specific data layout is 
//! documented in <ptf_types.h>.
//!
//! A LOD group contains some number of mesh references, sorted 
//! as coarsest to finest resolution. Any one of these meshes 
//! represent the same "thing," so only one of them should be drawn 
//! at any one time.
//!
//! A LOD group may also contain additional information, such as 
//! pointers to collision data.
typedef class ptf_lod_t *ptf_lod_handle;

//! Create a LOD instance. You need an instance to be able to read data 
//! into it, or add data and write out to a file.
//! \param o_lod The LOD instance handle will be stored here. You must 
//! call ptf_lod_destroy() on it once you're done with it.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_create(ptf_lod_handle *o_lod, ptf_char const * name);

//! Deallocate all data allocated by a LOD instance.
//! \param lod The allocated instance to destroy. It will not be valid 
//! after this call returns.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_destroy(ptf_lod_handle lod);

//! Read a ptLodData block from a PTF file and populate the LOD instance 
//! handle with the data found therein.
//! \param lod The LOD handle to populate with data from the file.
//! \param file The PTF file to read from.
//! \param ref The ptLodData type record index to read from.
//! \return ptf_err status.
//! \see ptf_lod_header
PTF_API ptf_err ptf_lod_read_from_file(ptf_lod_handle lod, 
    ptf_file_handle file, ptf_fixed_ref const &ref);

//! Given a populated LOD handle, and an existing LOD data record in a 
//! PTF file, replace that record with the data of the populated LOD.
//! The LOD in memory is left intact; this is a pure "write" operation.
//! \param lod The LOD to write to file.
//! \param file The file to receive the LOD data.
//! \param ref The index of the ptLodData record that will receive the data.
//! \return ptf_err status.
//! \see ptf_lod_header
PTF_API ptf_err ptf_lod_update_to_file(ptf_lod_handle lod, 
    ptf_file_handle file, ptf_fixed_ref const &ref);

//! Given a populated LOD handle, append a new ptLodData record to the 
//! given PTF file, and return the index of this new record.
//! \param lod The lod handle to write to the file.
//! \param file The file to receive the lod handle.
//! \param o_ref The index of the lod record will be returned here.
//! \return ptf_err status.
//! \see ptf_lod_header
PTF_API ptf_err ptf_lod_append_to_file(ptf_lod_handle lod, 
    ptf_file_handle file, ptf_fixed_ref *o_ref);

//! Given a ptf_lod entry, set the LOD table for that entry.
//! This is a type-safe wrapper for ptf_lod_set_semantic_items().
//! \param lod The allocated LOD handle.
//! \param table The LOD entries, sorted from coarsest (largest feature size) to finest.
//! \param count The number of entries in the LOD table.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_set_table(ptf_lod_handle lod,
    ptf_lod_table_entry const *table, ptf_16 count);

//! Given a ptf_lod entry, get the LOD table from that entry.
//! This is a type-safe wrapper for ptf_lod_get_data() using plsLodTable.
//! \param lod The allocated LOD handle.
//! \param table The LOD entries will be returned here.
//! \param tablecount The number of entries in the LOD table. If this is 
//! lower than the number of actual entries, only that many entries will be returned.
//! This value can be 0 to check how many entries there actually are.
//! \param o_count The actual number of entires is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_get_table(ptf_lod_handle lod,
    ptf_lod_table_entry *o_table, ptf_16 tablecount, ptf_16 *o_count);

//! Count the number of data instances within the LOD chunk. Arbitrary 
//! data types with specified semantics can be added to the chunk.
//! For defined semantics, see ptf_lod_semantic.
//! \param lod The allocated LOD handle to count semantics in.
//! \param o_count The number of semantics in the handle will be returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_count_semantics(ptf_lod_handle lod,
    ptf_32 *o_count);

//! Given an allocated LOD handle, get information about the N:th 
//! semantic, by index.
//! \param lod The allocated LOD handle.
//! \param index The index of the semantic to get information about.
//! This should be between 0 (inclusive) and ptf_lod_count_semantics()
//! (exclusive).
//! \param o_semantic The semantic identifier will be returned here. 
//! \see ptf_lod_semantic
//! \param o_count The number of entries of this semantic will be returned here.
//! \param o_itemsize The size of each item will be returned here, if 
//! fixed size. If variable size, this will contain 0.
//! \param o_datasize The total size of the data items, including padding, 
//! will be returned here. Because of padding, this may be larger than 
//! o_count * o_itemsize.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_get_semantic_at(ptf_lod_handle lod,
    ptf_32 index, ptf_32 *o_semantic, ptf_16 *o_count, ptf_16 *o_itemsize, 
    ptf_32 *o_datasize);

//! Given an allocated LOD handle, get information about a specific semantic, 
//! if present. This does a linear scan through the semantic table.
//! \param lod The allocated LOD handle.
//! \param semantic The semantic to get information about, if present.
//! \param i_index The index of the semantic within the semantic table in 
//! the LOD handle will be returned here.
//! \param o_count The number of entries of this semantic will be returned here.
//! \param o_itemsize The size of each item will be returned here, if 
//! fixed size. If variable size, this will contain 0.
//! \param o_datasize The total size of the data items, including padding, 
//! will be returned here. Because of padding, this may be larger than 
//! o_count * o_itemsize.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_get_semantic(ptf_lod_handle lod,
    ptf_32 semantic, ptf_32 *o_index, ptf_16 *o_count, ptf_16 *o_itemsize, 
    ptf_32 *o_datasize);

//! Given an allocated LOD handle, and a semantic item within the valid range 
//! for the handle, return all the items as a single block of data.
//! \param lod The allocated LOD handle to return data from.
//! \param index The index of the semantic to return data for.
//! \see ptf_lod_get_semantic() to get the index for a given semantic.
//! \param o_data The returned data will be stored in this buffer.
//! \param datasize At most this many bytes will be returned into the buffer.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_get_data(ptf_lod_handle lod,
    ptf_32 index, void *o_data, ptf_32 datasize);

//! Update all the semantic data for a given semantic for a given handle.
//! This replaces any previous data for that semantic.
//! \param lod The allocated handle to update data for.
//! \param semantic The actual semantic to update/replace/add.
//! \param data_size The amount of data to store. Does not need to be 
//! padded by the user, but will be padded to 8 by the storage.
//! \param src The data to store.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_set_semantic_blob(ptf_lod_handle lod,
    ptf_32 semantic, ptf_32 data_size, void const *src);

//! Update all the semantic data for a given semantic for a given handle.
//! This replaces any previous data for that semantic.
//! \param lod The allocated handle to update data for.
//! \param semantic The actual semantic to update/replace/add.
//! \param item_size The size of each item in the array.
//! \param num_items The number of items in the array.
//! \param src The data to store, which must be an array of data 
//! elements, each of which has size item_size.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_set_semantic_items(ptf_lod_handle lod,
    ptf_32 semantic, ptf_16 item_size, ptf_16 num_items, void const *src);

//! Set the given LOD to use the given collision data reference.
//! \param mesh The LOD to associate the collision data with.
//! \param coll Reference to the collision data record to use.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_set_collision(ptf_lod_handle lod, ptf_record_ref coll);

//! Return the given collision data used for the given LOD.
//! \param mesh The LOD the collision data is associated with.
//! \param o_coll The reference to the collision data record is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_lod_get_collision(ptf_lod_handle lod, ptf_record_ref * o_coll);

#endif  //  ptf_mesh_h
