
#if !defined(ptf_spheroid_h)
//! \internal header guard
#define ptf_spheroid_h

#include "ptf_file.h"
#include "ptf_types.h"
//! \addtogroup PtfGeometry The PTF Geometry Library

//! All spheroid data instances is identified by an abstract ptf_spheroid_handle.
//! These instances can exist separate from any ptf_file instance, and it
//! is up to the programmer to manage the lifetime of these objects. The 
//! API does provide functions to read and write these instances from/to 
//! PTF files.
typedef class ptf_spheroid_t *ptf_spheroid_handle;

//! Header for spheroid primitive data, as stored on disk.
struct ptf_spheroid_data {
  ptf_name      name;           //!< dataset identifier (48 bytes)
  ptf_fixed_ref material_arc;   //!< spheroid material archive
  ptf_16        material_id;    //!< spheroid material_id
  ptf_flag      surface;        //!< flags that spheroid represents the Earth's surface (should only be true for one shperoid per file)
  ptf_8         reserved1;      //!< reserved
  ptf_32        reserved2;      //!< reserved
  ptf_32        reserved3;      //!< reserved
  double        radius;         //!< spheroid major radius
  double        squash;         //!< spheroid squash is a scaling factor that is used to calculate 
                                //!< minor radius of the spheroid given the major radius
                                //!< formula for calculating the squash factor is 
                                //!< (major radius)/(major radius-minor radius)
};

//! Create a spheroid instance. You need an instance to be able to read data 
//! into it, or add data and write out to a file.
//! \param o_spheroid The spheroid instance handle will be stored here. You must 
//! call ptf_spheroid_destroy() on it once you're done with it.
//! \return ptf_err status.
PTF_API ptf_err ptf_spheroid_create(ptf_spheroid_handle *o_spheroid, ptf_char const * name);

//! Deallocate all data allocated by a spheroid instance.
//! \param spheroid The allocated instance to destroy. It will not be valid 
//! after this call returns.
//! \return ptf_err status.
PTF_API ptf_err ptf_spheroid_destroy(ptf_spheroid_handle spheroid);

//! Read a ptSpheroidData block from a PTF file and populate the spheroid instance 
//! handle with the data found therein.
//! \param spheroid The spheroid handle to populate with data from the file.
//! \param file The PTF file to read from.
//! \param ref The ptSpheroidData type record index to read from.
//! \return ptf_err status.
//! \see ptf_spheroid_data
PTF_API ptf_err ptf_spheroid_read_from_file(ptf_spheroid_handle spheroid, 
    ptf_file_handle file, ptf_fixed_ref const &ref);

//! Given a populated spheroid handle, and an existing spheroid data record in a 
//! PTF file, replace that record with the data of the populated spheroid.
//! The spheroid in memory is left intact; this is a pure "write" operation.
//! This will return an error if the spheroid isn't minimally consistent.
//! \param spheroid The spheroid to write to file.
//! \param file The file to receive the spheroid data.
//! \param ref The index of the ptSpheroidData record that will receive the data.
//! \return ptf_err status.
//! \see ptf_spheroid_data
PTF_API ptf_err ptf_spheroid_update_to_file(ptf_spheroid_handle spheroid, 
    ptf_file_handle file, ptf_fixed_ref const &ref);

//! Given a populated spheroid handle, append a new ptSpheroidData record to the 
//! given PTF file, and return the index of this new record.
//! This will return an error if the spheroid isn't minimally consistent.
//! \param spheroid The spheroid handle to write to the file.
//! \param file The file to receive the spheroid handle.
//! \param o_ref The index of the spheroid record will be returned here.
//! \return ptf_err status.
//! \see ptf_spheroid_data
PTF_API ptf_err ptf_spheroid_append_to_file(ptf_spheroid_handle spheroid, 
    ptf_file_handle file, ptf_fixed_ref *o_ref);

//! Store spheroid data
//! \param spheroid The spheroid object to copy data into.
//! \param data The spheroid data 
//! \return ptf_err status.
//! \see ptf_spheroid_data
PTF_API ptf_err ptf_spheroid_set_data(ptf_spheroid_handle spheroid, 
    ptf_spheroid_data const &data);

//! Retrieve data stored for a given spheroid object
//! \param spheroid The spheroid object to get data from.
//! \param o_data A pointer to the stored spheroid data will be returned here.
//! \return ptf_err status.
//! \see ptf_spheroid_data
PTF_API ptf_err ptf_spheroid_get_data(ptf_spheroid_handle spheroid, 
    ptf_spheroid_data * o_data);

#endif  //  ptf_spheroid_h
