#if !defined(ptf_str_h)
#define ptf_str_h

#if !defined(ptf_h)
#include "ptf.h"
#endif

#include <wchar.h>
//! \addtogroup PtfLibrary The PTF File Library
//! @{

//! Given a PTF UTF-8 string, convert it to an ISO 8859-1 string, substituting 
//! a given stand-in for non-ASCII characters.
//! \param src The input PTF string.
//! \param dst The output C string storage. Can be NULL to calculate how much 
//! storage would be needed.
//! \param count_dst The number of characters available at storage (including 
//! the terminating 0).
//! \param stand_in The character to substitute for for out-of-range characters.
//! \return The actual number of characters used in the output (or needed by the 
//! output), including terminating 0, whichever is greater.
//! \note This function will always zero-terminate the result, even when not 
//! being able to translate the entire string because of space.
PTF_API size_t ptf_str_to_char(ptf_char const * src, char * dst, size_t count_dst, char stand_in);

//! Given a PTF UTF-8 string, convert it to wide-char string, substituting 
//! a given stand-in for characters greater than 0xffff.
//! \param src The input PTF string.
//! \param dst The output wide string storage. Can be NULL to calculate how much 
//! storage would be needed.
//! \param count_dst The number of wide characters available at storage (including 
//! the terminating 0).
//! \param stand_in The character to substitute for for out-of-range characters.
//! \return The actual number of characters used in the output (or needed by the 
//! output), including terminating 0, whichever is greater.
//! \note This function will always zero-terminate the result, even when not 
//! being able to translate the entire string because of space.
PTF_API size_t ptf_str_to_wchar(ptf_char const * src, wchar_t * dst, size_t count_dst, wchar_t stand_in);

//! Given a regular ISO 8859-1 string, convert it to PTF.
//! \param str A 0-terminated ISO 8859-1 string.
//! \param dst A buffer to put the PTF string into.
//! \param count_dst The number of ptf_char elements in the buffer to use, including 
//! the terminating 0.
//! \return The actual number of characters used in the output (or needed by the 
//! output), including terminating 0, whichever is greater.
//! If the input data is not proper UTF-8, this function returns 0, and may 
//! overwrite some part of the buffer.
//! \note This function will always zero-terminate the result, even when not 
//! being able to translate the entire string because of space.
PTF_API size_t ptf_str_from_char(char const * str, ptf_char * dst, size_t count_dst);

//! Given a wide-char string, convert it to PTF.
//! \param str A 0-terminated wide character string.
//! \param dst A buffer to put the PTF string into.
//! \param count_dst The number of ptf_char elements in the buffer to use, including 
//! the terminating 0.
//! \return The actual number of characters used in the output (or needed by the 
//! output), including terminating 0, whichever is greater.
//! If the input data is not proper UTF-8, this function returns 0, and may 
//! overwrite some part of the buffer.
//! \note This function will always zero-terminate the result, even when not 
//! being able to translate the entire string because of space.
PTF_API size_t ptf_str_from_wchar(wchar_t const * str, ptf_char * dst, size_t count_dst);

//! Given a path name in ptf_char format, return the basename (minus any leading 
//! file path component) in ptf_char format.
//! \param path The initial path.
//! \param dst The destination for the file name part.
//! \param size The size of the destination buffer.
//! \return The number of characters really needed for the basename.
//! \note The output dst will always be zero terminated.
PTF_API size_t ptf_str_basename(ptf_char const * path, ptf_char * dst, size_t size);

//! Given two ptf_char strings, do a case-insensitive compare, returning 
//! -1, 0 or 1 based on lexicographic order.
//! \param a The first string.
//! \param b The second string.
//! \return -1 if a < b, 0 if a == b, 1 if a > b
//! \note The case insensitivity and collation order is really only correct 
//! for the C locale.
PTF_API int ptf_strcasecmp(ptf_char const * a, ptf_char const * b);

//! Copy a ptf_char string into the space given, making sure to always 
//! zero-terminate the output and never overwrite the buffer. If there is 
//! not enough space, only whole glyphs will be transferred, and an error 
//! (-1) will be returned. The output buffer is still zero terminated.
//! \param dst The destination buffer.
//! \param size The size (in bytes) of the destination buffer, including terminating zero.
//! \param src The source string.
//! \return -1 on truncation, 0 on success
PTF_API int ptf_strncpy(ptf_char * dst, size_t size, ptf_char const * src);

//! @}

#endif  //  ptf_str_h
