#if !defined(ptf_types_h)
//! \internal header guard
#define ptf_types_h

#if !defined(ptf_h)
#include "ptf.h"
#endif

//! \addtogroup PtfFormat The PTF File Format

//! PTF types define the majority of the structure and functionality of a PTF 
//! file. Record instances of types are stored in the file, and referenced through 
//! the TOC for that type. Additionally, the formats of some types contain 
//! references to other records of the same or other types.
//!
//! PTF types are divided into four ranges:
//! <ul><li>0-0x3fffffff: types used by the published specification. Arbitrated 
//!   through the standard document.</li>
//! <li>0x40000000-0x7fffffff: types used by at least one vendor in "production" use. 
//!   Ideally, arbitrated through a registry.</li>
//! <li>0x80000000-0xbfffffff: reserved; do not use.</li>
//! <li>0xc0000000-0xffffffff: experimental, non-production use. Not arbitrated.</li>
//! </ul>
//! The intention is for an experimental type to be possible to "promote" to a 
//! production type by masking off the high bit. Thus, if a vendor has a specific 
//! range of types allocated, the corresponding high-bit-set range should be the 
//! first choice for experiments by that vendor.
enum {
  //! ptNoType is not a valid type in and of itself. It may be 
  //! used as a signifier of no information.
  ptNoType = 0,
  //! The type of a annotation record. There may be multiple annotation 
  //! records within a single PTF record chunk (i e, it's an array of 
  //! such records, each of which is fixed size).
  //! \see struct ptf_annotation_entry
  ptAnnotation = 1,
  //! The type of an "empty" (unused) record. The rules for alignment 
  //! for these records is the same as for any other record, so the padding 
  //! between the end of a record and the end of the last page that the 
  //! record occupies is not considered an empty record.
  //! Note that you can't create or update records of this type through 
  //! the API; they are internally managed by the library. You can, however, 
  //! iterate over them, so you can calculate statistics.
  //! \see struct ptf_toc_entry
  ptEmptyRecord = 2,
  //! The type for a name for an external file, referenced within this file.
  //! External files are referenced by indexes into the external file records.
  //! Index 0 is reserved for the current file, and thus will be empty in the 
  //! TOC.
  ptFileDefinition = 3,
  //! The type for a block of Octree nodes. The first such block (at index 
  //! 1) is the root octree for a file aligned on octree boundaries. 
  //! The eight first such blocks, skipping node 1 (index 2 through 9) may be the 
  //! virtual root of a file, if it is aligned on a half octree block.
  ptOctreeNode = 4,
  //! The type for a "content description" record. These are pointed at by 
  //! content link indices in the Octree nodes.
  ptOctreeContent = 5,

  //! The type for a "bookmark" record.  Bookmarks are used to indicate geo-locations
  //! of interest within the file.
  ptBookmark = 6,

  //! Mesh data is stored in the PTF flattened mesh format.
  ptMeshData = 0x10,

  //! Raster elevation data is stored in the PTF orthogonally sampled format.
  ptElevationData = 0x11,

  //! Shape data defines implicit surfaces.
  ptShapeData = 0x12,

  //! Texture data can be stored within the PTF file for those cases where 
  //! you want to stream all data from a single file. You could also reference 
  //! external textures using ptTextureReference.
  //! \see ptf_texture_data
  ptTextureData = 0x13,

  //! Spheroid geometry
  ptSpheroidData = 0x14,

  //! Any kind of mesh data can be referenced using a mesh reference.
  //! The type of the data is inferred from the file name. Note that a 
  //! given PTF reader is not required to understand geometry in formats 
  //! other than PTM (which is a shallow wrapper on the ptMeshData record 
  //! format) and an external file PTF record reference.
  //! \see ptf_external_reference
  //! \see ptf_disk_file_header
  ptMeshReference = 0x20,

  //! Any kind of raster data can be referenced using a raster reference. 
  //! The specific correlation of the raster data to the world is up to 
  //! the raster data itself. The type of the raster data is inferred from 
  //! the file name. Note that a given PTF reader is not required to 
  //! understand raster in formats other than PTR (which is a shallow 
  //! wrapper on the ptRasterData record format) and an external file PTF 
  //! record reference.
  //! \see ptf_external_reference
  //! \see ptf_disk_file_header
  ptElevationReference = 0x21,

  //! Any kind of shape data can be referenced using a shape reference. 
  //! this can range from RenderMan scene descriptions to a simple ovoid 
  //! description (such as used to create the initial earth surface).
  //! Note that a given PTF reader is not required to understand shape 
  //! data in a format other than PTS (which is a thin wrapper on top of 
  //! the ptShapeData record format) and an external file PTF record 
  //! reference.
  //! \see ptf_external_reference
  //! \see ptf_disk_file_header
  ptShapeReference = 0x22,

  //! Any kind of texture file can be referenced by a texture reference. 
  //! The type of texture file is inferred from the name. However, the 
  //! only type of external texture data that is guaranteed to be parsed 
  //! by a PTF reader is an external PTF record reference, or the PTT file 
  //! format (which is a shallow wrapper around the PTF record format).
  //! \see ptf_disk_file_header
  ptTextureReference = 0x23,

  //! Material archives contain the materials used by geometry.
  ptMaterialArchive = 0x101,
  
  //! LOD data contains collections of mesh references for meshes that 
  //! each describe the same entity in different degrees of detail.
  ptLodData = 0x141,

  //! Opcode Collision data is stored in the PTF in raw byte format.
  //! It is the pre-processed AABB tree data from the OPCODE library, 
  //! as also used by the ODE library. This data is not necessarily 
  //! normative, and could be re-created using a scene-to-collision 
  //! compilation tool.
  ptOpcodeCollisionData = 0x142,
  
  //! Mesh or other geometry instance meta information contains data that 
  //! identifies pieces of a mesh into a whole, and also classification 
  //! data for that mesh instance.
  ptGeometryInstanceMetaData = 0x143,

  //! A layer description contains information about a collection of 
  //! mesh instances. It is referenced by the ptGeometryInstanceMetaData.
  ptLayerData = 0x144,

  //! Collection data contains collections of mesh references for meshes that 
  //! describe the connections among different octree nodes of the same "meta-mesh".
  ptCollectionData = 0x181,

  //! \{
  //! Sometimes it's useful to type pieces of data that are well known. For example,
  //! the annotation on references in the content type. For those purposes, the 
  //! following types are defined.
  
  //! A record_ref to a block
  ptRecordRef = 0x1000,
  //! A local_ref to a block
  ptLocalRef = 0x1001,
  //! A fixed_ref to a block
  ptFixedRef = 0x1002,
  //! A position (ptf_geopos)
  ptWorldPos = 0x1010,
  //! An orientation (ptf_ori)
  ptOrientation = 0x1011,
  //! A ptf_pos_ori
  ptPosOri = 0x1012,

  //! \}

  //! Extensions can be defined and published by anyone. There 
  //! may at some point be a central registry of what the extensions 
  //! mean (or at least who uses them). Please do not use without 
  //! first getting a specific range of types reserved.
  ptFirstExtension = 0x40000000,

  //! Extension range reserved for Makena Technologies, Inc.
  ptfMakenaExtensionStart = 0x50000000,

  //! Content represents areas of "old" terrain surface (i.e. non-PTF).
  ptMakenaOldTerrain = 0x50000000,

  ptfMakenaExtensionEnd = 0x500000FF,

  //! There is a range of types that is reserved and should not be 
  //! used at all; starting at ptFirstReserved up to ptFirstExperimental.
  ptFirstReserved = 0x80000000,

  //! For development experimentation, ptFirstExperimental and up can be 
  //! used. These types should NOT be used in production, or publicly 
  //! documented, as there is no arbitration. If you want to use an 
  //! experimental type in production, please promote the type to a 
  //! ptFirstExtension based type.
  ptFirstExperimental = 0xC0000000,


  ptTestType1 = 0xC0000001,     //!< For testing purposes, this record type can contain arbitrary data.
  ptTestType2 = 0xC0000002,     //!< For testing purposes, this record type can contain arbitrary data.
  ptTestType3 = 0xC0000003,     //!< For testing purposes, this record type can contain arbitrary data.
};

//! Records of type ptAnnotation contain name/value pairs. Note that strings are 
//! always UTF-8 encoded and 0 terminated.
struct ptf_annotation_entry {
  ptf_name                    name;
  ptf_value                   value;
};
typedef struct ptf_annotation_entry ptf_annotation_entry;

//! Records of type ptBookmark indicate a named location of interest in the file.
struct ptf_bookmark {
  ptf_name                    name;
  ptf_geopos                  pos;
  double                      radius;
  ptf_ori                     ori;
};
typedef struct ptf_bookmark ptf_bookmark;

//! The position and orientation of something. The position is in gecentric 
//! coordinate space. The orientation is local to that position (i e, the object 
//! turns around its local origin).
struct ptf_pos_ori {
  ptf_geopos                  pos;                   //!< where in global space it's centered
  ptf_ori                     ori;                   //!< how it's oriented (relative to object-local origin)
};

//! File definitions are used by record references. For example, if a record 
//! references "file 3" then information about what file that is is found in 
//! the ptFileDefinition record at TOC index 3.
//! It is expected that all file definition records refer to PTF files.
struct ptf_file_definition_record {
  ptf_name                    name;           //!< Canonical name of file (as found in header of that file)
  ptf_value                   rel_path;       //!< Relative path to the file when the definition was created.
  ptf_path                    fixed_path;     //!< Last path the file was found at (can be absolute).
};
typedef struct ptf_file_definition_record ptf_file_definition_record;

//! A UUID (also known as a GUID).
struct ptf_uuid {
  ptf_8                       data[16];       //!< The UUID data in network byte order
};

//! Octree blocks start with a block header, followed by a number of node records, 
//! followed by the same number of content record indices, followed by some padding 
//! (to alignment 16) followed by sub-block record refs. There may be more empty 
//! space between these data items, hence the block header contains actual offset 
//! values to locate the data within the block.
//! \see ptf_octree_node
struct ptf_octree_block_header {
  ptf_16                      num_nodes;      //!< Number of node records (and number of content pointers).
  ptf_16                      num_links;      //!< Number of link indices.
  ptf_32                      content_array_pointer;  //!< Offset from beginning of chunk to first content pointer. 
                                              //!< \see ptf_fixed_ref
  ptf_32                      link_pointer;   //!< Offset from beginning of chunk to first sub-block link pointer. 
                                              //!< \see ptf_remote_ref
  ptf_16                      level;          //!< How far down from the top of the octree this block is rooted.
  ptf_16                      node_array_pointer;   //!< Offset from beginning of chunk to first node struct.
};

//! A ptf_octree_node references child nodes within the same block, with an index 
//! between 1 and 4095. Separately indexed, for each node, within the block, there 
//! is also an index value for a node content type record.
//!
//! When the high bit is set of the child link, the value represents a sub-block 
//! reference, rather than a node within this block. Such a reference references 
//! another block, potentially in another file, pointing at the root node of that block.
//! The actual block reference is a ptf_remote_ref, indexed by the bottom 15 bits of 
//! the child pointer, into the link_pointer table of the ptf_octree_block_header.
//!
//! A value of 0 for a child means that there is no child.
struct ptf_octree_node {
  ptf_16                      child[2][2][2]; //!< Array indices: [right(1}/left{0)] [up(1)/down(0)] [hither(1)/yon(0}]
                                              //!< Note: If you want to store left-handed data in PTF, just slam the values 
                                              //!< into the file as-is, as the "mirroring" only happens in the interpretation
                                              //!< of the data -- the math will work equally well no matter what. The left-
                                              //!< handed-ness comes from the interpretation you make of the data.
                                              //!< "hither" really is an arbitrary name for "increasing Z".
};

//! \see ptf_child_index
enum ptf_child_bits {
  ptf_child_bit_hither = 1,
  ptf_child_bit_yon = 0,
  ptf_child_bit_top = 2,
  ptf_child_bit_bottom = 0,
  ptf_child_bit_right = 4,
  ptf_child_bit_left = 0
};

//! Identify children in an eight-sector sliced cube.
enum ptf_child_index {
  ptf_left_bottom_yon_child     = ( ptf_child_bit_left  | ptf_child_bit_bottom | ptf_child_bit_yon ),
  ptf_left_bottom_hither_child  = ( ptf_child_bit_left  | ptf_child_bit_bottom | ptf_child_bit_hither ),
  ptf_left_top_yon_child        = ( ptf_child_bit_left  | ptf_child_bit_top    | ptf_child_bit_yon ),
  ptf_left_top_hither_child     = ( ptf_child_bit_left  | ptf_child_bit_top    | ptf_child_bit_hither ),
  ptf_right_bottom_yon_child    = ( ptf_child_bit_right | ptf_child_bit_bottom | ptf_child_bit_yon ),
  ptf_right_bottom_hither_child = ( ptf_child_bit_right | ptf_child_bit_bottom | ptf_child_bit_hither ),
  ptf_right_top_yon_child       = ( ptf_child_bit_right | ptf_child_bit_top    | ptf_child_bit_yon ),
  ptf_right_top_hither_child    = ( ptf_child_bit_right | ptf_child_bit_top    | ptf_child_bit_hither )
};

//! The ptOctreeContent type data starts out with a header, describing what else
//! is in the record.
//! The "info_offset" values will be zero until such time as the additional 
//! information data format is defined.
struct ptf_node_content_header {
  ptf_32                      num_types;             //!< number of ptf_node_content_type
  ptf_32                      content_info_offset;   //!< where in the record content-specific info lives, or 0
};

//! For each type declared by the num_types field of ptf_node_content_header, 
//! there is one of these structs. They follow immediately after the 
//! ptf_node_content_header structure in the record.
//! This struct defines a specific thing being referenced -- for example, a 
//! given tree model, a given piece of terrain skin, etc.
struct ptf_node_content_type {
  ptf_record_ref              type;                  //!< type of data (mesh, raster, etc)
  ptf_record_ref              meta_data_ref;         //!< reference to record containing meta data for this content
  ptf_32                      num_instances;         //!< number of ptf_node_content_instance following
  ptf_32                      type_info_offset;      //!< where in the record instance-specific info lives, or 0
};

//! For each instance of the content type in ptf_node_content_type, there is 
//! instance information about the transformation to take the referenced 
//! data from object-local space to octree-global (geocentric) space.
//! The transform is such that an object-local vertex is first rotated 
//! using the orientation, and then translated using the position.
typedef struct ptf_pos_ori ptf_node_content_instance;

//! Node content annotation is used for the info_offset blocks.
struct ptf_node_content_annotation {
  ptf_32                      size;                  //!< size of data following (including size and type)
  ptf_type                    type;                  //!< type of data following, or 0 for "experimental."
};  //  followed by data

//! External references are used by ptf_type values ptMeshReference, 
//! ptRasterReference and ptShapeReference. Various file types can have 
//! additional data after the name, if defined in the standard. If you 
//! need additional data that's not in the standard, use an extension 
//! type record to store that information.
//! \see ptf_external_reference_ptf
struct ptf_external_reference {
  ptf_name                    filename;              //!< The file name of the file being referenced.
};

//! For external references to PTF files, the actual type/record being 
//! referenced is stored after the file name. Yes, this is somewhat 
//! redundant; the record_ref has a file reference, which refecences a 
//! ptf file reference record, which stores the file name again.
struct ptf_external_reference_ptf {
  ptf_external_reference      ptf;                   //!< The name of the file referenced.
  ptf_record_ref              record;                //!< The record being referenced.
};

struct ptf_semantic_ref {
  ptf_32          semantic;
  ptf_32          semantic_info;
  ptf_record_ref  reference;  // note: file_ref, because some semantics may point outside file
};

enum {
  pmsLodTable = 0x1,        //!< record_ref to the LOD table for this mesh
};


//! Storage of a mesh starts with the ptf_mesh_header, followed by 
//! some number of ptf_vertex_component_description structs, followed by 
//! some number of ptf_mesh_subset_description structs, followed by the 
//! vertex data (padded up to 8 byte size multiple), followed by each of 
//! the subset index lists (padded up to 8 byte multiple in size).
//! When storing a mesh in a file on disk, there is an additional file 
//! header known as a ptf_disk_file_header.
//! \see ptf_vertex_component_description, ptf_mesh_subset_description.
struct ptf_mesh_header {
  ptf_name                    name;                  //!< Meshes have names, given them by artists.
  ptf_8                       num_components;        //!< Number of vertex component descriptions. Range [1, 127]
  ptf_8                       num_edge_lists;        //!< Number of edge lists
  ptf_16                      num_subsets;           //!< Number of subset descriptions following vertex component descriptions. 
                                                     //!< Range [1, 4096]
  ptf_16                      vertex_stride;         //!< Size (in bytes) of each vertex. Range [1, 1024]
  ptf_16                      file_alignment;        //!< All data blocks are aligned to this value, as are individual vertex 
                                                     //!< components. Typical value will be 4 (or possibly 8).
  ptf_32                      data_offset;           //!< Offset from start of header for first semantic
                                                     //!< Set to sizeof(ptf_mesh_header) for now (other values are reserved).
  ptf_32                      vertex_count;          //!< Number of vertices in mesh. Range [1, 2 million]
  ptf_remote_ref              material_archive;      //!< Material archive used for materials in this mesh
  ptf_16                      num_semantics;         //!< Number of separate semantics (\see ptf_mesh_semantic)
  ptf_16                      spare1;                //!< Set to 0, ignore on read.
  float                       feature_size;          //!< "Feature size" of this mesh; typically the average of the shortest 
                                                     //!< non-degenerate edge within the mesh, and the median edge length 
                                                     //!< within the mesh, but other measures can be used.
};

//! Vertex component types. Also subset index element types, with the crucial different that 
//! subset index elements aren't scaled to a range (so pvtUInt16 is a valid index, pvtUInt16_8 isn't).
//! The scaling is intended to allow compact representation of vertex values that don't need full
//! 32-bit floating points, such as normal vectors or color values. Thus, pvtUint8 is ideal for 
//! color storage, because "0" means "0.0" and "255" means "1.0" as is the convention. U/V values 
//! could easily be stored as pvtInt16_4, which gives a range of [-8..8). Note the half-open interval 
//! there: the modified values use a power-of-two scaler, whereas the unmodified values use a 
//! scale that allows both 0 and 1 to be precisely represented.
//!
//! \see ptf_type_unpack().
enum {
  pvtInvalid = 0,
  pvtFloat32 = 1,   //!< IEEE 32-bit float values, with negative 0, denormals, +/-inf and NaN being illegal.
  pvtUInt32 = 2,    //!< Actual integral values.
  pvtInt32 = 3,     //!< Actual integral values.
  pvtUInt16 = 4,    //!< Assumed to be normalized in [0..1] range
  pvtInt16 = 5,     //!< Assumed to be normalized in [-1..1] range with -32768 being illegal (so 0 can be represented)
                    //!< Note the discrepancy with the 0..1 range of the unsigned type!
  pvtUInt8 = 6,     //!< Assumed to be normalized in [0..1] range
  pvtInt8 = 7,      //!< Assumed to be normalized in [-1..1] range with -128 being illegal (so 0 can be represented)
                    //!< Note the discrepancy with the 0..1 range of the unsigned type!

  pvtUInt16_4 = 0x14, //!< Assumed to be normalized in [0..16) range (note half-open interval)
  pvtInt16_4 = 0x15, //!< Assumed to be normalized in [-8..8) range (note half-open interval)
  pvtUInt8_4 = 0x16, //!< Assumed to be normalized in [0..16) range (note half-open interval)
  pvtInt8_4 = 0x17, //!< Assumed to be normalized in [-8..8) range (note half-open interval)

  pvtUInt16_8 = 0x24, //!< Assumed to be normalized in [0..256) range (note half-open interval)
  pvtInt16_8 = 0x25, //!< Assumed to be normalized in [-128..128) range (note half-open interval)
  pvtUInt8_8 = 0x26, //!< Assumed to be normalized in [0..256) range (note half-open interval)
  pvtInt8_8 = 0x27, //!< Assumed to be normalized in [-128..128) range (note half-open interval)

  pvtUInt16_12 = 0x34, //!< Assumed to be normalized in [0..4096) range (note half-open interval)
  pvtInt16_12 = 0x35, //!< Assumed to be normalized in [-2048..2048) range (note half-open interval)
  pvtUInt8_12 = 0x36, //!< Assumed to be normalized in [0..4096) range (note half-open interval)
  pvtInt8_12 = 0x37, //!< Assumed to be normalized in [-2048..2048) range (note half-open interval)

  pvtUInt16_16 = 0x44, //!< Assumed to be normalized in [0..65536) range (note half-open interval)
  pvtInt16_16 = 0x45, //!< Assumed to be normalized in [-32768..32768) range (note half-open interval)
  pvtUInt8_16 = 0x46, //!< Assumed to be normalized in [0..65536) range (note half-open interval)
  pvtInt8_16 = 0x47, //!< Assumed to be normalized in [-32768..32768) range (note half-open interval)

  pvtMaxType,
};

//! Vertex component usages
enum {
  pvuInvalid = 0,         //!< Could be used for unused padding within a vertex.
  pvuPosition = 1,
  pvuNormal = 2,          //!< Binormal along U == usage index 1, Tangent along V == usage index 2
  pvuTextureCoord = 3,
  pvuColor = 4,
  pvuIndex = 5,
  pvuWeight = 6,
  pvuGeneric = 7,         //!< A generic value with no specific semantic. Used sparingly, if at all.
  pvuMaxUsage,
};

//! For each vertex component, there is a description of the component in 
//! the mesh preamble, following the ptf_mesh_header. Each vertex component 
//! is assumed to be packed following the previous data aligned at the 
//! vertex_alignment alignment boundary. For example, a 6-byte vertex component 
//! in a 4-byte aligned vertex would actually take up 8 bytes.
//! The array of vertex components counts as one block for purposes of file alignment.
struct ptf_vertex_component_description {
  ptf_8                       usage;                  //!< For example, "position" or "binormal" or "diffuse color". 
                                                      //!< \see pvuPosition
  ptf_8                       usage_index;            //!< 0 for "the first," 1 for "the second" etc. In range [0,7]
  ptf_8                       type;                   //!< Typically float or unsigned byte. \see pvtFloat32 and pvtUInt8.
  ptf_8                       type_count;             //!< Number of components (3 for a normal, 2 for a UV coordinate, etc). 
                                                      //!< In range [0,4]
};

//! For each subset of the mesh, there is a subset description. The subset 
//! can reference a material (of type ptMaterial), and also defines what kind 
//! of primitive is involved (line list, triangle list, quad list, or higher 
//! order polygon list).
//!
//! Material references are handled specially in PTM files; the "material.index" 
//! value is actually an offset into the file data, where the material record 
//! data is stored, and the "material.file" value should be 0. When found in a 
//! PTF file, the "material" field is a true ptf_remote_ref to a record of type 
//! ptMaterial.
//!
//! \todo ptMaterial is no more; need to update this description!
//!
//! All ptf_mesh_subset_description instances count as one array for purposes of 
//! file alignment.
struct ptf_mesh_subset_description {
  ptf_32                      material_index;         //!< Index into material archive for the mesh to material used.
  ptf_16                      reserved_0;             //!< Set to 0.
  ptf_8                       primitive_type;         //!< 1 for points, 2 for lines, 3 for triangles, 4 for quads, ...
  ptf_8                       index_type;             //!< The only supported values are unsigned short, or unsigned int. 
                                                      //!< \see pvtUInt16 and pvtUInt32.
  ptf_32                      num_indices;            //!< Number of indices in the index buffer. Range [1, 2 million]
  ptf_32                      min_index;              //!< What is the minimum index value in this subset?
  ptf_32                      max_index;              //!< What is the maximum index value in this subset?
};

//! bounding plane values for retrieving an edge list from a mesh.
enum {
  MeshMinPlane = 0,
  MeshLeftPlane = 0,      //!< -X
  MeshRightPlane = 1,     //!< +X
  MeshBottomPlane = 2,    //!< -Y
  MeshTopPlane = 3,       //!< +Y
  MeshYonPlane = 4,       //!< -Z
  MeshHitherPlane = 5,    //!< +Z
  MeshMaxPlane
};

//! Each mesh has a list of edges, which are intended to hook up to edges on a neighboring 
//! mesh. This list is stored using the following structure.
struct ptf_mesh_edge_list_description {
  ptf_8                       plane_index;    //!< Octree bounding plane - implicitly defines normal. See "MeshLeftPlane" enum
  ptf_8                       reserved1;      //!< reserved
  ptf_16                      reserved2;      //!< reserved
  ptf_32                      plane_d;        //!< "d" parameter for bounding plane equation (ax + by + cz = d)
  ptf_32                      num_edges;      //!< Number of edges in this list
  ptf_32                      num_hier;       //!< number of parents in the hierarchy data
};

//! Each edge has the following information. \see ptf_mesh_edge_list_description.
struct ptf_mesh_edge {
  ptf_32                      edge_id;        //!< edge ID (unique per plane)
  ptf_32                      start_vtx;      //!< start vertex index
  ptf_32                      end_vtx;        //!< ending vertex index
  ptf_32                      material_index; //!< Index into material archive for the mesh to material used.
};

//! Vertex information is stored per subset, using a header and three arrays:
//! 1) An array of ptf_32, one per subset, storing the offset for each 
//! subset into the array of information records. One offset per subset; all 
//! information records for a subset are contiguous.
//! 2) An array of ptf_subset_vertex_information_record records, into which the 
//! previous offsets point. These in turn have size/offset pointers into the 
//! data array, which follows. This array is file aligned, but the array records 
//! are not.
//! 3) A data array, containing raw byte data. The base of this array is 
//! file aligned.
struct ptf_subset_information_header {
  ptf_32                      num_records;    //!< Number of ptf_subset_vertex_information_record records.
  ptf_32                      data_size;      //!< Number of bytes of raw data (referenced by the records).
};

//! \see struct ptf_subset_vertex_information_header.
struct ptf_subset_information_record {
  ptf_32                      type;           //!< type of this info record
  ptf_32                      size;           //!< size of data in the storage array
  ptf_32                      offset;         //!< offset into the storage array
  ptf_32                      reserved_0;     //!< write as 0, ignore on read
};

//! For subset information, there are various kinds of information stored
//! for different mesh subsets. The kinds of information include the absolute 
//! minimum value of the various vertex components and the absolute maximum.
enum {
  pviSubsetInfoMin = 1,                       //!< Minimum value of each vertex component over the subset
  pviSubsetInfoMax = 2,                       //!< Maximum value of each vertex component over the subset
};

//! A PTF content record disk file (such as a stand-alone mesh) will 
//! start out with a ptf_toc_entry (typedefed to ptf_disk_file_header).
//! To get a good magic number for use in file identification, the 
//! 'offset' member should be assigned the value PTF_DISK_FILE_OFFSET.
#define PTF_DISK_FILE_OFFSET 0x0a21746e6f63667470LL

//! The ptf_disk_file_header should contain the value PTF_DISK_FILE_OFFSET 
//! in the 'offset' member, and the actual type of the record being 
//! stored on disk in the 'type' member.
//! This is for files that store individual, non-inline meshes, rasters 
//! etc. File types PTM, PTR, PTS and PTT are examples of this.
//! Note that all offsets "from start of record" do NOT include the 
//! ptf_disk_file_header, only the data header (so the data values are 
//! identical inside a PTF file, and in a file on disk). In fact, you can 
//! mechanically translate between all-internalized data storage, and all-
//! externalized data storage.
typedef ptf_toc_entry ptf_disk_file_header;

//! Storage for a sample value for an elevation raster.  A union is used because raster elevation
//! instances can have different data storage types for their underlying samples.
typedef union {
  ptf_8   int8;
  ptf_16  int16;
  ptf_32  float32;
} ptf_elev_sample_val;

//! Header for raster elevation data, as stored on disk.
//! Coordinate system for raster elevation data is an implicit "flat" projected system in PTF geocentric 
//! space.
//! Sample values are signed.  Negative sample values indicate that the elevation is below sea level.
//! Data can be 8 bit int, 16 bit int, or 32 bit float.  Data values can be scaled in X, Y, and/or Z.
//! Data is stored as an array that follows the header.  Data is stored in row-order,
//! where the southernmost row is at the beginning of the data array and northernmost
//! row is at the end of the data.
//! Origin of the geographic coverage for this record is at the center of the block.
//! It is expected that a fully covered globe will contain some overlap at the corners 
//! between raster blocks, and that the raster blocks will touch at the centers of the 
//! edges, to allow for a full mapping of the globe. It is up to the PTF reader to 
//! determine what to do about the overlap, although the recommended approach is to 
//! use a cutting plane aligned through the shared center point to classify points as 
//! belonging to one side or the other.
struct ptf_elev_header {
  ptf_name                    name;           //!< dataset identifier (48 bytes)
  ptf_8                       sample_type;    //!< size and type of data samples (see "pvt" enum above)
  ptf_8                       flags;          //!< bit 0 = "has null data".  bits 1-7 are reserved
  ptf_16                      spare1;         //!< reserved

  ptf_32                      data_offset;    //!< offset to raster data from start of record
  ptf_32                      data_length;    //!< length of raster data (excluding header)

  ptf_16                      num_cols;       //!< number of columns of data
  ptf_16                      num_rows;       //!< number of rows of data

  ptf_elev_sample_val         null_data_val;  //!< sentinel value for samples that contain no data (32 bits)
                                              //!< this value is only meaningful if the "has null data" flag is set (above)

  ptf_32                      x_scale;        //!< horizontal scale: east-west (?)
  ptf_32                      y_scale;        //!< horizontal scale: north-south (?)
  ptf_32                      z_scale;        //!< vertical scale: up-down

  // dataset not required to be square (i.e. not required that #rows == #cols)
  // also samples are not required to be square
  // however, datasets and samples must be rectangular in the PTF coordinate system
  ptf_32                      extent_x;       //!< dataset half width (distance east/west from origin to edge of data)
  ptf_32                      extent_y;       //!< dataset half height (distance north/south from origin to edge of data)

  ptf_32                      proj_lon;       //!< projection origin longitude
  ptf_32                      proj_lat;       //!< projection origin latitude

  ptf_32                      z_offset;       //!< z offset for the sample data.  Useful for unsigned data.
  ptf_32                      spare2;         //!< reserved
};

//! Types of a piece of a material description. Materials are hierarchies of 
//! marshalled properties, which have well-defined binary types, or are 
//! collections of other properties (similar to XML node vs attribute). 
//! Properties are identified by name, and are typed.
//! Materials are marshalled using byte format, so no alignment rules are 
//! enforced; you will have to memcpy() to pack/unpack.
//! \see ptf_material_archive_create_material
//! \see ptf_material_archive_create
enum ptf_material_type {
  pmtNull = 0,
  pmtVariantData = 0x100,
  pmtFloatTuple = 0x101,
  pmtDoubleTuple = 0x102,
  pmtString = 0x103,
  pmtRecordRef = 0x104,
  pmtUint32Tuple = 0x105,
  pmtVariantMap = 0x200,
  pmtTextureMapExplicit = 0x201,
  pmtTextureMapImplicit = 0x202,
  pmtTextureMapGridded = 0x203,
  pmtMaterialMap = 0x204,
  pmtMaterial = 0x205,
  pmtMaterialArchive = 0x206,
  pmtShadowMap = 0x207,
  pmtShader = 0x208,
  pmtVariantAny = 0x300,
};

#pragma pack(push, 2)
//! Material archives have a simple header, followed by marshalling a number of 
//! serialized materials as simple child properties. Each material is serialized 
//! as a material property, which contains child properties for the actual data.
//! There are various hints in the header structs for properly being able to 
//! read future versions of material archives in older code. The "version_to_read" 
//! field assumes that these instructions are followed in older code.
//! You may not understand all of the new data, but you should know enough to be 
//! able to safely skip over it.
typedef struct ptf_material_archive_header ptf_material_archive_header;
struct ptf_material_archive_header {
  ptf_16            version_written;  //!< The version of the material format.
  ptf_16            version_to_read;  //!< The minimum version needed to understand the stream.
  ptf_16            extension_size;   //!< Size of extension data in header. Write as 0. Skip this much on read.
};
//! ptf_material_parameter_header supports properties that only have children (no data).
//! Note that properties can have data, or children, but not both.
typedef struct ptf_material_parameter_header ptf_material_parameter_header;
struct ptf_material_parameter_header {
  ptf_16            type;             //!< If high bit set (0x8000), then don't add to parent.
  ptf_16            prop_desc;        //!< If 0x0100 or higher, a custom parameter; name follows header as 
                                      //!< zero-terminated string. 
                                      //!< The lower byte contains the count of additional parameter bytes 
                                      //!< of the header (before the name) in this case.
                                      //!< If lower than 0x100, index into fixed-child-parameter array for type 
                                      //!< information.
  ptf_16            child_count;      //!< If 0, followed by data size. Else, can be any number.
};

//!< data_size        //!< Value must be < 0x8000.
                    //!< If value is 0x8000 or higher, read another ptf_16, and 
                    //!< combine the lowest 15 bits of that value with the lowest 
                    //!< 15 bits of this value. 
                    //!< size = (((data_size & 0x7fff) << 15) | (new_value & 0x7fff))
#pragma pack(pop)

//! Flags on a material, typically added as a Uint32 property named "flags".
enum ptf_material_flags {
  pmfNotCollidable = 0x1,             //!< Material does not prevent physical movement.
  pmfAlphaBlend = 0x2,                //!< Material is drawn with alpha blend. Framebuffer blend function is ONE,ONE_MINUS_SRC_ALPHA.
  pmfColorModulate = 0x4,             //!< Material is drawn with color modulation.
  pmfColorAdd = 0x8,                  //!< Material is drawn additively
};


//! Textures can be stored in a few formats. The only formats that a PTF reader 
//! is required to understand are RAW 32-bit BGRA, and the DXT1 and DXT5 compressed 
//! flavors, but more flavors are likely to be popular, so we also define how to 
//! store textures in a variety of formats.
//! Textures can also be stored as external file references.
//! The ptf_texture_data header is followed by an array of ptf_texture_mipmap_desc 
//! structures.
typedef struct ptf_texture_data ptf_texture_data;
struct ptf_texture_data {
  ptf_8             format;           //!< One of a few well-defined formats. Not all of the 
                                      //!< formats are guaranteed to be understood.
                                      //!< \see ptf_texture_data_format
  ptf_8             num_mipmaps;      //!< Number of MIP map levels stored in the file. 1 means there is only a 
                                      //!< single image.
  ptf_8             num_surfaces;     //!< 1 for regular texture, 6 for cube reflection map
  ptf_8             dimensions;       //!< 2 for 3D texture. Values 1, 2 and 3 are allowed.
  ptf_8             border;           //!< The width of MIP map level N is <tt>(((width - 2*border) >> N) + 2*border)</tt>.
                                      //!< The border size is actually included in the width, height and depth fields.
  ptf_8             reserved1[3];     //!< Write as 0, ignore on read.
  ptf_32            data_offset;      //!< Offset where texture data begins (from start of record).
  ptf_32            data_size;        //!< Size of all texture data in file.
  ptf_32            width;            //!< Width in texels of MIP map level 0. Includes any border.
  ptf_32            height;           //!< Height in texels of MIP map level 0. Includes any border. If 1D texture, 
                                      //!< value 1 should be written.
  ptf_32            depth;            //!< Depth in texels of MIP map level 0. Includes any border. If 2D texture, 
                                      //!< value 1 should be written.
  ptf_32            alignment;        //!< Each scan line is aligned to this many bytes. Must be a power of two. 
                                      //!< For compressed formats, is alignment of mip level images.
};

//! \see ptf_texture_data
//! There is at least one of these following the ptf_texture_data header, and 
//! usually more (for each surface, there are num_mipmaps of them in a row).
typedef struct ptf_texture_mipmap_desc ptf_texture_mipmap_desc;
struct ptf_texture_mipmap_desc {
  ptf_32            start_offset;     //!< Offset within chunk of texture data (i e 0 for the first mipmap)
  ptf_32            size;             //!< Size within chunk of texture data.
};


//! An enumeration of well-defined texel data formats.
enum ptf_texture_data_format {
  ptdNull = 0x0,                      //!< No data

  ptdRawBgra8 = 0x1,                  //!< Bytes in order B, G, R, A, 255 == 1.0, 0 = 0.0. Supported by all readers.
  ptdRawL8 = 0x2,                     //!< Single bytes; 255 = 1.0, 0 = 0.0. Supported by all readers.
  ptdRawBgra16 = 0x3,                 //!< 16-bit little-endian words in order B, G, R, A, 65535 == 1.0, 0 = 0.0
  ptdRawL16 = 0x4,                    //!< Single little-endian words, 65535 == 1.0, 0 = 0.0
  ptdRawBgraF16 = 0x5,                //!< little-endian 16-bit floats in order B, G, R, A (1s5e10m)
  ptdRawLF16 = 0x6,                   //!< Single little-endian 16-bit floats (1s5e10m)
  ptdRawBgraF32 = 0x7,                //!< little-endian 32-bit floats in order B, G, R, A (1s8e23m)
  ptdRawLF32 = 0x8,                   //!< Single little-endian 32-bit floats (1s8e23m)

  ptdDxt1 = 0x40,                     //!< Data compressed using S3TC compression, using the Microsoft Direct3D DXT1 
                                      //!< encoding. Only the compressed data, no headers. Very small MIP levels are 
                                      //!< rounded up to 4x4 pixels. Supported by all readers.
  ptdDxt5 = 0x41,                     //!< Data compressed using S3TC compression, using the Microsoft Direct3D DXT5 
                                      //!< encoding. Only the compressed data, no headers. Very small MIP levels are 
                                      //!< rounded up to 4x4 pixels. Supported by all readers.
  ptdJPEG = 0x42,                     //!< A complete JPG file, embedded in the record.
  ptdJPEG2000 = 0x43,                 //!< A complete JP2 file (basic proflie), embedded in the record.
  ptdPNG = 0x44,                      //!< A complete PNG file, embedded in the record.
};

//! A collection relates several meshes in different octree nodes together to make a "meta-mesh".
//! This is useful for example for a terrain skin that is distributed through many octree nodes, 
//! and in turn has different LOD bits in different levels. There will only be one collection 
//! reference in the ptContent chunk for the octree node, and the further subdivision of the 
//! meta-mesh will be stored in this collection (which in turn references ptMesh chunks for the 
//! actual geometry).
typedef struct ptf_collection_header ptf_collection_header;
struct ptf_collection_header {
  ptf_32            left;             //!< Left coordinate of topmost containing octree node for this collection
  ptf_32            bottom;           //!< Bottom coordinate of topmost containing octree node for this collection
  ptf_32            yon;              //!< Yon coordinate of topmost containing octree node for this collection
  ptf_16            octree_level;     //!< Topmost octree level that contains this collection (the starting point)
  ptf_16            num_nodes;        //!< Total number of nodes that make up this collection
  ptf_32            data_offset;      //!< Record offset to beginning of collection data
  ptf_type          type;             //!< The type of data organize (e g, ptMeshData).
  ptf_32            related;          //!< The number of related ptf_record_ref.
  ptf_32            related_offset;   //!< Record offset to beginning of related record data.
  ptf_name          name;             //!< Collection name
};

//! A LOD chunk is of type ptLodData, and defines a number of related references, all 
//! relating to the same model/mesh.
//! \note For terrain skin type geometry, there is a separate way of doing LOD. 
//! \see ptf_collection_header
typedef struct ptf_lod_header ptf_lod_header;
struct ptf_lod_header {
  ptf_16            num_semantics;    //!< Number of separate semantics (\see ptf_lod_semantic)
  ptf_16            data_offset;      //!< Offset from start of header for first semantic
  ptf_32            _reserved;        //!< Write as 0.
};

//! For each semantic within a LOD group, there is one of these headers, 
//! followed by the actual data items for that semantic. "Semantic" means 
//! something like "LOD mesh table" or "collision mesh reference".
//! \see ptf_lod_semantic
typedef struct ptf_lod_semantic_header ptf_lod_semantic_header;
struct ptf_lod_semantic_header {
  ptf_32            semantic_size;    //!< Size of total data for this semantic, starting at the start of this header.
                                      //!< Always a multiple of 8.
  ptf_32            item_semantic;    //!< The actual semantic being described.
  ptf_16            num_items;        //!< Number of items in the table
  ptf_16            item_size;        //!< Size of each item (for fixed-size data).
  ptf_32            _reserved;        //!< Write as 0.
};

//! For extensions, use the PTF type extension name space.
enum ptf_lod_semantic {
  plsNone                   = 0x0,          //!< nothing; dead data
  plsLodTable               = 0x1,          //!< struct ptf_lod_table_entry
  plsGroupName              = 0x2,          //!< ptf_name for the entire lod group
  plsOdeTrimeshCollision    = 0x40000001,   //!< record_ref to a "cooked" ODE trimesh collision chunk
  plsPhysXTrimeshCollision  = 0x40000002,   //!< record_ref to a "cooked" PhysX trimesh collision chunk
  plsExperimental           = 0xc0000001,   //!< for testing purposes
};

//! For plsLodTable semantic data, this is the data item stored.
//! The table is sorted in coarsest to finest resolution mesh (highest to lowest feature size).
typedef struct ptf_lod_table_entry ptf_lod_table_entry;
struct ptf_lod_table_entry {
  ptf_fixed_ref     mesh;                   //!< The mesh for this level of detail
  float             feature_size;           //!< The feature size of the mesh being referenced
};

//! For mesh instance meta data (one possible type for the record pointed at by a local_ref of 
//! a content reference annotation), the header is as follows. Within the chunk of data, the 
//! SEDRIS classification data can be found at the position (and with the size) indicated. 
//! Additionally, if the UUID is all 0, it means this mesh is free-standing, whereas if it is 
//! non-0, all other mesh instances with that same UUID are made of the same original "thing" 
//! (useful when using recursive octree LOD with binders for mesh content). 
typedef struct ptf_geometry_instance_meta_data_header ptf_geometry_instance_meta_data_header;
struct ptf_geometry_instance_meta_data_header {
  ptf_uuid          uuid;                             //!< The UUID of the original "thing," or all zeros.
  ptf_32            sedris_classification_offset;     //!< The offset within the data block (from the beginning); 0 means no 
                                                      //!< data.
  ptf_32            sedris_classification_size;       //!< The size after that offset; 0 means no data.
  ptf_32            layer_count;                      //!< How many layers are these instances members of? 0 for none.
  ptf_32            layer_offset;                     //!< Where in the data chunk are the layer identifying records?
                                                      //!< Each layer identifying record is a ptf_record_ref.
};

//! A layer is like a group of objects, where each instance in a content can be part of 
//! zero or more layers. \see ptLayerData
//! For a layer, we define a name and a SEDRIS classification (which is optional). 
//! You can also attach some number of key/value pairs (in text) to further describe 
//! the layer.
typedef struct ptf_layer_data_header ptf_layer_data_header;
struct ptf_layer_data_header {
  ptf_uuid          uuid;                             //!< The UUID of this layer, if the same layer is stored in more than 
                                                      //!< one file. All zeros means globally unique across files.
  ptf_name          name;                             //!< The user-visible name of the layer.
  ptf_32            sedris_classification_offset;     //!< If set, all members of the layer have the same classification.
  ptf_32            sedris_classification_size;       //!< Amount of SEDRIS classification data. If 0, no classification data.
  ptf_32            annotation_count;                 //!< Number of ptf_annotation_entry records in this data chunk.
  ptf_32            annotation_offset;                //!< The start of the first annotation record in this data chunk.
};


#endif  //  ptf_types_h
