
#include "ptf_content.h"

#include <vector>
#include <stdexcept>
#include <assert.h>

namespace {
  typedef ptf_node_content_instance reference_instance;
  struct content_reference {
    ptf_record_ref ref;
    ptf_record_ref meta_data_ref;
    size_t hdrpos;
    ptf_type data_type;
    std::vector<char> data;
    std::vector<reference_instance> instances;
  };

  ptf_32 append_padded(std::vector<char> & out, void const * data, size_t size)
  {
    if (out.size() + size > 0x10000000UL) {
      throw std::runtime_error("too large for machine");
    }
    size_t pos = out.size();
    size_t s8 = (size + 7) & -8;
    out.resize(pos + s8);
    memcpy(&out[pos], data, size);
    if (s8 > size) {
      memset(&out[pos + size], 0, s8-size);
    }
    return (ptf_32)pos;
  }
}


struct ptf_content_t {
  ptf_content_t() { sig = 'cnt '; }
  ~ptf_content_t() { sig = 0; }
  bool validate() { return sig == 'cnt '; }

  ptf_32 sig;

  ptf_err parse_data(void const *data, size_t size);
  ptf_err make_data(void **o_data, size_t *o_size);

  ptf_type data_type;
  std::vector<char> data;
  std::vector<content_reference> references;
  std::vector<char> outputblock;
};


ptf_err ptf_content_t::parse_data(void const *data, size_t osize)
{
  size_t size = osize;
  char const * base = (char const *)data;
  assert(references.size() == 0);
  if (size < sizeof(ptf_node_content_header)) {
    return peInvalidFile;
  }
  ptf_node_content_header nch = *(ptf_node_content_header const *)base;
  base += sizeof(ptf_node_content_header);
  size -= sizeof(ptf_node_content_header);
  if (nch.content_info_offset != 0) {
    if (nch.content_info_offset > osize-sizeof(ptf_node_content_annotation)) {
      return peInvalidFile;
    }
    ptf_node_content_annotation nca = *(ptf_node_content_annotation const *)
        ((char *)data + nch.content_info_offset);
    if (nca.size + nch.content_info_offset > osize) {
      return peInvalidFile;
    }
    this->data.resize(nca.size - sizeof(nca));
    if (this->data.size() > 0) {
      memcpy(&this->data[0], (char *)data + nch.content_info_offset + sizeof(nca), 
          this->data.size());
    }
    this->data_type = nca.type;
  }
  references.reserve(nch.num_types);
  for (ptf_32 t = 0; t < nch.num_types; ++t) {
    if (size < sizeof(ptf_node_content_type)) {
      return peInvalidFile;
    }
    ptf_node_content_type nct = *(ptf_node_content_type const *)base;
    base += sizeof(nct);
    size -= sizeof(nct);
    if (size < nct.num_instances * sizeof(ptf_node_content_instance)) {
      return peInvalidFile;
    }
    //  create this content reference
    content_reference cr0;
    cr0.ref = nct.type;
    cr0.meta_data_ref = nct.meta_data_ref;
    references.push_back(cr0);
    content_reference & cr = *references.rbegin();
    cr.instances.reserve(nct.num_instances);
    //  parse the annotation data, if present
    if (nct.type_info_offset != 0) {
      if (nct.type_info_offset > osize - sizeof(ptf_node_content_annotation)) {
        return peInvalidFile;
      }
      ptf_node_content_annotation nca = *(ptf_node_content_annotation const *)
        ((char *)data + nct.type_info_offset);
      if (nca.size + nct.type_info_offset > osize) {
        return peInvalidFile;
      }
      cr.data.resize(nca.size - sizeof(nca));
      if (cr.data.size()) {
        memcpy(&cr.data[0], (char *)data + nct.type_info_offset + sizeof(nca), cr.data.size());
      }
      cr.data_type = nca.type;
    }
    //  add all the instances -- file size is already verified
    for (ptf_32 i = 0; i < nct.num_instances; ++i) {
      ptf_node_content_instance nci = *(ptf_node_content_instance const *)base;
      base += sizeof(nci);
      size -= sizeof(nci);
      cr.instances.push_back(nci);
    }
  }
  return peSuccess;
}

ptf_err ptf_content_t::make_data(void **o_data, size_t *o_size)
{
  try {
    std::vector<char> & output = outputblock;
    output.resize(0);
    size_t siz = sizeof(ptf_node_content_header);
    siz += sizeof(ptf_node_content_type) * references.size();
    siz += (data.size()+7) & -8;
    for (size_t i = 0; i < references.size(); ++i) {
      siz += sizeof(ptf_node_content_instance) * references[i].instances.size();
      //  This is assuming that each reference has an annotation header, which 
      //  is conservative. But it doesn't matter much until we get to >1,000,000 
      //  references in a single content -- which we really shouldn't.
      siz += (references[i].data.size() + 15) & -8;
    }
    output.reserve(siz);

    //  create header
    ptf_node_content_header hdr;
    memset(&hdr, 0, sizeof(hdr));
    hdr.num_types = (ptf_32)references.size();
    size_t hdrpos = append_padded(output, &hdr, sizeof(hdr));
    //  foreach type, append type descriptor; remember where it was
    for (ptf_32 i = 0; i < hdr.num_types; ++i) {
      ptf_node_content_type nct;
      memset(&nct, 0, sizeof(nct));
      nct.num_instances = (ptf_32)references[i].instances.size();
      nct.type = references[i].ref;
      nct.meta_data_ref = references[i].meta_data_ref;
      references[i].hdrpos = append_padded(output, &nct, sizeof(nct));
      if (nct.num_instances > 0) {
        //  foreach instance, append instance descriptor
        append_padded(output, &references[i].instances[0], sizeof(references[i].instances[0])
            * nct.num_instances);
      }
    }
    //  append header annotation block; update header
    if (data.size()) {
      ptf_node_content_annotation nca;
      nca.size = (ptf_32)(sizeof(nca) + data.size());
      nca.type = data_type;
      size_t val = append_padded(output, &nca, sizeof(nca));
      append_padded(output, &data[0], data.size());
      ((ptf_node_content_header *)&output[hdrpos])->content_info_offset = (ptf_32)val;
    }
    //  foreach type, append type annotation block; remember where it was
    for (ptf_32 i = 0; i < hdr.num_types; ++i) {
      if (references[i].data.size()) {
        ptf_node_content_annotation nca;
        nca.size = (ptf_32)(sizeof(nca) + references[i].data.size());
        nca.type = references[i].data_type;
        size_t val = append_padded(output, &nca, sizeof(nca));
        append_padded(output, &references[i].data[0], references[i].data.size());
        assert(output.size() >= references[i].hdrpos + sizeof(ptf_node_content_type));
        assert(((ptf_node_content_type *)&output[references[i].hdrpos])->type_info_offset == 0);
        ((ptf_node_content_type *)&output[references[i].hdrpos])->type_info_offset = (ptf_32)val;
      }
    }

    *o_data = &output[0];
    *o_size = output.size();
  }
  catch (...) {
    return peTooBigForMachine;
  }

  return peSuccess;
}







ptf_err ptf_content_create(ptf_content_handle *o_cont)
{
  try {
    *o_cont = new ptf_content_t();
    return peSuccess;
  }
  catch (...) {
  }
  return peException;
}

ptf_err ptf_content_destroy(ptf_content_handle cont)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    delete cont;
    return peSuccess;
  }
  catch (...) {
  }
  return peException;
}

ptf_err ptf_content_read_from_file(ptf_content_handle cont, 
    ptf_file_handle fh, ptf_fixed_ref const &ref)
{
  void * data = 0;
  try {
    if (!cont->validate()) return peInvalidArguments;
    //  The call is guaranteed to clobber any previous data
    cont->~ptf_content_t();
    new(cont) ptf_content_t();
    //  Figure out this records contents
    ptf_64 size = 0;
    ptf_record_ref rref;
    memset(&rref, 0, sizeof(rref));
    rref.type = ptOctreeContent;
    rref.index = ref.index;
    ptf_err pe = ptf_file_get_record_size(fh, &rref, &size);
    if (pe == peSuccess) {
      if ((size > 0x10000000) || !(data = malloc((ptf_32)size))) {
        pe = peTooBigForMachine;
      }
      else {
        pe = ptf_file_read_record(fh, &rref, 0, data, size);
      }
    }
    if (pe == peSuccess) {
      //  Actually decode the data
      pe = cont->parse_data(data, (ptf_32)size);
    }
    if (data) {
      free(data);
    }
    return pe;
  }
  catch (...) {
    if (data) {
      free(data);
    }
  }
  return peException;
}

ptf_err ptf_content_update_to_file(ptf_content_handle cont, 
    ptf_file_handle fh, ptf_fixed_ref const &ref)
{
  void * data = 0;
  size_t size = 0;
  ptf_err pe;
  try {
    if (!cont->validate()) return peInvalidArguments;
    pe = cont->make_data(&data, &size);
    ptf_record_ref rref;
    memset(&rref, 0, sizeof(rref));
    rref.type = ptOctreeContent;
    rref.index = ref.index;
    if (pe == peSuccess) {
      pe = ptf_file_replace_record(fh, &rref);
    }
    if (pe == peSuccess) {
      pe = ptf_file_write_record(fh, data, (ptf_32)size);
    }
    if (pe == peSuccess) {
      pe = ptf_file_commit_record(fh, &rref);
    }
    return pe;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_append_to_file(ptf_content_handle cont, 
    ptf_file_handle fh, ptf_fixed_ref *o_ref)
{
  void * data = 0;
  size_t size = 0;
  ptf_err pe;
  try {
    if (!cont->validate()) return peInvalidArguments;
    pe = cont->make_data(&data, &size);
    if (pe == peSuccess) {
      ptf_32 cnt = 0;
      ptf_record_ref rref;
      if (pe == peSuccess) {
        pe = ptf_file_new_complete_record(fh, ptOctreeContent, data, (ptf_32)size, &rref);
        assert(rref.index > 0);
        o_ref->index = rref.index;
      }
    }
    return pe;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_get_annotation(ptf_content_handle cont, 
    void const **o_ptr, ptf_32 *o_size, ptf_type * o_type)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    *o_ptr = 0;
    *o_size = 0;
    *o_type = 0;
    if (cont->data.size()) {
      *o_ptr = &cont->data[0];
      *o_size = (ptf_32)cont->data.size();
      *o_type = cont->data_type;
    }
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_set_annotation(ptf_content_handle cont, 
    void const *ptr, ptf_32 size, ptf_type type)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    cont->data.resize(size);
    if (size > 0) {
      memcpy(&cont->data[0], ptr, size);
    }
    cont->data_type = type;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_count_references(ptf_content_handle cont, 
    ptf_32 *o_count)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    *o_count = (ptf_32)cont->references.size();
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_get_reference(ptf_content_handle cont, 
    ptf_32 index, ptf_record_ref *o_type)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    if (index >= cont->references.size()) {
      return peInvalidArguments;
    }
    *o_type = cont->references[index].ref;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_add_reference(ptf_content_handle cont, 
    ptf_record_ref const &ref, ptf_32 *o_index)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    //  no more than a million references in a single record
    // Ummm...you mean a hexmillion? -GRA
    if (cont->references.size() >= 0x100000) {
      return peTooBigForMachine;
    }
    content_reference cr;
    cr.ref = ref;
    cr.meta_data_ref.file = 0;
    cr.meta_data_ref.index = 0;
    cr.meta_data_ref.spare1 = 0;
    cr.meta_data_ref.type = ptNoType;
    cont->references.push_back(cr);
    *o_index = (ptf_32)cont->references.size() - 1;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_remove_reference(ptf_content_handle cont,
    ptf_32 index)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    if (index >= cont->references.size()) {
      return peInvalidArguments;
    }
    //  Don't copy construct all the references, just swap 
    //  with the hintermost reference and chop the vector, 
    //  because we don't promise anything about ordering.
    cont->references[index] = cont->references.back();
    cont->references.pop_back();
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_get_reference_meta_data(ptf_content_handle cont, 
    ptf_32 index, ptf_record_ref * o_ref)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    if (index >= cont->references.size()) {
      return peInvalidArguments;
    }
    *o_ref = cont->references[index].meta_data_ref;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_set_reference_meta_data(ptf_content_handle cont, 
    ptf_32 index, ptf_record_ref const & ref)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    if (index >= cont->references.size()) {
      return peInvalidArguments;
    }
    cont->references[index].meta_data_ref = ref;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_get_reference_annotation(ptf_content_handle cont, 
    ptf_32 index, void const **o_ptr, ptf_32 *o_size, ptf_32 * o_type)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    if (index >= cont->references.size()) {
      return peInvalidArguments;
    }
    if (cont->references[index].data.size()) {
      *o_ptr = &cont->references[index].data[0];
      *o_size = (ptf_32)cont->references[index].data.size();
      *o_type = cont->references[index].data_type;
      return peSuccess;
    }
    else {
      return peRecordNotFound;
    }
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_set_reference_annotation(ptf_content_handle cont, 
    ptf_32 index, void const *ptr, ptf_32 size, ptf_type type)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    if (index >= cont->references.size()) {
      return peInvalidArguments;
    }
    cont->references[index].data_type = type;
    cont->references[index].data.resize(size);
    if (size) {
      memcpy(&cont->references[index].data[0], ptr, size);
    }
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_count_reference_instances(ptf_content_handle cont, 
    ptf_32 index, ptf_32 *o_count)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    if (index >= cont->references.size()) {
      return peInvalidArguments;
    }
    *o_count = (ptf_32)cont->references[index].instances.size();
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_get_reference_instance(ptf_content_handle cont, 
    ptf_32 index, ptf_32 instance, ptf_geopos *o_pos, ptf_ori *o_ori)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    if (index >= cont->references.size()) {
      return peInvalidArguments;
    }
    if (instance >= cont->references[index].instances.size()) {
      return peInvalidArguments;
    }
    *o_pos = cont->references[index].instances[instance].pos;
    *o_ori = cont->references[index].instances[instance].ori;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_add_reference_instance(ptf_content_handle cont, 
    ptf_32 index, ptf_geopos const &pos, ptf_ori const &ori, ptf_32 *o_instance)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    if (index >= cont->references.size()) {
      return peInvalidArguments;
    }
    //  no more than a million instances in a single reference
    if (cont->references[index].instances.size() >= 0x100000) {
      return peTooBigForMachine;
    }
    reference_instance ri;
    ri.pos = pos;
    ri.ori = ori;
    *o_instance = (ptf_32)cont->references[index].instances.size();
    cont->references[index].instances.push_back(ri);
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_content_update_reference_instance(ptf_content_handle cont, 
    ptf_32 index, ptf_32 instance, ptf_geopos const &pos, ptf_ori const &ori)
{
  try {
    if (!cont->validate()) return peInvalidArguments;
    if (index >= cont->references.size()) {
      return peInvalidArguments;
    }
    if (instance >= cont->references[index].instances.size()) {
      return peInvalidArguments;
    }
    cont->references[index].instances[instance].pos = pos;
    cont->references[index].instances[instance].ori = ori;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

