#include "ptf.h"
#include "ptf_types.h"
#include "ptf_file.h"
#include "ptf_str.h"
#include "ptf_date.h"
#include "ptf_util.h"
#include "tiger.h"

using namespace impl;

//  This reference implementation of the PTF container format is designed to 
//  be transaction safe -- you can make edits to the container, and commit 
//  them, and any error on the way there will leave the container in a safe 
//  state. This is accomplished by appending all new data to the end of the 
//  container, and on flush, appending the update tables of content, flushing
//  the file (which at that point contains the old data, plus the new data), 
//  and then updating the meta-toc pointer, so it references the new TOC. 
//  At that point, the old data and TOCs become empty records, which can be 
//  re-used by some later operation, or removed by a compacting tool.


//  default alignment for files created new by this library
#define DEFAULT_ALIGNMENT 16
//  minimum allowed alignment
#define MIN_ALIGNMENT 8


#include <stdio.h>
#include <time.h>
#include <assert.h>
#include <string.h>

#include <vector>
#include <algorithm>

#ifdef PLATFORM_UNIX
  #define _wfopen fopen
#endif

struct toc_entry {
  toc_entry() {
    offset = 0;
    orig_size = 0;
    type = 0;
    dirty = false;
  }
  ptf_64 offset;
  ptf_32 orig_size;
  std::vector<ptf_toc_entry> toc;
  ptf_type type;
  bool dirty;
};

static ptf_32 ptfSig = (('p'<<24) | ('t'<<16) | ('f'<<8) | ' ');

struct ptf_file_t {
  public:
    ptf_file_t(ptf_file_accessors const * accessors);
    ~ptf_file_t();
    bool validate() { return sig_ == ptfSig; }
    ptf_err open(ptf_char const * path, ptf_open_mode mode, void* param);
    ptf_err close();
    ptf_err flush();
    ptf_err new_record(ptf_type type);
    ptf_err write_data(void const * data, ptf_32 size);
    ptf_err close_record(ptf_record_ref * o_ref);
    ptf_err get_toc(ptf_type type, std::vector<toc_entry>::iterator * o_toc);
    ptf_err get_or_make_toc(ptf_type type, std::vector<toc_entry>::iterator * o_toc);
    ptf_err append_toc(size_t at);
    ptf_err append_data(void const * data, ptf_64 size, ptf_64 * o_pos);
    ptf_err count_records(ptf_type type, ptf_32 * o_count);
    ptf_err read_record(ptf_type type, ptf_32 index, ptf_64 offset, void * buf, size_t size);
    ptf_err remove_record(ptf_type type, ptf_32 index);
    ptf_err get_record_size(ptf_type type, ptf_32 index, ptf_64 * o_size);
    ptf_err open_record(ptf_type type, ptf_32 index);
    ptf_err allocate_indices(ptf_type type, ptf_32 count, ptf_record_ref *o_ref);
    void compact_empty();
    ptf_err update_header(ptf_file_header const & hdr);

    ptf_file_accessors accessors_;
    ptf_32 sig_;
    bool update_;
    ptf_file_accessor_handle file_;
    ptf_file_header header_;
    zero_t<ptf_64> recPos_;
    zero_t<ptf_32> recSize_;
    zero_t<ptf_type> recType_;
    zero_t<ptf_32> editRecord_;

    tiger hasher_;

    std::vector<toc_entry> typeToc_;
    std::vector<ptf_char> path_;
    //  Octree handle, as used by the octree manager.
    void *oh_;
};



static void wpath_convert(wchar_t * path)
{
#if defined(_WIN32) || defined(_WIN64)
  //  convert /C/WINDOWS/SYSTEM into C:\WINDOWS\SYSTEM
  if (path[0] == '/' && path[1] != 0 && path[2] == '/') {
    path[0] = path[1];
    path[1] = ':';
  }
  while (*path) {
    if (*path == '/') {
      *path = '\\';
    }
    ++path;
  }
#endif
}

ptf_file_t::ptf_file_t(ptf_file_accessors const * accessors)
{
  sig_ = ptfSig;
  assert(sizeof(header_) == 256);
  file_ = 0;
  oh_ = 0;
  if(NULL == accessors) {
    memcpy( &accessors_, ptf_file_platform_accessors(), sizeof(ptf_file_accessors) );
  }
  else {
    memcpy( &accessors_, accessors, sizeof(ptf_file_accessors) );
  }
}

ptf_file_t::~ptf_file_t()
{
  compact_empty();
  flush();
  close();
  sig_ = 0;
}

ptf_err ptf_file_t::open(ptf_char const * path, ptf_open_mode mode, void* param)
{
  class close_file {
    public:
      close_file(ptf_file_accessor_handle * fpp, ptf_err (*close_func)(ptf_file_accessor_handle) ) : fpp_(fpp), close_func_( close_func ) {}
      ~close_file() { if (fpp_ && *fpp_) { close_func_(*fpp_); *fpp_ = 0; } }
      void reset() { fpp_ = 0; }
      ptf_file_accessor_handle * fpp_;
      ptf_err (*close_func_)(ptf_file_accessor_handle);
  };
  close_file noFile(&file_,accessors_.close);

  assert(!file_);
  memset(&header_, 0, sizeof(header_));
  recPos_ = 0;

  bool create = false;
  update_ = true;
  switch (mode) {
    case ptf_open_read: update_ = false; break;
    case ptf_open_update: break;
    case ptf_open_create: break; // needs special case
    case ptf_open_clobber: create = true; break;
    default: assert(!"bad ptf_open_mode");
      return peInvalidArguments;
  }
  file_ = accessors_.open(path, mode == ptf_open_create ? ptf_open_update : mode, param);

  if (!file_ && mode == ptf_open_create) {
    file_ = accessors_.open(path, mode, param);
    create = true;
  }

  if (!file_) {
    //TODO:  figure out why I couldn't open this file; could be permission
    return peFileNotFound;
  }
  else {
    //  verify it's a PTF file
    if (accessors_.seek(file_, 0, 0)) {
      return peIoError;
    }
    if (create) {
      //  Create new file -- fill out a basic header.
      ptf_str_from_char(PTF_MAGIC_CHARS, header_.magic, sizeof(header_.magic));
      if (ptf_str_basename(path, header_.file_name, sizeof(header_.file_name)) > sizeof(header_.file_name)) {
        return peInvalidArguments;
      }
      header_.spec_version = PTF_SPECIFICATION_VERSION;
      header_.min_version = PTF_MIN_VERSION;
      time_t t;
      time(&t);
      struct tm now = *gmtime(&t);
      ptf_date_from_tm(&now, &header_.cr_date);
      header_.alignment = DEFAULT_ALIGNMENT;
      header_.size = sizeof(header_);
      if (accessors_.write(&header_, sizeof(header_), file_) != peSuccess) {
        return peIoError;
      }
    }
    else {
      //  Open existing file -- verify that the header looks kosher.

      //  File must contain at least a header
      if (peSuccess != accessors_.read(&header_, sizeof(header_), file_)) {
        return peInvalidFile;
      }
      char buf[sizeof(header_.magic)];
      //  File must contain the right magic token
      if (8 != ptf_str_to_char(header_.magic, buf, 8, '?') || 
          strncmp(PTF_MAGIC_CHARS, buf, sizeof(buf))) {
        return peInvalidFile;
      }
      //  File must be readable by this version of the library
      if (header_.min_version > PTF_SPECIFICATION_VERSION) {
        return peInvalidFile;
      }
      //  Meta-TOC can't end beyond file size
      if (header_.meta_toc.offset+header_.meta_toc.size_disk > header_.size) {
        return peInvalidFile;
      }
      //  Alignment must be power of two
      if (header_.alignment & (header_.alignment-1)) {
        return peInvalidFile;
      }
      if (header_.alignment < MIN_ALIGNMENT) {
        return peInvalidFile;
      }
      //  Meta-TOC must be aligned (as must everything else)
      if (header_.meta_toc.offset & (header_.alignment-1)) {
        return peInvalidFile;
      }
      //  File size must be aligned (but record sizes don't need to)
      if (header_.size & (header_.alignment-1)) {
        return peInvalidFile;
      }
    }
  }

  //  Load the meta-toc and each type TOC
  if (header_.meta_toc.offset > 0 && header_.meta_toc.size_disk) {
    //  Read meta-TOC
    if (header_.meta_toc.offset + header_.meta_toc.size_disk > header_.size) {
      return peInvalidFile;
    }
    if (header_.meta_toc.offset & (header_.alignment-1)) {
      return peInvalidFile;
    }
    if (header_.meta_toc.compression || header_.meta_toc.encryption) {
      return peNotSupported;
    }
    if (header_.meta_toc.size_disk != header_.meta_toc.size_mem) {
      return peInvalidFile;
    }
    std::vector<ptf_meta_toc_entry> mte;
    if (accessors_.seek(file_, header_.meta_toc.offset, 0)) {
      return peIoError;
    }
    ptf_32 toRead = (ptf_32)header_.meta_toc.size_disk;
    if (toRead != (size_t)toRead) {
      return peTooBigForMachine;
    }
    mte.resize(header_.meta_toc.size_mem / sizeof(ptf_meta_toc_entry));
    ptf_err err = accessors_.read(&mte[0], header_.meta_toc.size_disk, file_);
    if (err != peSuccess) {
      return err;
    }

    // Check for valid hash
    {
      tiger hasher;
      hasher.hash_add_data( ( ptf_64 * )&mte[0], header_.meta_toc.size_disk );
      if( hasher.hash_complete() != header_.meta_toc.hash ) {
        return peInvalidFile;
      }
    }

    //  Read each type entry
    typeToc_.resize(mte.size());
    for (std::vector<ptf_meta_toc_entry>::iterator ptr = mte.begin(), end = mte.end();
        ptr != end; ++ptr) {
      if ((*ptr).offset & (header_.alignment - 1)) {
        return peInvalidFile;
      }
      if ((*ptr).encryption || (*ptr).compression) {
        return peNotSupported;
      }
      if ((*ptr).size_mem != (*ptr).size_disk) {
        return peInvalidFile;
      }
      toc_entry & te = typeToc_[ptr - mte.begin()];
      te.type = (*ptr).type;
      te.offset = (*ptr).offset;
      te.dirty = false;
      toRead = (ptf_32)(*ptr).size_disk;
      te.orig_size = (toRead + header_.alignment - 1) & ~(ptf_32)(header_.alignment - 1);
      if (toRead != (size_t)toRead) {
        return peTooBigForMachine;
      }
      te.toc.resize((*ptr).size_disk / sizeof(ptf_toc_entry));
      if ((*ptr).size_disk > 0) {
        if (accessors_.seek(file_, (*ptr).offset, 0)) {
          return peIoError;
        }
        err = accessors_.read(&te.toc[0], toRead, file_);
        if (err != peSuccess) {
          return err;
        }
        //  check for an invalid TOC entry
        {
          tiger hasher;
          hasher.hash_add_data( ( ptf_64 * )&te.toc[0], (*ptr).size_mem );
          ptf_64 storedHash, expectedHash;
          storedHash = (*ptr).hash;
          expectedHash = hasher.hash_complete();
          if( storedHash != expectedHash ) {
            return peInvalidFile;
          }
        }
        for (size_t n = te.toc.size(), i = 0; i != n; ++i) {
          if (te.toc[i].offset & (header_.alignment-1)) {
            return peInvalidFile;
          }
          if (!te.toc[i].offset && te.toc[i].size_disk) {
            return peInvalidFile;
          }
          if (te.toc[i].compression || te.toc[i].encryption) {
            return peNotSupported;
          }
          if (te.toc[i].size_disk != te.toc[i].size_mem) {
            return peInvalidFile;
          }
        }
      }
      else {
        //  if there's zero elements, there shouldn't be a toc offset
        if (te.offset) {
          return peInvalidFile;
        }
      }
    }
  }

  if (update_) {
    compact_empty();
  }
  
  //  remember how the file was opened
  path_.resize(strlen((char const *)path)+1);
  memcpy(&path_[0], path, path_.size());

  noFile.reset();
  return peSuccess;
}

ptf_err ptf_file_t::close()
{
  if (file_ != 0) {
    accessors_.close(file_);
    file_ = 0;
  }
  if (recPos_) {
    recPos_ = 0;
    return peInvalidSequence;
  }
  return peSuccess;
}

//  flush() is a little complex in order to implement proper 
//  transactional semantics. Either all changes are committed, or
//  no change is committed. This is mainly accomplished through 
//  re-locating the meta-TOC for each flush, as well as re-locating 
//  any type TOC that has changed. This means the old meta-TOC and 
//  type TOCs are still in the file, although marked as unused in a 
//  new empty record TOC. The single point of commit comes when the 
//  header is updated to point to the new meta-TOC instead of the 
//  old. If updating the file fails before that point through error 
//  or program termination, the header still points at the old, 
//  intact data, and there is some "random junk" (i e pending 
//  updates) after the end of the logically described file.
ptf_err ptf_file_t::flush()
{
  if (!file_) {
    return peInvalidSequence;
  }
  if (recPos_) {
    recPos_ = 0;
    return peInvalidSequence;
  }
  ptf_err ret = peSuccess;
  //  If read-only, then nothing can have changed
  if (update_) {
    std::vector<ptf_meta_toc_entry> mte;
    bool metaDirty = false;
    //  I may actually create the "empty records" type while doing this iteration
    for (size_t at = 0; at != typeToc_.size(); ++at) {
      std::vector<toc_entry>::iterator ptr = typeToc_.begin() + at;
      assert((*ptr).toc.size() <= 0xffffffffu);
      if ((*ptr).dirty) {
        metaDirty = true;
        //  re-write the EmptyRecord type last
        if ((*ptr).type != ptEmptyRecord) {
          ret = append_toc(ptr - typeToc_.begin());
          if (ret != peSuccess) {
            return ret;
          }
        }
        //  re-load the iterator after possible invalidation
        ptr = typeToc_.begin() + at;
      }
      ptf_meta_toc_entry pmte;
      memset(&pmte, 0, sizeof(pmte));
      pmte.type = (*ptr).type;
      pmte.offset = (*ptr).offset;
      pmte.size_mem = pmte.size_disk = (ptf_32)(*ptr).toc.size() * sizeof(ptf_toc_entry);
      tiger hasher;
      hasher.hash_add_data( (ptf_64*)&(*ptr).toc[0], pmte.size_mem );
      pmte.hash = hasher.hash_complete();
      mte.push_back(pmte);
    }
#if !defined(NDEBUG)
    for (size_t n = typeToc_.size(), i = 0; i != n; ++i) {
      //  ensure that every TOC entry with data has an offset
      assert(!typeToc_[i].toc.size() || typeToc_[i].offset || typeToc_[i].type == ptEmptyRecord);
    }
#endif
    if (metaDirty) {
      //  write the empty record TOC
      std::vector<toc_entry>::iterator toc;
      ret = get_or_make_toc(ptEmptyRecord, &toc);
      if (ret != peSuccess) {
        return ret;
      }
      //  Manually clear out the record range for the 
      //  current meta-TOC.
      if (header_.meta_toc.size_mem) {
        ptf_toc_entry eent;
        memset(&eent, 0, sizeof(eent));
        eent.offset = header_.meta_toc.offset;
        eent.size_mem = eent.size_disk = (header_.meta_toc.size_mem + header_.alignment - 1)
            & ~(ptf_32)(header_.alignment - 1);
        (*toc).toc.push_back(eent);
      }
      size_t at = toc - typeToc_.begin();
      ret = append_toc(at);
      if (ret != peSuccess) {
        //  undo the meta-TOC clearing operation
        if (header_.meta_toc.size_mem) {
          typeToc_[at].toc.pop_back();
        }
        return ret;
      }
      assert(mte.size() <= 0xffffffffU);
      //  so, now update the mte entry for the empty record type
      assert(mte[at].type == ptEmptyRecord);
      mte[at].offset = typeToc_[at].offset;
      mte[at].size_mem = mte[at].size_disk = (ptf_32)typeToc_[at].toc.size() * sizeof(ptf_toc_entry);
      //  write the meta-TOC
#if !defined(NDEBUG)
      for (size_t n = mte.size(), i = 0; i != n; ++i) {
        assert(!(mte[i].offset & (header_.alignment-1)));
        assert(!mte[i].offset == !mte[i].size_disk);
      }
#endif
      ptf_64 nmt;
      for (size_t i = 0; i < mte.size(); ++i) {
        //  right now, compression or encryption are not supported
        assert(!mte[i].compression && !mte[i].encryption);
        assert((mte[i].size_disk == mte[i].size_mem) || mte[i].compression);
        {
          tiger hasher;
          hasher.hash_add_data( (ptf_64*)&typeToc_[i].toc[0], mte[i].size_mem );
          mte[i].hash = hasher.hash_complete();
        }
      }
      ret = append_data(&mte[0], mte.size()*sizeof(ptf_meta_toc_entry), &nmt);
      if (ret != peSuccess) {
        return ret;
      }
      header_.meta_toc.offset = nmt;
      header_.meta_toc.size_mem = header_.meta_toc.size_disk = (ptf_32)mte.size() * sizeof(ptf_meta_toc_entry);
      header_.meta_toc.compression = header_.meta_toc.encryption = 0;

      {
        tiger hasher;
        hasher.hash_add_data( (ptf_64*)&mte[0], header_.meta_toc.size_mem );
        header_.meta_toc.hash = hasher.hash_complete();
      }
    }
#if !defined(NDEBUG)
    for (size_t n = typeToc_.size(), i = 0; i != n; ++i) {
      //  ensure that every TOC entry with data has an offset
      assert(!typeToc_[i].toc.size() || typeToc_[i].offset);
    }
#endif
    //  Pad the end of the file to a multiple of the page size
    if (accessors_.seek(file_, 0, 2)) {
      return peIoError;
      ptf_64 pos = accessors_.tell(file_);
      size_t toWrite = (size_t)(((pos + header_.alignment - 1) & ~(ptf_64)(header_.alignment - 1)) - pos);
      if (toWrite > 0) {
        //  zero pad up to page size
        void * data = calloc(toWrite, 1 );
        if (!data) {
          return peOutOfMemory;
        }
        ret = accessors_.write(data, toWrite, file_);
        if (ret != peSuccess) {
          return ret;
        }
      }
    }
    //  Flush the dirty data to disk, making sure that it's consistent 
    //  when the meta_toc pointer is changed to point to the new data.
    ret = accessors_.sync(file_);
    if (ret != peSuccess) {
      return ret;
    }
    //  Commit the transaction by writing the new header and flushing
    if (accessors_.seek(file_, 0, 0)) {
      return peIoError;
    }
    time_t t;
    time(&t);
    struct tm tm = *gmtime(&t);
    ptf_date_from_tm(&tm, &header_.mod_date);
    if (accessors_.write(&header_, sizeof(header_), file_) != peSuccess) {
      return peIoError;
    }
    ret = accessors_.sync(file_);
    if (ret != peSuccess) {
      return ret;
    }
    //  The TOC is now clean -- mark it as such
    for (std::vector<toc_entry>::iterator ptr = typeToc_.begin(), end = typeToc_.end();
        ptr != end; ++ptr) {
      (*ptr).dirty = false;
    }
  }
  return ret;
}

ptf_err ptf_file_t::new_record(ptf_type type)
{
  if (!update_) {
    return peReadOnly;
  }
  if (recPos_) {
    recPos_ = 0;
    return peInvalidSequence;
  }
  if (type == ptEmptyRecord) {
    //  I can't allow editing of the empty record contents
    return peInvalidArguments;
  }
  std::vector<toc_entry>::iterator ptr;
  ptf_err err = get_or_make_toc(type, &ptr);
  if (err != peSuccess) {
    return err;
  }
  if ((*ptr).toc.size() == 0xffffffffU) {
    //  can't create more records than that in the file format
    return peTooBigForMachine;
  }
  //  allocate space in the TOC (but make it empty)
  editRecord_ = (ptf_32)(*ptr).toc.size();
  ptf_toc_entry empty;
  memset(&empty, 0, sizeof(empty));
  hasher_.hash_init();
  (*ptr).toc.push_back(empty);

  //  remember that this is what's being edited
  recPos_ = header_.size;
  recSize_ = 0;
  recType_ = type;

  return peSuccess;
}

ptf_err ptf_file_t::write_data(void const * data, ptf_32 size)
{
  if (!recPos_) {
    return peInvalidSequence;
  }
  assert(update_);
  if (accessors_.seek(file_, recPos_+(ptf_64)recSize_, 0)) {
    recPos_ = 0;
    return peIoError;
  }
  ptf_err err = accessors_.write(data, size, file_);
  if (err == peSuccess) {
    recSize_ += size;
    hasher_.hash_add_data( ( ptf_64 * )data, size );
  }
  return err;
}

ptf_err ptf_file_t::close_record(ptf_record_ref * o_ref)
{
  if (!recPos_) {
    return peInvalidSequence;
  }
  ptf_64 pos = recPos_;
  recPos_ = 0;
  assert(update_);
  std::vector<toc_entry>::iterator ptr, etoc;
  ptf_err err = get_or_make_toc(ptEmptyRecord, &etoc);
  if (err != peSuccess) {
    return err;
  }
  err = get_toc((ptf_32)recType_, &ptr);
  if (err != peSuccess) {
    return err;
  }
  assert((*ptr).toc.size() > editRecord_);
  ptf_toc_entry pte;
  memset(&pte, 0, sizeof(pte));
  pte.offset = pos;
  pte.size_disk = pte.size_mem = (ptf_32)recSize_;
  pte.hash = hasher_.hash_complete();
  hasher_.hash_init();
  if ((*ptr).toc[editRecord_].size_disk) {
    (*etoc).toc.push_back((*ptr).toc[editRecord_]);
    (*etoc).dirty = true;
  }
  (*ptr).toc[editRecord_] = pte;
  (*ptr).dirty = true;
  //  Align file size to alignment boundary
  header_.size = (pos+recSize_+header_.alignment-1) & ~((ptf_64)header_.alignment-1);

  if ( o_ref ) {
    memset(o_ref, 0, sizeof(o_ref));
    o_ref->file = 0;
    o_ref->index = (ptf_32)editRecord_;
    o_ref->type = (ptf_32)recType_;
  }
  return peSuccess;
}

ptf_err ptf_file_t::get_toc(ptf_type type, std::vector<toc_entry>::iterator * o_toc)
{
  std::vector<toc_entry>::iterator ptr = typeToc_.begin(), end = typeToc_.end();
  while (ptr != end) {
    if ((*ptr).type == type) {
      *o_toc = ptr;
      return peSuccess;
    }
    ++ptr;
  }
  return peTypeNotFound;
}

ptf_err ptf_file_t::get_or_make_toc(ptf_type type, std::vector<toc_entry>::iterator * o_toc)
{
  if (recPos_ != 0) {
    recPos_ = 0;
    return peInvalidSequence;
  }
  std::vector<toc_entry>::iterator ptr = typeToc_.begin(), end = typeToc_.end(), ret = end;
  while (ptr != end) {
    if ((*ptr).type == type) {
      assert(ret == end);
      ret = ptr;
    }
    ++ptr;
  }
  if (ret == end) {
    ptf_64 siz = typeToc_.size();
    //  verify that we're not trying to add more than 0xffffffff types
    if ((ptf_64)(size_t)(siz+1) != siz+1) {
      return peTooBigForMachine;
    }
    toc_entry te;
    te.type = type;
    //  The zeroth TOC slot is always marked empty.
    te.toc.push_back(ptf_toc_entry());
    typeToc_.push_back(te);
    ret = --typeToc_.end();
  }
  else {
    ptf_64 siz = (*ret).toc.size();
    //  verify that we're not trying to add more than 0xffffffff items
    if ((ptf_64)(size_t)(siz+1) != siz+1) {
      return peTooBigForMachine;
    }
  }
  *o_toc = ret;
  return peSuccess;
}

ptf_err ptf_file_t::append_toc(size_t at)
{
  //  Create an annotation in the empty records table about the 
  //  location of the old TOC.
  std::vector<toc_entry>::iterator eiter;
  ptf_err err = get_or_make_toc(ptEmptyRecord, &eiter);
  if (err != peSuccess) {
    return err;
  }
  //  Get the toc iterator after allocating the empty record TOC
  std::vector<toc_entry>::iterator toc = typeToc_.begin() + at;
  ptf_toc_entry eent;
  memset(&eent, 0, sizeof(eent));
  eent.offset = (*toc).offset;
  eent.size_disk = eent.size_mem = (*toc).orig_size;
  if (eent.offset) {
    (*eiter).toc.push_back(eent);
  }

#if !defined(NDEBUG)
  for (size_t n = (*toc).toc.size(), i = 0; i !=n; ++i) {
    assert(!((*toc).toc[i].offset & (header_.alignment-1)));
    assert((*toc).toc[i].offset != 0 || (*toc).toc[i].size_mem == 0);
  }
#endif

  //  Write the TOC contents at the end of the file.
  ptf_64 offset = 0;
  ptf_32 size = (ptf_32)(*toc).toc.size() * (ptf_32)sizeof(ptf_toc_entry);
  if ((*toc).toc.size()) {
    for (size_t i = 0; i < (*toc).toc.size(); ++i) {
      assert(!(*toc).toc[i].compression && !(*toc).toc[i].encryption);
      assert(((*toc).toc[i].size_disk == (*toc).toc[i].size_mem) || (*toc).toc[i].compression);
    }
    err = append_data(&(*toc).toc[0], size, &offset);
    if (err != peSuccess) {
      if (eent.offset) {
        (*eiter).toc.pop_back();  //  undo annotation
      }
      return err;
    }
    assert(offset != 0);
  }
  assert(!(offset & (header_.alignment-1)));
  //  Remember where this TOC now lives.
  (*toc).offset = offset;
  (*toc).orig_size = (size + header_.alignment - 1) & ~(ptf_32)(header_.alignment - 1);

  assert(typeToc_[at].offset != 0 || !typeToc_[at].toc.size());

  return peSuccess;
}

ptf_err ptf_file_t::append_data(void const * data, ptf_64 size, ptf_64 * o_pos)
{
  assert(!(header_.size & (ptf_64)(header_.alignment - 1)));
  ptf_64 pos = header_.size;
  if (accessors_.seek(file_, header_.size, 0)) {
    return peIoError;
  }
  ptf_err err = accessors_.write(data, size, file_);
  if (err != peSuccess) {
    return err;
  }
  header_.size = (header_.size + size + header_.alignment - 1) & ~(ptf_64)(header_.alignment - 1);
  *o_pos = pos;
  return peSuccess;
}

ptf_err ptf_file_t::count_records(ptf_type type, ptf_32 * o_count)
{
  std::vector<toc_entry>::iterator toc;
  ptf_err ret = get_toc(type, &toc);
  if (ret != peSuccess) {
    return ret;
  }
  size_t size = (*toc).toc.size();
  //  make sure file hasn't grown beyond bounds
  assert((ptf_32)size == size);
  *o_count = (ptf_32)size;
  return peSuccess;
}

ptf_err ptf_file_t::read_record(ptf_type type, ptf_32 index, ptf_64 offset, void * buf, size_t size)
{
  std::vector<toc_entry>::iterator toc;
  ptf_err ret = get_toc(type, &toc);
  if (ret != peSuccess) {
    return ret;
  }
  if (index >= (*toc).toc.size()) {
    return peRecordNotFound;
  }
  ptf_toc_entry & te = (*toc).toc[index];
  if (te.offset == 0) {
    return peRecordNotFound;
  }
  if (offset + size > te.size_disk) {
    return peInvalidArguments;
  }
  if (te.offset+offset+size > header_.size) {
    return peInvalidFile;
  }
  if (accessors_.seek(file_, te.offset+offset, 0)) {
    return peIoError;
  }
  ret = accessors_.read(buf, size, file_);
  if (ret != peSuccess) {
    return ret;
  }

  // For now, always check the hash
  if( true ) {
    tiger hasher;
    unsigned char tempBuf[ 64 * 1024 ];
    ptf_64 verifyOffset = 0;

    while( ( verifyOffset + sizeof( tempBuf ) ) <= te.size_disk ) {
      accessors_.seek(file_, te.offset+verifyOffset, 0);
      ret = accessors_.read(tempBuf, sizeof( tempBuf ), file_);
      if (ret != peSuccess) {
        return ret;
      }
      hasher.hash_add_data( (ptf_64*)tempBuf, sizeof( tempBuf ) );
      verifyOffset += sizeof( tempBuf );
    }

    ptf_64 leftover = te.size_disk - verifyOffset;
    assert( leftover < sizeof( tempBuf ) );
    if( leftover ) {
      accessors_.seek(file_, te.offset+verifyOffset, 0);
      ret = accessors_.read(tempBuf, leftover, file_);
      if (ret != peSuccess) {
        return ret;
      }
      hasher.hash_add_data( (ptf_64*)tempBuf, leftover );
    }

    ptf_64 expectedHash = hasher.hash_complete();
    if ( te.hash !=  expectedHash ) {
      return peInvalidFile;
    }
  }
  return peSuccess;
}

ptf_err ptf_file_t::remove_record(ptf_type type, ptf_32 index)
{
  if (index == 0) {
    //  record 0 is always removed
    return peInvalidArguments;
  }
  //  Get the empty record toc before the other, as it may 
  //  allocate the type and invalidate other iterators.
  std::vector<toc_entry>::iterator toc, etoc;
  ptf_err ret = get_or_make_toc(ptEmptyRecord, &etoc);
  if (ret != peSuccess) {
    return ret;
  }
  //  Find the type to remove. get_toc() doesn't invalidate iterators.
  ret = get_toc(type, &toc);
  if (ret != peSuccess) {
    return ret;
  }
  if (index >= (*toc).toc.size()) {
    return peRecordNotFound;
  }
  //  Re-claim space only if there is actual space to re-claim.
  ptf_toc_entry & te = (*toc).toc[index];
  if (te.offset && te.size_mem) {
    //  remember the record location as an empty space
    ptf_toc_entry pte;
    memset(&pte, 0, sizeof(pte));
    pte.offset = te.offset;
    pte.size_disk = pte.size_mem = (te.size_mem + header_.alignment - 1) & ~(ptf_32)(header_.alignment - 1);
    (*etoc).toc.push_back(pte);
    //  mark as dirty
    (*etoc).dirty = true;
  }
  //  mark the record as deleted
  te.offset = 0;
  te.size_mem = te.size_disk = 0;
  (*toc).dirty = true;
  return peSuccess;
}

ptf_err ptf_file_t::get_record_size(ptf_type type, ptf_32 index, ptf_64 * o_size)
{
  std::vector<toc_entry>::iterator toc;
  //  Find the type to check.
  ptf_err ret = get_toc(type, &toc);
  if (ret != peSuccess) {
    return ret;
  }
  if (index >= (*toc).toc.size()) {
    return peRecordNotFound;
  }
  if ((*toc).toc[index].offset == 0) {
    return peRecordNotFound;
  }
  *o_size = (*toc).toc[index].size_disk;
  return peSuccess;
}

ptf_err ptf_file_t::open_record(ptf_type type, ptf_32 index)
{
  if (!update_) {
    return peReadOnly;
  }
  if (recPos_) {
    recPos_ = 0;
    return peInvalidSequence;
  }
  if (type == ptEmptyRecord) {
    //  I can't allow editing of the empty record contents
    return peInvalidArguments;
  }
  if (index == 0) {
    //  Record 0 is always invalid
    return peInvalidArguments;
  }
  std::vector<toc_entry>::iterator ptr;
  ptf_err err = get_or_make_toc(type, &ptr);
  if (err != peSuccess) {
    return err;
  }
  if (index >= (*ptr).toc.size()) {
    return peRecordNotFound;
  }
  editRecord_ = index;

  //  remember that this is what's being edited
  recPos_ = header_.size;
  recSize_ = 0;
  recType_ = type;

  return peSuccess;
}

ptf_err ptf_file_t::allocate_indices(ptf_type type, ptf_32 count, ptf_record_ref *o_ref)
{
  if (!update_) {
    return peReadOnly;
  }
  std::vector<toc_entry>::iterator ptr;
  ptf_err err = get_or_make_toc(type, &ptr);
  if (err != peSuccess) return err;
  memset(o_ref, 0, sizeof(*o_ref));
  o_ref->type = type;
  o_ref->index = (ptf_32)(*ptr).toc.size();
  if ((ptf_64)count + (ptf_64)(*ptr).toc.size() > 0xffffffffUL) {
    return peTooBigForMachine;
  }
  ptf_toc_entry pte;
  memset(&pte, 0, sizeof(pte));
  (*ptr).toc.insert((*ptr).toc.end(), count, pte);
  (*ptr).dirty = true;
  return peSuccess;
}

static int compare_toc(ptf_toc_entry const & ta, ptf_toc_entry const & tb)
{
  if (ta.offset < tb.offset) return true;
  return false;
}

//  Compact the empty list to avoid lots of little 
//  blocks that are contiguous.
void ptf_file_t::compact_empty()
{
  std::vector<toc_entry>::iterator ptr;
  ptf_err err = get_toc(ptEmptyRecord, &ptr);
  if (err != peSuccess) {
    //  probably a new file, or a file without empty ranges
    return;
  }
  if ((*ptr).toc.size() < 2) {
    //  nothing to do
    return;
  }
  //  Nobody will reference empty chunks by index, so it's OK to 
  //  re-order them.
  bool dirty = false;
  std::sort((*ptr).toc.begin(), (*ptr).toc.end(), compare_toc);
  for (size_t i = 0; i < (*ptr).toc.size()-1;) {
    if ((*ptr).toc[i].offset + (*ptr).toc[i].size_mem == (*ptr).toc[i+1].offset) {
      (*ptr).toc[i].size_mem += (*ptr).toc[i+1].size_mem;
      (*ptr).toc[i].size_disk += (*ptr).toc[i+1].size_disk;
      (*ptr).toc.erase((*ptr).toc.begin() + i + 1);
      dirty = true;
    }
    else {
      ++i;
    }
  }
  if (dirty) {
    (*ptr).dirty = true;
  }
}

ptf_err ptf_file_t::update_header(ptf_file_header const & hdr)
{
  if (hdr.range.x_min > hdr.range.x_max) return peInvalidArguments;
  if (hdr.range.y_min > hdr.range.y_max) return peInvalidArguments;
  if (hdr.range.z_min > hdr.range.z_max) return peInvalidArguments;
  if (!hdr.file_name[0]) return peInvalidArguments;
  memcpy(header_.file_name, hdr.file_name, sizeof(header_.file_name));
  header_.file_name[sizeof(header_.file_name)-1] = 0;
  memcpy(&header_.range, &hdr.range, sizeof(header_.range));
  memcpy(&header_.cr_date, &hdr.cr_date, sizeof(header_.cr_date));
  memcpy(&header_.mod_date, &hdr.mod_date, sizeof(header_.mod_date));
  memcpy(&header_.schema, &hdr.schema, sizeof(header_.schema));
  update_ = true;
  return peSuccess;
}







ptf_err ptf_file_open(ptf_char const * path, ptf_open_mode mode, ptf_file_handle * o_file)
{
  return ptf_file_open_accessors(path, mode, ptf_file_platform_accessors(), NULL, o_file);
}

ptf_err ptf_file_open_accessors(ptf_char const * path, ptf_open_mode mode, ptf_file_accessors const * accessors, void * param, ptf_file_handle * o_file)
{
  try {
    *o_file = 0;
    ptf_file_t * ft = new ptf_file_t(accessors);
    ptf_err ret = ft->open(path, mode, param);
    if (ret == peSuccess) {
      *o_file = ft;
    }
    else {
      delete ft;
    }
    return ret;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_close(ptf_file_handle file)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    ptf_err ret = file->flush();
    delete file;
    return ret;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_flush(ptf_file_handle file)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    return file->flush();
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_count_types(ptf_file_handle file, ptf_32 * o_count)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    *o_count = (ptf_32)file->typeToc_.size();
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_get_type_at(ptf_file_handle file, ptf_32 index, ptf_type * o_type)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    if (index >= file->typeToc_.size()) {
      return peInvalidArguments;
    }
    *o_type = file->typeToc_[index].type;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_count_type_records(ptf_file_handle file, ptf_type type, ptf_32 * o_count)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    return file->count_records(type, o_count);
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_get_record_size(ptf_file_handle file, ptf_record_ref const * ref, ptf_64 * o_size)
{
  if (ref->file != 0) {
    return peInvalidArguments;
  }
  try {
    if (!file->validate()) return peInvalidArguments;
    return file->get_record_size(ref->type, ref->index, o_size);
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_read_record(ptf_file_handle file, ptf_record_ref const * ref, ptf_64 offset, void * buf, ptf_64 size)
{
  if ((size_t)size != size) {
    return peTooBigForMachine;
  }
  if (ref->file != 0) {
    return peInvalidArguments;
  }
  try {
    if (!file->validate()) return peInvalidArguments;
    return file->read_record(ref->type, ref->index, offset, buf, (size_t)size);
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_new_record(ptf_file_handle file, ptf_type type)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    return file->new_record(type);
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_write_record(ptf_file_handle file, void const * data, ptf_32 size)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    return file->write_data(data, size);
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_commit_record(ptf_file_handle file, ptf_record_ref * o_ref)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    return file->close_record(o_ref);
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_file_new_complete_record(ptf_file_handle file, ptf_type type, void const * data, ptf_32 size, ptf_record_ref * o_ref)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    ptf_err err = ptf_file_new_record(file, type);
    if (err != peSuccess) return err;
    err = ptf_file_write_record(file, data, size);
    if (err != peSuccess) return err;
    err = ptf_file_commit_record(file, o_ref);
    return err;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_remove_record(ptf_file_handle file, ptf_record_ref const * ref)
{
  if (ref->file != 0) {
    return peInvalidArguments;
  }
  try {
    if (!file->validate()) return peInvalidArguments;
    return file->remove_record(ref->type, ref->index);
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_allocate_records(ptf_file_handle file, ptf_type type, ptf_32 count, ptf_record_ref *o_ref)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    if (type == ptEmptyRecord) return peInvalidArguments;
    if (count < 1) return peInvalidArguments;
    return file->allocate_indices(type, count, o_ref);
  }
  catch (...) {
    return peException;
  }
}

//  replace can be implemented transactionally
ptf_err ptf_file_replace_record(ptf_file_handle file, ptf_record_ref const * ref)
{
  if (ref->file != 0) {
    return peInvalidArguments;
  }
  try {
    if (!file->validate()) return peInvalidArguments;
    return file->open_record(ref->type, ref->index);
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_make_reference(ptf_file_handle file, ptf_file_definition_record const * def, ptf_32 * o_index)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    if( def->name[0] == '\0' ) {
      return peInvalidArguments;
    }
    ptf_file_definition_record fdr;
    ptf_record_ref rr;
    ptf_err ret = peSuccess;
    ptf_32 cnt = 0;
    ret = ptf_file_count_type_records(file, ptFileDefinition, &cnt);
    if (ret == peTypeNotFound) {
      ret = peSuccess;
      cnt = 0;
    }
    if (ret != peSuccess) {
      return ret;
    }
    for (ptf_32 i = 1; i < cnt; ++i) {
      rr.file = 0;
      rr.spare1 = 0;
      rr.type = ptFileDefinition;
      rr.index = i;
      ret = ptf_file_read_record(file, &rr, 0, &fdr, sizeof(fdr));
      if (ret == peRecordNotFound) {
        //  empty record
        continue;
      }
      if (ret != peSuccess) {
        return ret;
      }
      if (!ptf_strcasecmp(def->name, fdr.name)) {
        *o_index = i;
        return peSuccess;
      }
    }
    ret = ptf_file_new_complete_record(file, ptFileDefinition, def, sizeof(*def), &rr);
    if (ret == peSuccess) {
      *o_index = rr.index;
    }
    return ret;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_get_reference(ptf_file_handle file, ptf_32 index, ptf_file_definition_record * o_def)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    ptf_record_ref rr;
    rr.file = 0;
    rr.spare1 = 0;
    rr.type = ptFileDefinition;
    rr.index = index;
    ptf_err ret = ptf_file_read_record(file, &rr, 0, o_def, sizeof(*o_def));
    return ret;
  }
  catch (...) {
    return peException;
  }
}

static int sniff_ptf_file(ptf_char const * wpath, ptf_char const * canon, ptf_file_accessors const * accessors, void * param )
{
  // if wpath is empty string, return false
  char cpath[1025];
  if( ptf_str_to_char( wpath, cpath, 1024, '?' ) == 1 )
    return -1;

  //  check the path
  ptf_file_accessor_handle f = accessors->open(wpath, ptf_open_read, param);

  if (!f) return -1;
  //  read the header
  ptf_file_header hdr;
  memset(&hdr, 0, sizeof(hdr));
  if (accessors->read(&hdr, sizeof(hdr), f) != peSuccess) {
    accessors->close(f);
    return -1;
  }
  //  test for magic signature
  char buf[sizeof(hdr.magic)];
  //  File must contain the right magic token
  if (8 != ptf_str_to_char(hdr.magic, buf, 8, '?') || 
      strncmp(PTF_MAGIC_CHARS, buf, sizeof(buf))) {
    return peInvalidFile;
  }
  //  make sure a mal-formed header can't cause a buffer overrun
  memset(&hdr.file_name[sizeof(hdr.file_name)], 0, 6);
  if (ptf_strcasecmp(canon, hdr.file_name)) {
    accessors->close(f);
    return -1;
  }
  //  that's good enough for now -- we found our file!
  accessors->close(f);
  return 0;
}

static int sniff_ptf_path_file(ptf_char const * path, ptf_char const * file, ptf_file_definition_record * def, ptf_file_accessors const * accessors, void* param)
{
  wchar_t wname[PTF_PATHSIZE+1] = L"/";
  size_t wn = 0;
  if (file) {
    wn = ptf_str_to_wchar(file, &wname[1], PTF_PATHSIZE, '?');
    if (wn >= PTF_PATHSIZE) {
      return -1;  //  path won't fit
    }
    ++wn;
  }
  wchar_t wpath[PTF_PATHSIZE];
  size_t n = ptf_str_to_wchar(path, wpath, PTF_PATHSIZE, '?');
  if (n > PTF_PATHSIZE-wn) {
    //  path doesn't fit
    return -1;
  }
  memcpy(&wpath[n-1], wname, sizeof(wchar_t)*wn);
  wpath_convert(wpath);

  ptf_char ppath[PTF_PATHSIZE+1];
  ptf_str_from_wchar( wpath, ppath, PTF_PATHSIZE );

  if (sniff_ptf_file(ppath, def->name, accessors, param)) {
    //  not there, or wrong file
    return -1;
  }
  //  Update the definition record with what we found.
  if (ptf_str_from_wchar(wpath, def->fixed_path, sizeof(def->fixed_path)) > sizeof(def->fixed_path)) {
    return -1;
  }
  return 0;   //  success!
}

ptf_err ptf_reference_resolve_file(ptf_file_definition_record * def, ptf_char const * const * path)
{
  return ptf_reference_resolve_file_accessors(def, path, ptf_file_platform_accessors(), NULL);
}

ptf_err ptf_reference_resolve_file_accessors(ptf_file_definition_record * def, ptf_char const * const * path, ptf_file_accessors const * accessors, void* param)
{
  ptf_char const * const * search = path;
  // get the relative path as stored in the file definition record
  ptf_char * filename = def->rel_path;
  char cFilename[1025];
  if( ptf_str_to_char( filename, cFilename, 1024, '?' ) == 1 ) {
    // if relative path is not available, use the name
    filename = def->name;
  }

  //  first, try the current working directory
  if (!sniff_ptf_path_file((ptf_char const *)".", filename, def, accessors, param)) {
    return peSuccess;
  }

  //  next, try each of the search paths
  if (search) {
    while (*search) {
      if (!sniff_ptf_path_file(*search, filename, def, accessors, param)) {
        return peSuccess;
      }
      ++search;
    }
  }

  //  next, try the full path as stored in the record
  if (!sniff_ptf_path_file(def->fixed_path, 0, def, accessors, param)) {
    return peSuccess;
  }

  //  finally, give up
  return peFileNotFound;
}

ptf_err ptf_file_get_header(ptf_file_handle file, ptf_file_header * o_hdr)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    *o_hdr = file->header_;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_update_header(ptf_file_handle file, ptf_file_header const * i_hdr)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    return file->update_header(*i_hdr);
  }
  catch (...) {
    return peException;
  }
}

ptf_err platform_fread64(void * buffer, ptf_64 len, ptf_file_accessor_handle param)
{
  return fread64(buffer, len, (FILE*)param);
}

ptf_err platform_fseek64(ptf_file_accessor_handle param, ptf_64 offset, int whence)
{
  return (ptf_err)fseek64((FILE*)param, offset, whence);
}

ptf_err platform_fwrite64(void const * buffer, ptf_64 len, ptf_file_accessor_handle param)
{
  return fwrite64(buffer, len, (FILE*)param);
}

ptf_64 platform_ftell64(ptf_file_accessor_handle param)
{
  return ftell64((FILE*)param);
}

ptf_err platform_fsync64(ptf_file_accessor_handle param)
{
  return fsync64((FILE*)param);
}

ptf_file_accessor_handle platform_fopen(ptf_char const * path, ptf_open_mode mode, void* param)
{
  char const * cm = 0;
  switch (mode) {
    case ptf_open_read: cm = "rb"; break;
    case ptf_open_update: cm = "r+b"; break;
    case ptf_open_create: cm = "w+b"; break; 
    case ptf_open_clobber: cm = "w+b"; break;
    default: assert(!"bad ptf_open_mode");
      return 0;
  }
#ifdef PLATFORM_UNIX
  FILE* file = fopen((char*)path, (char*)cm);
#else
  wchar_t wpath[256];
  ptf_str_to_wchar(path, wpath, 256, '?');
  wpath_convert(wpath);

  wchar_t m[8];
  ptf_str_to_wchar((const ptf_char*)cm, m, 8, '?');

  FILE* file = _wfopen(wpath, m);
#endif
  return file;
}

ptf_err platform_fclose(ptf_file_accessor_handle param)
{
  return (ptf_err)fclose((FILE*)param);
}

ptf_file_accessors const * ptf_file_platform_accessors()
{
  static ptf_file_accessors defaultAccessors = {
    platform_fread64,
    platform_fseek64,
    platform_fwrite64,
    platform_ftell64,
    platform_fsync64,
    platform_fopen,
    platform_fclose };

  return &defaultAccessors;
}
ptf_err ptf_file_get_open_path(ptf_file_handle file, ptf_char *o_buf, size_t buf_size)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    size_t s = file->path_.size();
    if (s >= buf_size) return peOutOfMemory;
    memcpy(o_buf, &file->path_[0], s+1);
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_file_get_definition_record(ptf_file_handle file, ptf_file_definition_record *io_def)
{
  try {
    if (!file->validate()) return peInvalidArguments;
    memset(io_def, 0, sizeof(*io_def));
    ptf_err err = ptf_file_get_open_path(file, io_def->fixed_path, sizeof(io_def->fixed_path));
    if (err != peSuccess) return err;
    char const *sep = strrchr((char const *)io_def->fixed_path, '/');
#if defined(WIN32)
    if (!sep) sep = strrchr((char const *)io_def->fixed_path, '\\');
#endif
    if (!sep) sep = (char const *)io_def->fixed_path; else ++sep;
    size_t s = strlen(sep);
    if (s >= sizeof(io_def->name)) return peInvalidArguments;
    memcpy(io_def->name, sep, s+1);
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}
//  Secret function, used by ptf_octree.
void *ptf_file_get_octree_handle(ptf_file_handle file)
{
  if (!file->validate()) return 0;
  return file->oh_;
}

void ptf_file_set_octree_handle(ptf_file_handle file, void *oh)
{
  if (!file->validate()) return;
  file->oh_ = oh;
}

#if !defined(NDEBUG)
#include <assert.h>

class PathTester {
  public:
    PathTester() {
      wchar_t foo[100] = L"/C/WINDOWS/SYSTEM";
      wpath_convert(foo);

#if defined(_WIN32) || defined(_WIN64)
      assert(!wcscmp(foo, L"C:\\WINDOWS\\SYSTEM"));
#else
      assert(!wcscmp(foo, L"/C/WINDOWS/SYSTEM"));
#endif

      wcscpy(foo, L"./relative/path");
      wpath_convert(foo);
#if defined(_WIN32) || defined(_WIN64)
      assert(!wcscmp(foo, L".\\relative\\path"));
#else
      assert(!wcscmp(foo, L"./relative/path"));
#endif

      wcscpy(foo, L"../relative/path");
      wpath_convert(foo);
#if defined(_WIN32) || defined(_WIN64)
      assert(!wcscmp(foo, L"..\\relative\\path"));
#else
      assert(!wcscmp(foo, L"../relative/path"));
#endif

      wcscpy(foo, L"/");
      wpath_convert(foo);
#if defined(_WIN32) || defined(_WIN64)
      assert(!wcscmp(foo, L"\\"));
#else
      assert(!wcscmp(foo, L"/"));
#endif
    }
};
PathTester unitTestPath;

#endif

