
#include "ptf_material.h"
#include "ptf_str.h"
#include "ptf_util.h"
using namespace impl;

#include <string>
#include <vector>
#include <assert.h>

enum {
  ArchiveMaterialCountMax = 0xffff,
  MaterialParameterCountMax = 0xffff,
  MaterialParameterNameLengthMax = 0x7f,
  MaterialPropertyMaxDataSize = 0x7fff,
};

struct m_param_desc {
  char const *        name;
  ptf_material_type   type;
  bool                multiple;
  bool                optional;
};

m_param_desc mp_float =     { "float",          pmtFloatTuple,          };
m_param_desc mp_double =    { "double",         pmtDoubleTuple,         };
m_param_desc mp_string =    { "string",         pmtString,              };
m_param_desc mp_record =    { "record",         pmtRecordRef,           };
m_param_desc mp_uint32 =    { "uint32",         pmtUint32Tuple,         };
m_param_desc mp_explicit =  { "explicit_map",   pmtTextureMapExplicit,  };
m_param_desc mp_implicit =  { "implicit_map",   pmtTextureMapImplicit,  };
m_param_desc mp_gridded =   { "gridded_map",    pmtTextureMapGridded,   };
m_param_desc mp_map =       { "material_map",   pmtMaterialMap,         };
m_param_desc mp_material =  { "material",       pmtMaterial,            };
m_param_desc mp_archive =   { "archive",        pmtMaterialArchive,     };
m_param_desc mp_shadowmap = { "shadow_map",     pmtShadowMap,           };
m_param_desc mp_shader =    { "shader",         pmtShader,              };

m_param_desc mp_explicit_children[] = {
  { "texture_data",         pmtRecordRef,                               },
  { "texture_map_name",     pmtString,              false,  true,       },
  { "physical_scale_u",     pmtFloatTuple,                              },
  { "physical_scale_v",     pmtFloatTuple,                              },
};

m_param_desc mp_implicit_children[] = {
  { "texture_map",          pmtTextureMapExplicit,                      },
  { "reference_pos",        pmtDoubleTuple,                             },
  { "u_vector",             pmtFloatTuple,                              },
  { "v_vector",             pmtFloatTuple,                              },
  { "clamp_behavior",       pmtString,                                  },
};

m_param_desc mp_gridded_children[] = {
  { "texture_map",          pmtTextureMapImplicit,  true,               },
};

m_param_desc mp_map_children[] = {
  { "material_map",         pmtTextureMapImplicit,                      },
  { "texture_map",          pmtTextureMapExplicit,  true,               },
};

m_param_desc mp_any[] = {
  { "",                     pmtVariantAny,          true,               },
};

m_param_desc mp_materials[] = {
  { "",                     pmtMaterial,            true,               },
};

m_param_desc mp_shader_children[] = {
  { "shader_name",          pmtString,                                  },
  { "shader_properties",    pmtMaterial,                                },
};




struct ptf_material_parameter_t {
    ptf_material_parameter_t(
        ptf_material_parameter_t *i_parent, 
        m_param_desc const &i_you, 
        m_param_desc const *i_children, 
        size_t n_children,
        char const *i_name);
    virtual ~ptf_material_parameter_t();

    static const ptf_32 matSig = 'm'<<24 | 'a'<<16 | 't'<<8 | ' ';
    bool validate() { return sig == matSig; }
    ptf_material_parameter_t *find_parameter(ptf_char const *name);
    ptf_err remove_parameter(ptf_material_parameter_t *param);
    m_param_desc const *find_param_desc(ptf_char const *name);

    virtual ptf_err marshal_to(block_t & b);
    virtual ptf_err marshal_from(block_t & b, ptf_material_parameter_header const & hdr);

    ptf_16 find_prop_desc_index();

    ptf_32 sig;
    ptf_material_parameter_t *parent;
    m_param_desc me;
    m_param_desc const *child_descs;
    size_t n_child_descs;

    std::vector<ptf_material_parameter_t *> children;
    std::vector<char> data;
    std::string name;
};

struct ptf_material_t : public ptf_material_parameter_t {
    ptf_material_t(ptf_material_archive_t *i_parent, char const *i_name);

};

struct ptf_material_archive_t : public ptf_material_parameter_t {
    ptf_material_archive_t();
    ~ptf_material_archive_t();

    ptf_material_t * find_material(ptf_char const *name)
      {
        return static_cast<ptf_material_t *>(find_parameter(name));
      }
    ptf_err remove_material(ptf_material_t * mtl)
      {
        return remove_parameter(mtl);
      }

    virtual ptf_err marshal_to(block_t & b);
    //  this hides the inherited function!
    ptf_err marshal_from(block_t & b, ptf_16 * o_version);
};


struct material_type_parameter {
  ptf_material_type type;
  m_param_desc const & param;
  m_param_desc const * children;
  size_t n_children;
};
material_type_parameter parameter_types[] = {
  { pmtFloatTuple,          mp_float,       0,                      0, },
  { pmtDoubleTuple,         mp_double,      0,                      0, },
  { pmtString,              mp_string,      0,                      0, },
  { pmtRecordRef,           mp_record,      0,                      0, },
  { pmtUint32Tuple,         mp_uint32,      0,                      0, },
  { pmtTextureMapExplicit,  mp_explicit,    mp_explicit_children,   sizeof(mp_explicit_children)/sizeof(mp_explicit_children[0]), },
  { pmtTextureMapImplicit,  mp_implicit,    mp_implicit_children,   sizeof(mp_implicit_children)/sizeof(mp_implicit_children[0]), },
  { pmtTextureMapGridded,   mp_gridded,     mp_gridded_children,    sizeof(mp_gridded_children)/sizeof(mp_gridded_children[0]), },
  { pmtMaterialMap,         mp_map,         mp_map_children,        sizeof(mp_map_children)/sizeof(mp_map_children[0]), },
  { pmtMaterial,            mp_material,    mp_any,                 sizeof(mp_any)/sizeof(mp_any[0]), },
  { pmtMaterialArchive,     mp_archive,     mp_materials,           sizeof(mp_materials)/sizeof(mp_materials[0]), },
  { pmtShadowMap,           mp_shadowmap,   mp_explicit_children,   sizeof(mp_explicit_children)/sizeof(mp_explicit_children[0]), },
  { pmtShader,              mp_shader,      mp_shader_children,     sizeof(mp_shader_children)/sizeof(mp_shader_children[0]), },
};

material_type_parameter const * find_parameter(ptf_material_type i_type)
{
  for (size_t i = 0; i < sizeof(parameter_types)/sizeof(parameter_types[0]); ++i) {
    if (parameter_types[i].type == i_type) {
      return &parameter_types[i];
    }
  }
  return 0;
}

ptf_material_parameter_t *new_parameter(ptf_material_type i_type, ptf_material_parameter_t *i_parent, 
    char const *i_name)
{
  material_type_parameter const * mtp = find_parameter(i_type);
  if (mtp) {
    return new ptf_material_parameter_t(i_parent, mtp->param, mtp->children, mtp->n_children, i_name);
  }
  assert(!"bad parameter type in new_parameter()");
  return 0;
}

ptf_material_parameter_t *new_parameter_type(ptf_material_parameter_t * i_parent, m_param_desc const & me, 
    m_param_desc const * children, size_t child_count)
{
  return new ptf_material_parameter_t(i_parent, me, children, child_count, me.name);
}


ptf_material_parameter_t::ptf_material_parameter_t(
    ptf_material_parameter_t *i_parent, 
    m_param_desc const &i_you, 
    m_param_desc const *i_children, 
    size_t n_children,
    char const *i_name)
  : parent(i_parent)
  , me(i_you)
  , child_descs(i_children)
  , n_child_descs(n_children)
  , name(i_name)
{
  assert(n_children < 0x100);
  assert(me.type >= pmtVariantMap || n_children == 0);
  //  pre-create the required child parameters, although they will be empty
  for (size_t i = 0; i < n_children; ++i) {
    if (!i_children[i].multiple && !i_children[i].optional) {
      children.push_back(new_parameter(i_children[i].type, this, i_children[i].name));
    }
  }
  sig = matSig;
}

ptf_material_parameter_t::~ptf_material_parameter_t()
{
  for (size_t i = 0; i < children.size(); ++i) {
    delete children[i];
  }
  sig = 0;
}

ptf_material_parameter_t *ptf_material_parameter_t::find_parameter(ptf_char const *name)
{
  for (size_t i = 0; i < children.size(); ++i) {
    if (children[i] && !ptf_strcasecmp(name, (ptf_char const *)children[i]->name.c_str())) {
      return children[i];
    }
  }
  return 0;
}

ptf_err ptf_material_parameter_t::remove_parameter(ptf_material_parameter_t *param)
{
  for (size_t i = 0; i < children.size(); ++i) {
    if (children[i] == param) {
      children[i] = 0;
      delete param;
      if(me.type != pmtMaterialArchive) {
        children.erase( children.begin()+i );
      }
      return peSuccess;
    }
  }
  return peInvalidArguments;
}

m_param_desc const *ptf_material_parameter_t::find_param_desc(ptf_char const *name)
{
  for (size_t i = 0; i < n_child_descs; ++i) {
    if (!ptf_strcasecmp(name, (ptf_char const *)child_descs[i].name)) {
      return &child_descs[i];
    }
  }
  for (size_t i = 0; i < children.size(); ++i) {
    if (!ptf_strcasecmp(name, (ptf_char const *)children[i]->name.c_str())) {
      return &children[i]->me;
    }
  }
  return 0;
}


ptf_err ptf_material_parameter_t::marshal_to(block_t & b)
{
  if (children.size()) {
    if (data.size() > 0) {
      return peInvalidFile;
    }
  }
  else {
    // TODO: Implement writing larger chunks of data
    if (data.size() > MaterialPropertyMaxDataSize) {
      return peInvalidFile;
    }
  }
  ptf_material_parameter_header hdr;
  memset(&hdr, 0, sizeof(hdr));
  hdr.type = me.type;
  hdr.child_count = (ptf_16)children.size();
  hdr.prop_desc = find_prop_desc_index();
  if (!b.put(hdr)) return peOutOfMemory;
  if (hdr.prop_desc >= 0x100) {
    if (!b.put((ptf_char const *)name.c_str())) return peOutOfMemory;
  }
  if (!hdr.child_count) {
    ptf_32 size = (ptf_32)data.size();
    ptf_16 s;
    ptf_16 orval = 0;
    if( size ) {
      while( size ) {
        s = ( size >> 15 ) & 0x7fff;
        size <<= 15;
        if( s ) {
          s |= orval;
          if (!b.put(s)) return peInvalidFile;
          orval = 0x8000;
        }
      }
      if (!b.put_bytes(&data[0], data.size())) return peOutOfMemory;
    }
    else {
      // Write out zero-size data if that's the case
      s = 0;
      if (!b.put(s)) return peInvalidFile;
    }
  }
  for (size_t i = 0; i < children.size(); ++i) {
    if((me.type == pmtMaterialArchive) && !children[i]) {
      bool valid = false;
      if (!b.put(valid)) return peOutOfMemory;
    }
    else {
      bool valid = true;
      if (!b.put(valid)) return peOutOfMemory;
      ptf_err e = children[i]->marshal_to(b);
      if (e != peSuccess) return e;
    }
  }
  return peSuccess;
}

ptf_err ptf_material_parameter_t::marshal_from(block_t & b, ptf_material_parameter_header const & hdr)
{
  if (!hdr.child_count) {
    //  decode a data block
    ptf_16 s;
    ptf_32 size = 0;
    do {
      if (!b.get(s)) return peInvalidFile;
      size = (size << 15) | (s & 0x7fff);
    } while (s & 0x8000);
    if (s > MaterialPropertyMaxDataSize) {
      // TODO: Implement reading larger chunks of data
      //  Skip a block that's larger than what we know how to deal with.
      //  This will leave an empty property in the property tree, which 
      //  will lead to data loss if it's saved out again. It's not recommended 
      //  to save out data stored in files of versions later than what 
      //  the current library is.
      if (!b.seek(size, 1)) return peInvalidFile;
    }
    else if (size > 0) {
      size_t pos = b.pos();
      char const * base = (char const *)b.data();
      data.resize(size);
      if (!b.get_bytes(&data[0], size)) return peInvalidFile;
    }
  }
  else {
    //  decode some number of children
    for (size_t i = 0; i < hdr.child_count; ++i) {
      //  this ought to be shared with material archive
      bool valid = false;
      if (!b.get(valid)) return peInvalidFile;
      if (!valid) continue;
      ptf_material_parameter_header hdr;
      if (!b.get(hdr)) return peInvalidFile;
      ptf_char name[256];
      bool unknown = true;
      if (hdr.prop_desc >= 0x100) {
        //  skip extension data
        if (!b.seek(hdr.prop_desc & 0xff, 1)) return peInvalidFile;
        //  read name
        if (!b.get(name, 256)) return peInvalidFile;
      }
      else {
        //  find name based on index
        if (hdr.prop_desc >= n_child_descs || !child_descs[hdr.prop_desc].name[0]) {
          sprintf((char *)name, "param_%d_type_%d", (int)children.size(), hdr.prop_desc);
        }
        else {
          ptf_strncpy(name, 256, (ptf_char const *)child_descs[hdr.prop_desc].name);
          if (hdr.type != child_descs[hdr.prop_desc].type 
              && child_descs[hdr.prop_desc].type != pmtVariantAny) {
            return peInvalidFile;
          }
          unknown = false;
        }
      }
      ptf_material_parameter_t * nu_param = find_parameter(name);
      if (nu_param == 0) {
        if (::find_parameter((ptf_material_type)hdr.type) != 0) {
          unknown = false;
        }
      }
      if (nu_param == 0 || nu_param->me.multiple) {
        if (unknown) {
          //  I'll have to manufacture the type desc myself
          m_param_desc desc;
          desc.multiple = (hdr.child_count || (hdr.prop_desc & pmtVariantMap));
          desc.name = (char const *)name;
          desc.type = (ptf_material_type)hdr.type;
          nu_param = new_parameter_type(this, desc, desc.multiple ? mp_any : 0, 
              desc.multiple ? sizeof(mp_any)/sizeof(mp_any[0]) : 0);
        }
        else {
          nu_param = new_parameter((ptf_material_type)hdr.type, this, (char const *)name);
        }
        if(nu_param && hdr.prop_desc < 0x100) {
          nu_param->me.multiple |= child_descs[hdr.prop_desc].multiple;
        }
        children.push_back(nu_param);
      }
      //  else I'm just re-creating a standard child parameter
      ptf_err e = nu_param->marshal_from(b, hdr);
      if (e != peSuccess) return e;
    }
  }
  return peSuccess;
}

ptf_16 ptf_material_parameter_t::find_prop_desc_index()
{
  if (!parent) return 0x100;  //  can't be a desc based property
  for (size_t i = 0; i < parent->n_child_descs; ++i) {
    if (!ptf_strcasecmp((ptf_char const *)name.c_str(), 
        (ptf_char const *)parent->child_descs[i].name)) {
      //  All known parameters must be named
      assert(parent->child_descs[i].name[0]);
      return (ptf_16)i;
    }
  }
  return 0x100;
}


ptf_material_t::ptf_material_t(ptf_material_archive_t *i_parent, char const *i_name)
  : ptf_material_parameter_t(i_parent, mp_material, mp_any, sizeof(mp_any)/sizeof(mp_any[0]), i_name)
{
}


ptf_material_archive_t::ptf_material_archive_t()
  : ptf_material_parameter_t(0, mp_archive, mp_materials, sizeof(mp_materials)/sizeof(mp_materials[0]), "archive")
{
}

ptf_material_archive_t::~ptf_material_archive_t()
{
}

ptf_err ptf_material_archive_t::marshal_to(block_t & b)
{
  ptf_material_archive_header hdr;
  memset(&hdr, 0, sizeof(hdr));
  hdr.version_written = 1;
  hdr.version_to_read = 1;
  if (!b.put(hdr)) return peOutOfMemory;
  return ptf_material_parameter_t::marshal_to(b);
}

ptf_err ptf_material_archive_t::marshal_from(block_t & b, ptf_16 * o_version)
{
  ptf_material_archive_header hdr;
  memset(&hdr, 0, sizeof(hdr));
  if (!b.get(hdr)) return peInvalidFile;
  *o_version = hdr.version_written;
  if (hdr.version_written == 0 || hdr.version_to_read > PTF_CURRENT_MATERIAL_VERSION) {
    return peInvalidVersion;
  }
  //  skip extension data, if present
  if (!b.seek(hdr.extension_size, 1)) return peIoError;
  ptf_material_parameter_header phdr;
  if (!b.get(phdr)) return peInvalidFile;
  if (!phdr.child_count || phdr.type != pmtMaterialArchive || phdr.prop_desc < 0x100) {
    return peInvalidFile;
  }
  char data[256];
  if (!b.seek((phdr.prop_desc & 0xff), 1)) return peInvalidFile;
  if (!b.get((ptf_char *)data, 256)) return peInvalidFile;
  name = data;
  return ptf_material_parameter_t::marshal_from(b, phdr);
}




ptf_err ptf_material_archive_create(ptf_material_archive *o_arc)
{
  try {
    *o_arc = new ptf_material_archive_t();
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

ptf_err ptf_material_archive_destroy(ptf_material_archive arc)
{
  try {
    if (!arc->validate()) return peInvalidArguments;
    delete arc;
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

ptf_err ptf_material_archive_append_to_file(ptf_material_archive arc,
    ptf_file_handle file, ptf_fixed_ref *o_ref)
{
  try {
    if (!arc->validate()) return peInvalidArguments;
    block_t b;
    ptf_err e = arc->marshal_to(b);
    if (e != peSuccess) return e;
    ptf_record_ref rref;
    e = ptf_file_new_complete_record(file, ptMaterialArchive, b.data(),
        (ptf_32)b.size(), &rref);
    o_ref->index = rref.index;
    return e;
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_material_archive_update_to_file(ptf_material_archive arc,
    ptf_file_handle file, ptf_fixed_ref const &ref)
{
  try {
    if (!arc->validate()) return peInvalidArguments;
    block_t b;
    ptf_err e = arc->marshal_to(b);
    if (e != peSuccess) return e;
    ptf_record_ref rref;
    memset(&rref, 0, sizeof(rref));
    rref.type = ptMaterialArchive;
    rref.index = ref.index;
    e = ptf_file_replace_record(file, &rref);
    if (e != peSuccess) return e;
    e = ptf_file_write_record(file, b.data(), (ptf_32)b.size());
    if (e != peSuccess) return e;
    e = ptf_file_commit_record(file, &rref);
    return e;
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_material_archive_read_from_file(ptf_material_archive arc,
    ptf_file_handle file, ptf_fixed_ref const &ref, ptf_16 * o_version)
{
  try {
    if (!arc->validate()) return peInvalidArguments;
    if (arc->me.type != pmtMaterialArchive) return peInvalidArguments;
    *o_version = 0;
    ptf_record_ref rref;
    memset(&rref, 0, sizeof(rref));
    rref.type = ptMaterialArchive;
    rref.index = ref.index;
    ptf_64 size;
    ptf_err e = ptf_file_get_record_size(file, &rref, &size);
    if (e != peSuccess) return e;
    if (size > 0x1000000) {
      return peTooBigForMachine;
    }
    block_t b(size);
    void * base = b.extend((ptf_32)size);
    if (!base) return peOutOfMemory;
    e = ptf_file_read_record(file, &rref, 0, base, (ptf_32)size);
    if (e != peSuccess) return e;
    if (!b.seek(0, 0)) return peIoError;
    return arc->marshal_from(b, o_version);
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}


ptf_err ptf_material_archive_create_material(ptf_material_archive arc, 
    ptf_char const *name, ptf_material_handle *o_mat)
{
  try {
    if (!arc->validate()) return peInvalidArguments;
    if (arc->me.type != pmtMaterialArchive) return peInvalidArguments;
    if (arc->find_material(name) != 0) {
      return peDuplicateName;
    }
    if (arc->children.size() > ArchiveMaterialCountMax) {
      return peTooBigForMachine;
    }
    *o_mat = new ptf_material_t(arc, (char const *)name);
    arc->children.push_back(*o_mat);
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}


ptf_err ptf_material_remove(ptf_material_handle mtl)
{
  try {
    if (!mtl->validate()) return peInvalidArguments;
    if (mtl->me.type != pmtMaterial) return peInvalidArguments;
    ptf_err e = static_cast<ptf_material_archive_t *>(mtl->parent)->remove_parameter(mtl);
    return e;
  }
  catch (...) {
    return peException;
  }
}


ptf_err ptf_material_archive_find_material(ptf_material_archive arc,
    ptf_char const *name, ptf_material_handle *o_mat)
{
  try {
    if (!arc->validate()) return peInvalidArguments;
    if (arc->me.type != pmtMaterialArchive) return peInvalidArguments;
    *o_mat = static_cast<ptf_material_t *>(arc->find_parameter(name));
    if (!*o_mat) return peNameNotFound;
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}


ptf_err ptf_material_archive_get_material(ptf_material_archive arc,
    ptf_32 index, ptf_material_handle *o_mat)
{
  try {
    if (!arc->validate()) return peInvalidArguments;
    if (arc->me.type != pmtMaterialArchive) return peInvalidArguments;
    if (index >= arc->children.size()) return peInvalidArguments;
    *o_mat = static_cast<ptf_material_t *>(arc->children[index]);
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}


ptf_err ptf_material_archive_count_materials(ptf_material_archive arc,
    ptf_32 * o_count)
{
  if (!arc->validate()) return peInvalidArguments;
  if (arc->me.type != pmtMaterialArchive) return peInvalidArguments;
  return ptf_parameter_count_parameters(arc, o_count);
}

ptf_err ptf_material_get_name(ptf_material_handle mat,
    ptf_char const **m_name)
{
  if (!mat->validate()) return peInvalidArguments;
  if (mat->me.type != pmtMaterial) return peInvalidArguments;
  try {
    *m_name = (ptf_char const *)mat->name.c_str();
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

ptf_err ptf_material_count_parameters(ptf_material_handle mat,
    ptf_32 * o_count)
{
  if (!mat->validate()) return peInvalidArguments;
  if (mat->me.type != pmtMaterial) return peInvalidArguments;
  return ptf_parameter_count_parameters(mat, o_count);
}


ptf_err ptf_material_get_parameter(ptf_material_handle mat,
    ptf_32 index, ptf_material_parameter *o_param)
{
  if (!mat->validate()) return peInvalidArguments;
  if (mat->me.type != pmtMaterial) return peInvalidArguments;
  return ptf_parameter_get_parameter(mat, index, o_param);
}

ptf_err ptf_parameter_find_parameter(ptf_material_parameter param,
    ptf_char const *i_name, ptf_material_parameter *o_param)
{
  *o_param = 0;
  try {
    if (!param->validate()) return peInvalidArguments;
    for (size_t i = 0; i < param->children.size(); ++i) {

      ptf_char *childName = (ptf_char *)param->children[i]->name.c_str();

      if (!ptf_strcasecmp(i_name, (ptf_char const *)param->children[i]->name.c_str())) {
        *o_param = param->children[i];
        return peSuccess;
      }
    }
  }
  catch (...) {
    return peException;
  }
  return peNameNotFound;
}

ptf_err ptf_material_find_parameter(ptf_material_handle mat,
    ptf_char const *name, ptf_material_parameter *o_param)
{
  if (!mat->validate()) return peInvalidArguments;
  if (mat->me.type != pmtMaterial) return peInvalidArguments;
  return ptf_parameter_find_parameter(mat, name, o_param);
}

ptf_err ptf_material_add_parameter(ptf_material_handle mat,
    ptf_char const *name, ptf_material_type type, ptf_material_parameter *o_param)
{
  if (!mat->validate()) return peInvalidArguments;
  if (mat->me.type != pmtMaterial) return peInvalidArguments;
  return ptf_parameter_add_parameter(mat, name, type, o_param);
}

ptf_err ptf_parameter_add_parameter(ptf_material_parameter param,
    ptf_char const *name, ptf_material_type type, ptf_material_parameter *o_param)
{
  try {
    if (!param->validate()) return peInvalidArguments;
    if (param->data.size()) {
      return peInvalidSequence;
    }
    if (!name[0] || strlen((char const *)name) > MaterialParameterNameLengthMax) {
      return peInvalidArguments;
    }
    if (param->children.size() >= MaterialParameterCountMax) {
      return peTooBigForMachine;
    }
    ptf_material_parameter_t *mp = param->find_parameter(name);
    if (mp && !mp->me.multiple) {
      //  Because I pre-create all known children that are not 
      //  multiple-instance, I have to return known empty children 
      //  here, rather than disallowing their creation.
      if (mp->children.size() == 0 && mp->data.size() == 0) {
        *o_param = mp;
        return peSuccess;
      }
      return peDuplicateName;
    }
    m_param_desc const *desc = param->find_param_desc(name);
    bool has_any = false;
    for (size_t i = 0; i < param->n_child_descs; ++i) {
      if (param->child_descs[i].type == pmtVariantAny &&
          !param->child_descs[i].name[0]) {
        has_any = true;
        break;
      }
    }
    if (desc != 0 && type != desc->type) {
      return peInvalidArguments;
    }
    else if (desc == 0 && !has_any) {
      //  todo: at some point, I should allow adding arbitrary parameters to certain specific types.
      return peInvalidArguments;
    }
    else {
      *o_param = new_parameter(type, param, (char const *)name);
      if(desc) {
        (*o_param)->me.multiple |= desc->multiple;
      }
      param->children.push_back(*o_param);
    }
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}


ptf_err ptf_parameter_remove(ptf_material_parameter param)
{
  try {
    if (!param->validate()) return peInvalidArguments;
    ptf_err e = param->parent->remove_parameter(param);
    return e;
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}


ptf_err ptf_parameter_get_value(ptf_material_parameter param,
    void *data, size_t *ioSize)
{
  try {
    if (!param->validate()) return peInvalidArguments;
    size_t s = *ioSize;
    *ioSize = param->data.size();
    if (s < *ioSize) {
      return peOutOfMemory;
    }
    if (*ioSize) {
      memcpy(data, &param->data[0], *ioSize);
    }
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}


ptf_err ptf_parameter_set_value(ptf_material_parameter param,
    void const *data, size_t size)
{
  try {
    if (!param->validate()) return peInvalidArguments;
    if (param->children.size()) {
      return peInvalidSequence;
    }
    param->data.resize(size);
    if (size) {
      memcpy(&param->data[0], data, size);
    }
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}


ptf_err ptf_parameter_count_parameters(ptf_material_parameter param,
    ptf_32 *o_count)
{
  try {
    if (!param->validate()) return peInvalidArguments;
    if(param->me.type != pmtMaterialArchive) {
      *o_count = (ptf_32)param->children.size();
    }
    else {
      ptf_32 count = 0;
      for (size_t i = 0; i < param->children.size(); ++i) {
        if (param->children[i]) {
          ++count;
        }
      }
      *o_count = count;
    }
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}


ptf_err ptf_parameter_get_parameter(ptf_material_parameter param,
    ptf_32 index, ptf_material_parameter *o_param)
{
  try {
    if (!param->validate()) return peInvalidArguments;
    if (index >= param->children.size()) {
      return peInvalidArguments;
    }
    *o_param = param->children[index];
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}


ptf_err ptf_parameter_get_info(ptf_material_parameter param,
    ptf_char const **o_name, ptf_material_type *o_type, size_t *o_size)
{
  try {
    if (!param->validate()) return peInvalidArguments;
    *o_name = (ptf_char const *)param->name.c_str();
    *o_type = param->me.type;
    *o_size = param->data.size();
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

