
#include "ptf_texture.h"

#include <string.h>

#include <vector>
#include <string>


struct ptf_texture_t {
  ptf_texture_t() { memset(&header, 0, sizeof(header)); sig = 'tex '; }
  ~ptf_texture_t() { clear(); sig = 0; }
  void clear();
  bool validate() { return sig == 'tex '; }

  ptf_32 sig;
  ptf_texture_data header;
  //  This CAN'T be good...
  std::vector<std::vector<std::vector<char> > > data;
};

void ptf_texture_t::clear()
{
  std::vector<std::vector<std::vector<char> > >().swap(data);
}

static ptf_err ptf_file_write_pad(ptf_file_handle file, size_t amount)
{
  static char buf[256];
  while (amount > 0) {
    size_t towrite = amount;
    if (towrite > 256) {
      towrite = 256;
    }
    ptf_err e = ptf_file_write_record(file, buf, (ptf_32)towrite);
    if (e != peSuccess) {
      return e;
    }
    amount -= towrite;
  }
  return peSuccess;
}

static ptf_err validate_header(ptf_texture_data const *header)
{
  if (header->alignment == 0) {
    return peInvalidArguments;
  }
  if (header->border > 1) {
    return peInvalidArguments;
  }
  if (header->depth > 0x8000) {
    return peTooBigForMachine;
  }
  if (header->dimensions == 0 || header->dimensions > 3) {
    return peInvalidArguments;
  }
  if (header->format == ptdNull) {
    return peInvalidArguments;
  }
  if (header->height > 0x8000) {
    return peTooBigForMachine;
  }
  int maxmip = 0;
  ptf_32 val = header->height | header->width | header->depth;
  while (val > 0) {
    val = val >> 1;
    ++maxmip;
  }
  if (header->num_mipmaps > maxmip || header->num_mipmaps == 0) {
    return peInvalidArguments;
  }
  if (header->num_surfaces != 1 && header->num_surfaces != 6) {
    return peInvalidArguments;
  }
  if (header->width > 0x8000) {
    return peTooBigForMachine;
  }
  if (header->alignment & (header->alignment-1)) {
    return peInvalidArguments;
  }
  return peSuccess;
}

ptf_err ptf_texture_create(ptf_texture_handle *o_tex)
{
  try {
    *o_tex = new ptf_texture_t;
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

ptf_err ptf_texture_destroy(ptf_texture_handle tex)
{
  try {
    if (!tex->validate()) return peInvalidArguments;
    delete tex;
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

ptf_err ptf_texture_resize(ptf_texture_handle tex, ptf_texture_data const * header)
{
  try {
    if (!tex->validate()) return peInvalidArguments;
    ptf_err e = validate_header(header);
    if (e != peSuccess) {
      return e;
    }
    tex->clear();
    tex->header = *header;
    memset(&tex->header.reserved1, 0, sizeof(tex->header.reserved1));

    tex->data.resize(header->num_surfaces);
    for (size_t i = 0; i < header->num_surfaces; ++i) {
      tex->data[i].resize(header->num_mipmaps);
    }
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

ptf_err ptf_texture_set_data(ptf_texture_handle tex, ptf_8 surface, ptf_8 mipmap,
    void const * data, size_t size)
{
  try {
    if (!tex->validate()) return peInvalidArguments;
    if (size > 0x10000000) {
      return peTooBigForMachine;
    }
    if (surface >= tex->header.num_surfaces) {
      return peInvalidArguments;
    }
    if (mipmap >= tex->header.num_mipmaps) {
      return peInvalidArguments;
    }
    tex->data[surface][mipmap].resize(size);
    if (size > 0) {
      memcpy(&tex->data[surface][mipmap][0], data, size);
    }
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

ptf_err ptf_texture_get_info(ptf_texture_handle tex, ptf_texture_data *o_header)
{
  try {
    if (!tex->validate()) return peInvalidArguments;
    *o_header = tex->header;
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

ptf_err ptf_texture_get_data(ptf_texture_handle tex, ptf_8 surface, ptf_8 mipmap,
    void const **o_data, size_t *o_size)
{
  try {
    if (!tex->validate()) return peInvalidArguments;
    if (surface >= tex->header.num_surfaces) {
      return peInvalidArguments;
    }
    if (mipmap >= tex->header.num_mipmaps) {
      return peInvalidArguments;
    }
    *o_size = tex->data[surface][mipmap].size();
    if (*o_size) {
      *o_data = &tex->data[surface][mipmap][0];
    }
    else {
      *o_data = 0;
    }
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

ptf_err ptf_texture_read_from_file(ptf_texture_handle tex, ptf_file_handle file, ptf_fixed_ref const &fref)
{
  try {
    if (!tex->validate()) return peInvalidArguments;
    ptf_err e;
    ptf_64 size = 0;
    ptf_record_ref ref;
    memset(&ref, 0, sizeof(ref));
    ref.type = ptTextureData;
    ref.index = fref.index;
    e = ptf_file_get_record_size(file, &ref, &size);
    if (e != peSuccess) {
      return e;
    }
    if (size > 0x10000000) {
      return peTooBigForMachine;
    }
    if (size < sizeof(ptf_texture_data)) {
      return peInvalidFile;
    }
    ptf_texture_data header;
    e = ptf_file_read_record(file, &ref, 0, &header, sizeof(header));
    if (e != peSuccess) {
      return e;
    }
    if ((ptf_64)header.data_offset + header.data_size > size) {
      return peInvalidFile;
    }
    if (header.data_offset < sizeof(header) + sizeof(ptf_texture_mipmap_desc) * 
        header.num_mipmaps * header.num_surfaces) {
      return peInvalidFile;
    }
    e = validate_header(&header);
    if (e != peSuccess) {
      return e;
    }
    std::vector<ptf_texture_mipmap_desc> mips;
    mips.resize((ptf_32)header.num_mipmaps * (ptf_32)header.num_surfaces);
    e = ptf_file_read_record(file, &ref, sizeof(header), &mips[0], sizeof(ptf_texture_mipmap_desc)*mips.size());
    if (e != peSuccess) {
      return e;
    }
    size_t i = 0;
    for (i = 0; i < mips.size(); ++i) {
      if ((ptf_64)mips[i].start_offset + mips[i].size > header.data_size) {
        return peInvalidFile;
      }
    }
    e = ptf_texture_resize(tex, &header);
    if (e != peSuccess) {
      return e;
    }
    i = 0;
    for (ptf_8 surface = 0; surface < header.num_surfaces; ++surface) {
      for (ptf_8 mipmap = 0; mipmap < header.num_mipmaps; ++mipmap) {
        tex->data[surface][mipmap].resize(mips[i].size);
        if (mips[i].size) {
          e = ptf_file_read_record(file, &ref, (ptf_64)header.data_offset + mips[i].start_offset, 
              &tex->data[surface][mipmap][0], mips[i].size);
          if (e != peSuccess) {
            return peInvalidFile;
          }
        }
        ++i;
      }
    }
  }
  catch (...) {
    return peException;
  }
  return peSuccess;
}

static ptf_err write_texture(ptf_file_handle file, ptf_texture_handle tex)
{
  try {
    if (!tex->validate()) return peInvalidArguments;
    ptf_err e;
    std::vector<ptf_texture_mipmap_desc> mips;

    mips.resize((ptf_32)tex->header.num_mipmaps * (ptf_32)tex->header.num_surfaces);

    tex->header.data_offset = sizeof(tex->header);
    tex->header.data_offset += (ptf_32)mips.size()*sizeof(ptf_texture_mipmap_desc);
    size_t padding = ((tex->header.data_offset + tex->header.alignment - 1)
        & -tex->header.alignment) - tex->header.data_offset;
    tex->header.data_offset += (ptf_32)padding;

    size_t i = 0;
    ptf_32 start_offset = 0;
    for (ptf_8 surface = 0; surface < tex->header.num_surfaces; ++surface) {
      for (ptf_8 mipmap = 0; mipmap < tex->header.num_mipmaps; ++mipmap) {
        mips[i].start_offset = start_offset;
        mips[i].size = (ptf_32)(tex->data[surface][mipmap].size() + tex->header.alignment - 1)
            & -tex->header.alignment;
        start_offset += mips[i].size;
        ++i;
      }
    }
    tex->header.data_size = start_offset;

    e = ptf_file_write_record(file, &tex->header, sizeof(tex->header));
    if (e != peSuccess) {
      return e;
    }

    e = ptf_file_write_record(file, &mips[0], (ptf_32)mips.size()*sizeof(ptf_texture_mipmap_desc));
    if (e != peSuccess) {
      return e;
    }

    //  If file alignment is bigger than my current position, round up.
    e = ptf_file_write_pad(file, padding);
    if (e != peSuccess) {
      return e;
    }

    i = 0;
    for (ptf_8 surface = 0; surface < tex->header.num_surfaces; ++surface) {
      for (ptf_8 mipmap = 0; mipmap < tex->header.num_mipmaps; ++mipmap) {
        if (mips[i].size) {
          e = ptf_file_write_record(file, &tex->data[surface][mipmap][0],
              (ptf_32)tex->data[surface][mipmap].size());
          if (e != peSuccess) {
            return e;
          }
          e = ptf_file_write_pad(file, mips[i].size - tex->data[surface][mipmap].size());
          if (e != peSuccess) {
            return e;
          }
        }
        ++i;
      }
    }
  }
  catch (...) {
    return peException;
  }

  return peSuccess;
}

ptf_err ptf_texture_append_to_file(ptf_texture_handle tex, ptf_file_handle file, ptf_fixed_ref *o_fr)
{
  try {
    if (!tex->validate()) return peInvalidArguments;
    ptf_record_ref rref;
    ptf_err e = ptf_file_new_record(file, ptTextureData);
    if (e != peSuccess) {
      return e;
    }
    e = write_texture(file, tex);
    if (e != peSuccess) {
      return e;
    }
    memset(&rref, 0, sizeof(rref));
    e = ptf_file_commit_record(file, &rref);
    if (e == peSuccess) {
      o_fr->index = rref.index;
    }
    return e;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_texture_update_to_file(ptf_texture_handle tex, ptf_file_handle file, ptf_fixed_ref const &ref)
{
  try {
    if (!tex->validate()) return peInvalidArguments;
    ptf_record_ref rref;
    memset(&rref, 0, sizeof(rref));
    rref.type = ptTextureData;
    rref.index = ref.index;
    ptf_err e = ptf_file_replace_record(file, &rref);
    if (e != peSuccess) {
      return e;
    }
    e = write_texture(file, tex);
    if (e != peSuccess) {
      return e;
    }
    e = ptf_file_commit_record(file, &rref);
    return e;
  }
  catch (...) {
    return peException;
  }
}
